package sita;

import java.awt.Color;
import java.awt.Point;
import java.util.Hashtable;
import java.util.ArrayList;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.Collections;
import java.util.Enumeration;
import javax.swing.JSplitPane;
import org.sanskritlibrary.CatalogRecord;
import org.sanskritlibrary.FSM;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.ResourceInputStream;
import org.sanskritlibrary.ResourceInputStreamImpPath;
import org.sanskritlibrary.TabInfo;
import org.sanskritlibrary.Text;
import org.sanskritlibrary.webservice.WebServices;

/*
puranas            59
BagavadgItA        20
itihAsasamuccaya    1
BAgavataSlokawIkA   1
azwAdaSaSlokigItA   1
mahABArata         51
uttaragItA          1
SivasahasranAma     1
 */
public class ApplicationState
{
    static String[] encodingTypes = new String[] {"deva", "roman", "slp1", "telugu"};
    static String virtualDirectoryName = "virtual";

    ArrayList<String> textNames;
    Hashtable<String, ArrayList<String>> manuscripts = new Hashtable<String, ArrayList<String>>();

    public Hashtable<String, ArrayList<String>> texts;
    BasicConfig basicConfig;
    Catalog catalog;
    Pages pages;

    String currentTextType;
    CatalogRecord currentCatalogRecord;
    String currentFileName;
    private boolean displaySuccessful = false;      // set to true after first manuscript page with digital text displayed
    private boolean useDefaultWidth = false;

    int lastTextIdx = -1;
    int lastManuscriptIdx = -1;
    int lastPageIdx = -1;

    double splitDivider = 0.0;
    int orientation = -1;
    int fontsize = 0;
    int currentTabNumber = 0;
    String[] currentEncodings = null;

    int nextAnnotationNumber = 0;

    StringBuilder sb = new StringBuilder();

    String lastText;
    String lastManuscript;
    boolean manuscriptChanged = true;
    String lastPage;
    String currentAnnotationDirectory;
    String currentAnnotationPage;
    String imageFilename;
    boolean allowModification = false;
    private int moveToAnnotationNumber = -1;
    private boolean reviewOnly = true;
    
    Annotations annotations = new Annotations(this);
    private AnnotationInfo currentAnnotation;
    Annotated annotater;

    public Hashtable<String, ArrayList<VerseAnnotation>> verseIndex = new Hashtable<String, ArrayList<VerseAnnotation>>();
    public ArrayList<Hashtable<String, String>> annotationsSummary;
    public ArrayList<Hashtable<String, String>> commentsSummary;
    public ArrayList<RawComment> commentsList;
    public ArrayList<VerseAnnotationCollection> differencesSummary;
    Hashtable<String, Boolean> fileExists = new Hashtable<String, Boolean>();
    private ArrayList<TextMap>[] textmaps = null;
    private ArrayList<String>[] tablines = null;
    private Hashtable<String, TransformMap> transformMaps = null;

    public ApplicationState(Annotated annotater) throws Exception
    {
        this.annotater = annotater;
        loadState();            // last page processed.
        basicConfig = new BasicConfig("config.xml", "machineGenerated.xml");
        pages =  new Pages(String.format("%s/pages.xml", basicConfig.getCatalogDirectory()));
        catalog = new Catalog(String.format("%s/catalog.xml", basicConfig.getCatalogDirectory()), pages, basicConfig.getDefaultEncoding());
        currentAnnotation = new AnnotationInfo();          // assume first tab is the default when program first runs.
        reviewOnly = basicConfig.getShowReviewOnly();   // If review only command is visible, force user to turn off review only.
    }

    public void setReviewOnly (boolean b) {
        reviewOnly = b;
    }

    public boolean getReviewOnly () {
        return reviewOnly;
    }

    public boolean getShowReviewOnly () {
        return basicConfig.getShowReviewOnly();
    }

    public int getCurrentAnnotationTabNumber () {
        return currentAnnotation.getTabNumber();
    }

    public void setTextMap (ArrayList<TextMap> textmap) {
        textmaps[getCurrentTabNumber()] = textmap;
    }

    public TabInfo getCurrentTabInfo () {
        return getCurrentCatalogRecord().tabs.get(currentTabNumber);
    }


    public ArrayList<TextMap> getTextMap () {
        return textmaps[getCurrentTabNumber()];
    }
    
    public void setTextLines (ArrayList<String> lines) {
        tablines[getCurrentTabNumber()]= lines;
    }


    public ArrayList<String> getTextLines () {
        return tablines[getCurrentTabNumber()];
    }
    


    public int getNumTabs () {
        return currentCatalogRecord.tabs.size();
    }

    public String getTabName (int tabnum) {
        return currentCatalogRecord.tabs.get(tabnum).name;
    }


    public void setCurrentTabNumber (int tabNumber) {
        currentTabNumber = tabNumber;
    }

    public int getCurrentTabNumber () {
        return currentTabNumber;
    }

    public void setDisplaySuccessful(boolean b)
    {
        displaySuccessful = b;
    }

    public boolean isMacOSX()
    {
        return System.getProperty("os.name").equals("Mac OS X");
    }

    public ArrayList<String> getAllManuscriptDirectories()
    {
        ArrayList<String> a = new ArrayList<String>();
        Hashtable<String, DefaultInfo> defaults = basicConfig.getDefaults();
        Enumeration<String> en = defaults.keys();
        while (en.hasMoreElements())
        {
            DefaultInfo di = defaults.get(en.nextElement());
            a.add(di.dirname);
        }

        Collections.sort(a);

        return a;
    }

    public int getDefaultWidth()
    {
        DefaultInfo di = basicConfig.getDefaultInfo(lastManuscript);
        if (di == null)
            return annotater.getDefaultDisplayWidth();
        else
            return di.resolvedWidth;
    }

    public Color getDefaultPointColor()
    {
        DefaultInfo di = basicConfig.getDefaultInfo(lastManuscript);
        if (di == null)
            return Color.red;
        else
            return di.pointColor;
    }


    public double getDefaultDivider()
    {
        if (splitDivider != 0.0)         // use temporary value
            return splitDivider;
        else
        {
            DefaultInfo di = basicConfig.getDefaultInfo(lastManuscript);
            if (di == null)
                return 0.5;
            else
                return di.divider;
        }
    }

    public int getDefaultOrientation()
    {
        if (orientation != -1)
            return orientation;         // use temporary value;
        else
        {
            DefaultInfo di = basicConfig.getDefaultInfo(lastManuscript);
            if (di == null)
                return JSplitPane.HORIZONTAL_SPLIT;
            else
            {
                if (di.orientation == '|')
                    return JSplitPane.HORIZONTAL_SPLIT;
                else
                    return JSplitPane.VERTICAL_SPLIT;
            }
        }
    }

    public int getDefaultFontsize()
    {
        if (fontsize != 0)
            return fontsize;
        else
        {
            DefaultInfo di = basicConfig.getDefaultInfo(lastManuscript);
            if (di == null)
                return 12;
            else
                return di.fontsize;

        }
    }

    // Called if encoding changed from menu bar. Will only be called if there are no xml files in the current manuscript directory that contain selected text offsets
    public void setEncoding(String encoding) throws Exception
    {
        currentEncodings[currentTabNumber] = encoding;
        saveEncodingFile(getAnnotationsCurrentDirectory(), currentEncodings);
    }


    public int getNumberOfEncodings () {
        return this.currentCatalogRecord.tabs.get(currentTabNumber).legalEncodings.length;
    }

    public String getEncoding()
    {
        return currentEncodings[currentTabNumber];            // this should always equal the contents of the "encodingType" file in the current manuscript directory.
    }

    public boolean isDisplaySuccessful()
    {
        return displaySuccessful;
    }

    public ArrayList<CommentInfo> getCommentTags()
    {
        return basicConfig.getCommentInfo();
    }

    public CommentInfo getCommentTag(String tag)
    {
        for (CommentInfo ci : basicConfig.getCommentInfo())
        {
            if (ci.tag.equals(tag))
                return ci;
        }

        return null;        // should never happen
    }

    public boolean updateLocalVersionsOfDefaults() throws Exception
    {
        boolean rc = manuscriptChanged;
        if (manuscriptChanged)
        {
            DefaultInfo di = basicConfig.getDefaultInfo(lastManuscript);
            if (di == null)
            {
                orientation = JSplitPane.HORIZONTAL_SPLIT;
                splitDivider = 0.5;
                fontsize = 12;
            }
            else
            {
                orientation = di.orientation == '|' ? JSplitPane.HORIZONTAL_SPLIT : JSplitPane.VERTICAL_SPLIT;
                splitDivider = di.divider;
                fontsize = di.fontsize;
            }

            // get encoding to use, possibly creating a "encodingType" file
            currentEncodings = readEncodingFile(getAnnotationsCurrentDirectory());
            setPossibleEncodings();
            currentAnnotation.setEncoding(getEncoding());

            manuscriptChanged = false;
        }

        return rc;
    }

    private void setPossibleEncodings() {
        Hashtable<String, Integer> encodings;
        String[] physicalFiles = new File(basicConfig.getTextDirectory()).list();
        String vd = String.format("%s/%s", basicConfig.getTextDirectory(), virtualDirectoryName);
        File vdf = new File(vd);
        if (!vdf.exists()) {
            vdf.mkdirs();
        }

        String[] virtualFiles = new File(vd).list();
        ArrayList<String> files = new ArrayList<String>();
        for (String s : physicalFiles) {
            files.add(s);
        }
        for (String s: virtualFiles) {
            files.add(s);
        }

        for (TabInfo tab : currentCatalogRecord.tabs) {
            encodings = new Hashtable<String, Integer>();
            int numfiles = 0;
            for (Text text : tab.texts) {
                if (text.type == Text.Type.TextRef) {
                    try {
                        numfiles++;
                        String filename = text.ref;
                        String pattern = String.format("^%s\\.(.*)", filename);
                        Regex re = new Regex("", pattern);
                        for (String file : files) {
                            re.reset(file);
                            if (re.find()) {
                                String ext = re.getGroup(1);
                                Integer count = encodings.get(ext);
                                if (count == null) {
                                    count = new Integer(1);
                                } else {
                                    count = new Integer(count.intValue() + 1);
                                }
                                encodings.put(ext, count);
                            }
                        }
                    } catch (Exception ex) {
                        System.err.println(ex.getMessage());
                    }
                }
            }
            
            Enumeration<String> en = encodings.keys();
            try {
                while (en.hasMoreElements()) {
                    String key = en.nextElement();
                    if (encodings.get(key) != numfiles) {
                        encodings.remove(key);
                    }
                }
            } catch (Exception ex) {
                Logger.writeLog(ex.getMessage());
            }

            if (encodings.size() == 0) {
                Logger.writeLog(String.format("Tab %s of %s does not have an encoding shared by all its files.", tab.name, currentFileName));
            }
            tab.setLegalEncodings(encodings);
        }
    }


    public boolean getEncodingExists(String enc)
    {
        TabInfo ti = currentCatalogRecord.tabs.get(currentTabNumber);
        for (String le : ti.legalEncodings) {
            if (le.equals(enc)) {
                return true;
            }
        }
        return false;
    }

    public String getImageFilename(String manuscriptDirectory, String abbrev)
    {
        int n = 0;
        abbrev = abbrev.trim();
        if (manuscriptDirectory == null)
            manuscriptDirectory = lastManuscript;
        ArrayList<String> ps = currentCatalogRecord.getPages();
        for (String p : ps)
        {
            if (p.endsWith(abbrev))
                return p;
        }

        return null;
    }

    public void clearAnnotations () {
        for (AnnotationInfo ai : annotations.getAnnotations()) {
            if (ai != currentAnnotation) {
                annotations.removeAnnotation(ai);
            }
        }
    }

    public String getComment()
    {
        return currentAnnotation.getComment();
    }

    public void setComment(String comment)
    {
        currentAnnotation.setComment(comment);
    }


    public void setUseDefaultWidth(boolean b)
    {
        useDefaultWidth = b;
    }

    public boolean getUseDefaultWidth()
    {
        return basicConfig.getUseDefaultWidth();
    }

    public int getResolvedWidth()
    {
        if (useDefaultWidth)
            return getDefaultWidth();
        else
            return annotations.getResolvedWidth();
    }

    // If b is true then return width that is currently on screen. Otherwise possibly return default width.
    public int getResolvedWidth(boolean b)
    {
        if (!b)
            return getResolvedWidth();
        else
            return annotations.getResolvedWidth();
    }

    public void setResolvedWidth(int resolvedWidth)
    {
        annotations.setResolvedWidth(resolvedWidth);
    }


    public void forceReloadOfAnnotations()
    {
        currentAnnotationDirectory = null;
    }

    public Annotations getAnnotations() throws Exception
    {
        useDefaultWidth = basicConfig.getUseDefaultWidth();     // in case zooming in and out on last page forced it to be false.

        nextAnnotationNumber = 0;
        if (!getAnnotationsCurrentDirectory().equals(currentAnnotationDirectory) || !lastPage.equals(currentAnnotationPage))
        {
            String subdir = String.format("%s/%s", lastText, lastManuscript);
            annotations.loadAnnotations(basicConfig.getAnnotationsRootDirectory(), subdir, lastPage + ".xml");
            if (annotations.getResolvedWidth() == 0)            // probably won't ever happen
                annotations.setResolvedWidth(annotater.getDefaultDisplayWidth());

            currentAnnotationDirectory = getAnnotationsCurrentDirectory();
            currentAnnotationPage = lastPage;

            if (currentAnnotation == null) {
                currentAnnotation = new AnnotationInfo();       // add an empty annotation
            }
            else {
                currentAnnotation.clear();              // one less thing to garbage collect.
                currentAnnotation.setTabNumber(getCurrentTabNumber());
            }

            annotations.add(currentAnnotation);
        }
        
        return annotations;
    }

    public String getAnnotationsRootDirectory()
    {
        return basicConfig.getAnnotationsRootDirectory();
    }

    public String getAnnotationsCurrentDirectory()
    {
        return String.format("%s/%s/%s", basicConfig.getAnnotationsRootDirectory(), lastText, lastManuscript);
    }

    public String getAnnotationsBackupDirectory()
    {
        return basicConfig.getAnnotationsBackupDirectory();
    }

    public boolean haveSelectedText() throws Exception
    {
        return haveSelectedText(getAnnotationsCurrentDirectory());
    }

    public boolean haveSelectedText(String dirname) throws Exception
    {
        TextAnalyzer ta = new TextAnalyzer(this);
        return ta.haveSelectedText(dirname);
    }


    public AnnotationInfo getAnnotation(int n)
    {
        return annotations.getAnnotation(n);
    }

    public String[] getTabEncodings () {

        String[] a = new String[getNumTabs()];
        for (int i=0; i<a.length; i++) {
            a[i] = "roman";
        }
        return a;
    }

    public void denormalizeAnnotations(int actualWidth)
    {
        int size = annotations.getSize();
        for (int i=0; i<size; i++)
        {
            AnnotationInfo ai = annotations.getAnnotation(i);
            if (!ai.isEmpty())
                ai.denormalize(actualWidth, annotations.getResolvedWidth());      // map points from actualWidth down to resolutionWidth
        }
    }

    public void normalizeAnnotations(int resolutionWidth, int actualWidth)
    {
        int size = annotations.getSize();
        for (int i=0; i<size; i++)
        {
            AnnotationInfo ai = annotations.getAnnotation(i);
            if (!ai.isEmpty())
                ai.normalize(resolutionWidth, actualWidth);      // map point from resolution width up to actualWidth
        }
    }

    // normalize the current annotation using actualWidth and oldWidth, then denormalize it using actualWidth and newWidth
    public void adjustAnnotations(int actualWidth, int oldWidth, int newWidth)
    {
        int size = annotations.getSize();
        for (int i = 0; i < size; i++)
        {
            AnnotationInfo ai = annotations.getAnnotation(i);
            if (!ai.isEmpty())
                ai.adjustPoints(actualWidth, oldWidth, newWidth);
        }
    }

    public int getNumAnnotations()
    {
        int num = 0;
        if (annotations != null)
        {
            int size = annotations.getSize();
            for (int i=0; i<size; i++)
            {
                AnnotationInfo ai = annotations.getAnnotation(i);
                if (!ai.isEmpty())
                    num++;
            }
        }
        return num;
    }


    private void loadState()
    {
        try
        {
            BufferedReader br = IOAssistance.openReader("state.txt");
            String s = br.readLine();
            String[] f = s.split(",");
            if (f.length == 3)
            {
                lastTextIdx = Integer.parseInt(f[0]);
                lastManuscriptIdx = Integer.parseInt(f[1]);
                lastPageIdx = Integer.parseInt(f[2]);
            }
            br.close();
        }
        catch (Exception ex)        // no big deal.
        {
            lastTextIdx = -1;
            lastManuscriptIdx = -1;
            lastPageIdx = -1;
        }
    }

    public boolean getCanCloseWindow()
    {
        return basicConfig.getCanCloseWindow();
    }

    public boolean getSearchWithSLP1()
    {
        return basicConfig.getSearchWithSLP1();
    }

    public void saveState() throws Exception
    {
            FileWriter fw = IOAssistance.openUncachedWriter("state.txt");
            fw.write(String.format("%d,%d,%d", lastTextIdx, lastManuscriptIdx, lastPageIdx));
            fw.close();
    }

    public void saveDefaults(int dividerLocation, int frameWidth, int resolvedWidth, Color pointColor, int orientation) throws Exception
    {
        basicConfig.saveDefaults(lastManuscript, dividerLocation, frameWidth, resolvedWidth, pointColor, orientation, getDefaultFontsize());
        basicConfig.toXML();
    }


    public boolean reformat(String oldEncoding, String newEncoding, StringBuilder result)
    {
        TextAnalyzer ta = new TextAnalyzer(this);
        return ta.reformat(oldEncoding, newEncoding, result);
    }
    
    public void createVerseIndex(boolean useCurrentManuscript) throws Exception
    {
        TextAnalyzer ta = new TextAnalyzer(this);
        ta.createVerseIndex(verseIndex, useCurrentManuscript);
    }

    public void generateAnnotationsSummary(boolean currentManuscriptOnly) throws Exception
    {
        TextAnalyzer ta = new TextAnalyzer(this);
        annotationsSummary = ta.generateAnnotationsSummary(getAnnotationsCurrentDirectory(), currentManuscriptOnly);
    }

    public void generateCommentsSummary(boolean currentManuscriptOnly) throws Exception
    {
        TextAnalyzer ta = new TextAnalyzer(this);
        commentsSummary = ta.generateCommentsSummary(getAnnotationsCurrentDirectory(), currentManuscriptOnly);
    }

    public void generateDifferencesSummary(boolean currentManuscriptOnly) throws Exception
    {
        TextAnalyzer ta = new TextAnalyzer(this);
        differencesSummary = ta.generateDifferencesSummary(getAnnotationsCurrentDirectory(), getAnnotationsBackupDirectory(), currentManuscriptOnly);
    }

    public void searchForText(boolean useCurrentManuscript, String text, ArrayList<VerseAnnotationEx> foundImages) throws Exception
    {
        TextAnalyzer ta = new TextAnalyzer(this);
        ta.searchForText(useCurrentManuscript, text, foundImages);
    }


    public void saveEncodingFile(String path, String newEncoding) throws Exception {
        currentEncodings[currentTabNumber] = newEncoding;
        saveEncodingFile(path, currentEncodings);
    }

    public void saveEncodingFile(String path, String[] newEncodings) throws Exception
    {
        StringBuilder bfr = new StringBuilder();
        String delim = "";
        for (String e : newEncodings) {
            bfr.append(String.format("%s%s", delim, e));
            delim =",";
        }
        FileWriter fw = IOAssistance.openUncachedWriter(path + "/encodingType");
        fw.write(bfr.toString());
        fw.close();
    }

    public String[] readEncodingFile(String path) throws Exception
    {
        String[] encodings = null;
        String fn = path + "/encodingType";
        File f = new File(fn);
        if (f.exists())
        {
            BufferedReader br = IOAssistance.openReader(fn);
            String encoding = br.readLine();
            encodings = encoding.split(",");
            br.close();
        }
        else            // save an "encodingType" file with the default encoding
        {
            File p = new File(path);
            p.mkdirs();
            encodings = new String[this.getNumTabs()];
            int i = 0;
            for (TabInfo ti : this.currentCatalogRecord.tabs) {
                encodings[i++] = ti.defaultEncoding;
            }
            saveEncodingFile(path, encodings);
        }

        return encodings;
    }


    public String getDefaultEncoding()
    {
        return basicConfig.getDefaultEncoding();
    }

    public void setCurrentTextType(String name)
    {
        currentTextType = name;
    }

    public String getCurrentTextType()
    {
        return currentTextType;
    }

    public void setCurrentCatalogRecord(CatalogRecord cr)
    {
        currentCatalogRecord = cr;
        textmaps = new ArrayList[cr.tabs.size()];
        tablines = new ArrayList[cr.tabs.size()];
    }

    public CatalogRecord getCurrentCatalogRecord()
    {
        return currentCatalogRecord;
    }

    public void setCurrentFileName(String name)
    {
        currentFileName = name;
    }

    public String getCurrentFileName()
    {
        return currentFileName;
    }

    
    public void setLastTextIdx(int idx)
    {
        lastTextIdx = idx;
    }

    public int getLastTextIdx()
    {
        return lastTextIdx;       // e.g., BG
    }


    public void setLastText(String name)
    {
        lastText = name;
    }

    public String getLastText()
    {
        return lastText;       // e.g., BG
    }


    public void setLastManuscriptIdx(int idx)
    {
        lastManuscriptIdx = idx;
    }

    public int getLastManuscriptIdx()
    {
        return lastManuscriptIdx;      // e.g., 2366
    }


    public void setLastManuscript(String name)
    {
        if (!name.equals(lastManuscript))
            manuscriptChanged = true;
        lastManuscript = name;
    }

    public String getLastManuscript()
    {
        return lastManuscript;      // e.g., 2366
    }

    public void setLastPageIdx(int idx)
    {
        lastPageIdx = idx;
    }

    public int getLastPageIdx()
    {
        return lastPageIdx;       // page 3
    }

    public void setLastPage(String name)
    {
        lastPage = name;
    }

    public String getLastPage()
    {
        return lastPage;       // page 3
    }

    public void update() throws Exception
    {
        basicConfig.toXML();
    }

    public String getManuscriptRootDirectory()
    {
        return basicConfig.getManuscriptRootDirectory();
    }

    public void setManuscriptRootDirectory(String dir) throws Exception
    {
        basicConfig.setManuscriptRootDirectory(dir);
    }

    public String getBackupManuscriptRootDirectory()
    {
        return basicConfig.getAnnotationsBackupDirectory();
    }

    public void setBackupManuscriptRootDirectory(String dir) throws Exception
    {
        basicConfig.setAnnotationsBackupDirectory(dir);
    }

    public String getCatalogDirectory()
    {
        return basicConfig.getCatalogDirectory();
    }

    public void setCatalogDirectory(String dir) throws Exception
    {
        basicConfig.setCatalogDirectory(dir);
    }

    /*

    public String getBackupDirectory()
    {
        return basicConfig.getBackupDirectory();
    }

    public void setBackupDirectory(String dir) throws Exception
    {
        basicConfig.setBackupDirectory(dir);
    }
     */

    public String getTextDirectory()
    {
        return basicConfig.getTextDirectory();
    }

    public void setTextDirectory(String dir) throws Exception
    {
        basicConfig.setTextDirectory(dir);
    }

    public void setAnnotationsBackupDirectory (String dir) throws Exception {
        basicConfig.setAnnotationsBackupDirectory(dir);
    }

    public ArrayList<String> getManuscriptTypes()
    {
        if (textNames == null)
            textNames = catalog.getManuscriptTypes();

        return textNames;
    }

    public String getManuscriptType () {
        return currentCatalogRecord.manuscriptType;
    }


    public ArrayList<CatalogRecord> getDirectories(String textName)
    {
        return catalog.getDirectories(textName);
    }

// Current annotation support
    public void setMoveToAnnotationNumber(int n)
    {
        this.moveToAnnotationNumber = n;
    }

    public int getMoveToAnnotationNumber()
    {
        return this.moveToAnnotationNumber;
    }

    public AnnotationInfo getCurrentAnnotation()
    {
        return currentAnnotation;
    }

    public int getNumberOfCurrentAnnotation()
    {
        int n = 0;

        if (!currentAnnotation.isEmpty())
        {
            int cnt = annotations.getAnnotations().size();
            for (int i=0; i<cnt; i++)
            {
                if (annotations.getAnnotation(i) == currentAnnotation)
                    return i + 1;
            }
        }

        return n;
    }

    public boolean hasManuscriptAnnotation()
    {
        return currentAnnotation.hasManuscriptAnnotation();
    }

    public void setPoint1(Point p)
    {
        currentAnnotation.setPoint1(p);
    }

    public void setPoint1(int x, int y)
    {
        currentAnnotation.setPoint1(x, y);
    }

    public Point getPoint1()
    {
        return currentAnnotation.getPoint1();
    }

    public void setPoint2(Point p)
    {
        currentAnnotation.setPoint2(p);
    }
    
    public void setPoint2(int x, int y)
    {
        currentAnnotation.setPoint2(x, y);
    }

    public Point getPoint2()
    {
        return currentAnnotation.getPoint2();
    }


    /*
    public void setTextStart(Point start)
    {
        currentAnnotation.setTextStart(start);
    }
     */

    public TextPoint getTextStart(ArrayList<TextMap> textmap)
    {
        return currentAnnotation.getTextStart(textmap);
    }

    public void clearTextMarkers () {
        currentAnnotation.clearTextMarkers();
    }

    public void addTextMarker (String filename, TextPoint start, TextPoint end) {
        currentAnnotation.addTextMarker(filename, start, end);
    }
/*
    public void setTextEnd(int end)
    {
        currentAnnotation.setTextEnd(end);
    }
*/
    
    public TextPoint getTextEnd(ArrayList<TextMap> textmap)
    {
        return currentAnnotation.getTextEnd(textmap);
    }

    public void setLeftBracketExists(boolean leftBracketExists)
    {
        currentAnnotation.setLeftBracketExists(leftBracketExists);
    }

    public boolean getLeftBracketExists()
    {
        return currentAnnotation.getLeftBracketExists();
    }

    public void setRightBracketExists(boolean finishExists)
    {
        currentAnnotation.setRightBracketExists(finishExists);
    }

    public boolean getRightBracketExists()
    {
        return currentAnnotation.getRightBracketExists();
    }

    public boolean swapBrackets()
    {
        return currentAnnotation.swapBrackets();
    }

    public void setBoxExists(boolean boxExists)
    {
        currentAnnotation.setBoxExists(boxExists);
    }

    public boolean getBoxExists()
    {
        return currentAnnotation.getBoxExists();
    }

    public int getPointSize()
    {
        return annotations.getPointSize();
    }

    public void setPointColor(Color pointColor)
    {
        annotations.setPointColor(pointColor);
    }

    public Color getPointColor()
    {
        return annotations.getPointColor();
    }

    public void clearCurrentConfiguration()
    {
        currentAnnotation.clear();
    }

    public void setCurrentAnnotationInfo(AnnotationInfo ai)
    {
        currentAnnotation = ai;
    }

    public boolean isCurrentAnnotationEmpty()
    {
        setComment(annotater.getComment());
        return currentAnnotation.isEmpty();
    }

    public void removeCurrentAnnotation()
    {
        if (currentAnnotation.isEmpty())        // already empty
            return;

        annotations.removeAnnotation(currentAnnotation);
        resetCurrentAnnotation();       // find or create another empty annotation
    }

    public void resetCurrentAnnotation()
    {
        currentAnnotation = null;
        int size = annotations.getSize();
        for (int i=0; i<size; i++)
        {
            AnnotationInfo ai = annotations.getAnnotation(i);
            if (ai.isEmpty())
            {
                currentAnnotation = ai;         // use already existing empty one
                break;
            }
        }

        if (currentAnnotation == null)          // no already existing empty one, so create new one and add it to annotations
        {
            currentAnnotation = new AnnotationInfo();
            annotations.add(currentAnnotation);
        }
    }

    public void saveAnnotations(int resolvedWidth, int actualWidth) throws Exception {
        if (!reviewOnly) {
            setComment(annotater.getComment());         // update current annotation with more recent comment, if any.
            normalizeAnnotations(resolvedWidth, actualWidth);
            annotations.toXml(resolvedWidth);
        }
    }

    public void setNextAnnotationNumber(int n) {
        nextAnnotationNumber = n;
    }

    public int getNextAnnotationNumber() {
        return nextAnnotationNumber;
    }

    public Annotated getAnnotater() {
        return annotater;
    }

    public void setImageFilename(String imageFilename) {
        this.imageFilename = imageFilename;
    }

    public String getImageFilename() {
        return imageFilename;
    }

    public void setAllowModification(boolean b)
    {
        allowModification = b;
    }

    public boolean getAllowModification()
    {
        return allowModification;
    }

    public void setSplitDivider(int dividerLocation, int frameWidth)
    {
        splitDivider = (double)dividerLocation/(double)frameWidth;
    }

    public double getSplitDivider()
    {
        return splitDivider;
    }

    public void setOrientation(int orientation)
    {
        this.orientation = orientation;
    }

    public int getOrientation()
    {
        return orientation;
    }

    public void setFontSize(int fontsize)
    {
        this.fontsize = fontsize;
    }

    public boolean getUseCDATA()
    {
        return basicConfig.getUseCDATA();
    }

    public ArrayList<SegmentInfo> getVirtualFile (String name) {
        return basicConfig.getVirtualFile(name);
    }
    
    public ArrayList<OffsetInfo> getOffsets (String name) {
        return basicConfig.getOffsets(name);
    }

    public void updateVirtualFiles() throws Exception {
        Hashtable<String, FSM[]> fsmh = null;
        ResourceInputStream ris = null;
        StringBuilder errors = new StringBuilder();

        Hashtable<String, ArrayList<SegmentInfo>> virtualFiles = basicConfig.getVirtualFiles();
        Hashtable<String, ArrayList<OffsetInfo>> oldVirtualFileOffsets = basicConfig.getVirtualFileOffsets();
        Hashtable<String, ArrayList<OffsetInfo>> virtualFileOffsets = new Hashtable<String, ArrayList<OffsetInfo>>();
        ArrayList<OffsetInfo> offsets = null;

        // Create virtual subdirectory if it doesn't exist.
        File virtualDirectory = new File(String.format("%s/virtual", getTextDirectory()));
        if (!virtualDirectory.exists()) {
            virtualDirectory.mkdirs();
        }

        // Get the supported encodings
        ArrayList<String> supportedEncodings = basicConfig.getSupportedEncodings();
        int numEncodings = supportedEncodings.size();

        // Create TransformMaps for all supported encodings
        if (transformMaps == null) {
            transformMaps = new Hashtable<String, TransformMap>();
            fsmh = new Hashtable<String, FSM[]>();
            ris = new ResourceInputStreamImpPath("transcoders");
            for (String encodingName : supportedEncodings) {
                if (!encodingName.startsWith("slp")) {
                    transformMaps.put(encodingName,
                            new TransformMap(String.format("*:%s->slp1", encodingName), ris, fsmh));
                }
            }
        }

        Enumeration<String> en = virtualFiles.keys();
        while (en.hasMoreElements()) {              // for each virtual file type (no extension)
            String key = en.nextElement();
            for (int i = 0; i < numEncodings; i++) {  // for each supported encoding (extension)
                String encodingName = supportedEncodings.get(i);
                offsets = new ArrayList<OffsetInfo>();
                if (i == 0) {       // save offsets for first encoding. Offset should be the same for all others.
                    virtualFileOffsets.put(key, offsets);
                }
                ArrayList<SegmentInfo> sis = virtualFiles.get(key);     // segment info is the same for each encoding type
                String vf = String.format("%s/%s.%s",
                        virtualDirectory.getAbsolutePath(), key, encodingName);
                BufferedWriter bw = IOAssistance.openWriter(vf);

                int lineno = 0;         // line number in virtual file
                for (SegmentInfo si : sis) {        // for each segment
                    OffsetInfo oi = new OffsetInfo(lineno);
                    offsets.add(oi);            // segment starts at this line number
                    if (si.isBookmark() || si.isComment()) {        // if it is a bookmark or comment segment
                        String[] lines = si.content.split("\n");
                        for (String line : lines) {
                            lineno++;
                            bw.write(line);
                            bw.write("\n");
                        }
                    } else if (si.isText()) {
                        String tf = String.format("%s/%s.%s", getTextDirectory(), si.source, encodingName);

                        if (new File(tf).exists() == false) {
                            Utils.displayWarning(String.format("%s does not exist. Virtual file %s.%s not created",
                                    tf, key, encodingName));
                            bw.close();
                            new File(vf).delete();      // source missing, can't create virtual file for this encoding type.
                            break;
                        }

                        BufferedReader br = IOAssistance.openReader(tf);    // open the source file
                        Regex re = null;
                        String line = null;
                        if (si.matcher != null) {
                            re = new Regex("", si.matcher);
                        }
                        boolean firstLineFound = false;
                        boolean lastLineFound = false;
                        boolean writeLine = false;
                        int firstLine = -1;
                        int lastLine = -1;
                        int nextLine = 0;
                        while ((line = br.readLine()) != null && !lastLineFound) {
                            if (re != null) {
                                re.reset(line);
                                if (re.find()) {
                                    String verseNumber = re.getGroup(0);
                                    if (si.mapspecs != null && si.mapspecs.get(encodingName) != null) {
                                        verseNumber = WebServices.transformString(verseNumber,
                                                transformMaps.get(encodingName));
                                    }
                                    if (!firstLineFound) {
                                        firstLineFound = verseNumber.equals(si.start);
                                        if (firstLineFound) {
                                            firstLine = nextLine;
                                            writeLine = true;    // once this is set true it won't be set back to false in this loop. Instead when lastLineFound becomes true loop will exit.
                                        }
                                    }
                                    lastLineFound = verseNumber.equals(si.end);
                                }
                            } else {
                                writeLine = true;
                            }

                            lastLine = nextLine++;
                            if (writeLine) {
                                bw.write(line);
                                bw.write("\n");
                                lineno++;
                            }
                        }
                        br.close();
                        oi.sourceFile = si.source;
                        oi.startLineInSourceFile = firstLine;
                        oi.endLineInSourceFile = lastLine;
                    }
                }
                bw.close();   // close the virtual file

                // Now verify that new offsets for the just created virtual file match old offsets (if any)
                if (i == 0) {       // only check first offset, assume that other encodings have the same number of lines.
                    if (!compare(oldVirtualFileOffsets.get(key), offsets)) {
                        if (errors.length() == 0) {
                            errors.append("Warnings reported during creation of virtual files\n");
                        }
                        errors.append(String.format("New virtual file %s has different layout than old one\n", key));
                        errors.append("Offsets before: ");
                        String delim = "";
                        for (OffsetInfo oinfo : oldVirtualFileOffsets.get(key)) {
                            errors.append(String.format("%s%d", delim, oinfo.offsetInVirtualFile));
                            delim = ", ";
                        }

                        errors.append("\nOffsets after: ");
                        delim = "";
                        for (OffsetInfo oinfo : offsets) {
                            errors.append(String.format("%s%d", delim, oinfo.offsetInVirtualFile));
                            delim = ", ";
                        }
                        errors.append("\n");
                    }
                }
            }
        }

        if (errors.length() != 0) {
            Logger.writeLog(errors.toString());
            Utils.displayWarning("Warnings generated. See log.txt.");
        } else {
            Utils.displayWarning("Virtual files successfully created.");
        }
        basicConfig.setVirtualFileOffset(virtualFileOffsets);
    }

    private boolean compare (ArrayList<OffsetInfo> oldOffsets, ArrayList<OffsetInfo> newOffsets) {
        if (oldOffsets == null) {
            return true;
        } else if (newOffsets.size() != oldOffsets.size()) {
            return false;
        } else {
            int len = newOffsets.size();
            for (int i=0; i<len; i++) {
                OffsetInfo newoi = newOffsets.get(i);
                OffsetInfo oldoi = newOffsets.get(i);
                if (newoi.startLineInSourceFile != oldoi.startLineInSourceFile ||
                    newoi.endLineInSourceFile != oldoi.endLineInSourceFile) {
                    return false;
                }
            }
        }

        return true;
    }

    public void getBookmarks(ArrayList<BookmarkInfo> bookmarks, ArrayList<TextMap> textMap, ArrayList<String> lines) {
        int len = textMap.size();
        StringBuilder sbb = new StringBuilder();
        for (TextMap tm : textMap) {
            if (tm.type == Text.Type.Bookmark) {
                sbb.setLength(0);
                for (int i = tm.tabStartrow; i <= tm.tabEndrow; i++) {
                    sbb.append(lines.get(i));
                }
                bookmarks.add(new BookmarkInfo(sbb.toString(), tm.tabStartrow));
            }
        }
    }

    @Override public String toString () {
        return annotations.toString();
    }
}
