package sita;

import java.awt.Color;
import java.util.Hashtable;
import java.util.ArrayList;
import java.io.File;
import javax.swing.JSplitPane;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;

public class BasicConfig extends DefaultHandler //implements ErrorHandler, ContentHandler, DTDHandler, EntityResolver
{
    private boolean dirty = false;
    private Locator locator;
    private String defaultEncoding;
    private ArrayList<String> elements;
    private Hashtable<String, DefaultInfo> defaults;

    private String annotationsRootDirectory = "annotations";
    private String virtualFileName = null;
    private String segmentContent = "";
    private String segmentType = null;
    private String segmentSource = null;
    private String segmentMatcher = null;
    private String segmentStart = null;
    private String segmentEnd= null;
    private Hashtable<String, Boolean> segmentMapSpecs = new Hashtable<String, Boolean>();

    private Hashtable<String, ArrayList<SegmentInfo>> virtualFiles;
    private ArrayList<String> virtualFileNames = new ArrayList<String>();
    private ArrayList<SegmentInfo> virtualFile;
    private ArrayList<CommentInfo> commentInfo;
    private String supportedEncoding = "";
    private ArrayList<String> supportedEncodings = new ArrayList<String>();
    private CommentInfo currentComment = null;
    private boolean useCDATA = false;
    private boolean canCloseWindow = true;
    private boolean searchWithSLP1 = true;
    private boolean useDefaultWidth = false;
    private boolean showReviewOnly = false;
    private MachineGeneratedParser mgp;
    ApplicationState appstate;

    public BasicConfig(String filename, String defaultsFile) throws Exception
    {
        // Get defaults from defaults.xml
        mgp = new MachineGeneratedParser(defaultsFile);
        defaults = mgp.getDefaults();
        
        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();

        // Parse the input
        SAXParser saxParser = factory.newSAXParser();
        saxParser.parse( new File(filename), this);
    }

    public Hashtable<String, ArrayList<SegmentInfo>> getVirtualFiles () {
        return virtualFiles;
    }

    public ArrayList<SegmentInfo> getVirtualFile(String name) {
        return virtualFiles.get(name);
    }

    public ArrayList<String> getSupportedEncodings () {
        return supportedEncodings;
    }

    public Hashtable<String, ArrayList<OffsetInfo>> getVirtualFileOffsets () {
        return mgp.getVirtualFileOffsets();
    }

    public ArrayList<OffsetInfo> getOffsets(String name) {
        return mgp.getVirtualFileOffsets().get(name);
    }

    public void setVirtualFileOffset (Hashtable<String, ArrayList<OffsetInfo>> vfo) throws Exception {
        mgp.setVirtualFileOffsets(vfo);
        dirty = true;
        toXML();
    }

    public DefaultInfo getDefaultInfo(String name)
    {
        return defaults.get(name);
    }

    public String getDefaultEncoding()
    {
        return defaultEncoding;
    }

    public boolean getUseDefaultWidth()
    {
        return useDefaultWidth;
    }

    public Hashtable<String, DefaultInfo> getDefaults()
    {
        return defaults;
    }


    public void saveDefaults(String name, int dividerLocation, int frameWidth, int resolvedWidth, Color pointColor, int orientation, int fontsize)
    {
        double divider = (double)dividerLocation/(double)frameWidth;
        char orient = '|';
        if (orientation == JSplitPane.VERTICAL_SPLIT)
            orient = '_';
        DefaultInfo di = new DefaultInfo(name, divider, pointColor, resolvedWidth, orient, fontsize);
        defaults.put(name, di);
        dirty = true;
    }

    public void setAnnotationsRootDirectory(String annotationsRootDirectory)
    {
        this.annotationsRootDirectory = annotationsRootDirectory;
    }

    public String getAnnotationsRootDirectory()
    {
        return annotationsRootDirectory;
    }

    public void setManuscriptRootDirectory(String manuscriptRootDirectory) throws Exception
    {
        mgp.setManuscriptRootDirectory(manuscriptRootDirectory);
        dirty = true;
    }

    public String getManuscriptRootDirectory()
    {
        return mgp.getManuscriptRootDirectory();
    }

    public void setAnnotationsBackupDirectory(String annotationsBackupDirectory) throws Exception
    {
        mgp.setAnnotationsBackupDirectory(annotationsBackupDirectory);
        dirty = true;
    }

    public String getAnnotationsBackupDirectory()
    {
        return mgp.getAnnotationsBackupDirectory();
    }

    public void setCatalogDirectory(String catalogDirectory)
    {
        mgp.setCatalogDirectory(catalogDirectory);
        dirty = true;
    }

    public String getCatalogDirectory()
    {
        return mgp.getCatalogDirectory();
    }


    /*
    public void setBackupDirectory(String backupDirectory) throws Exception
    {
        mgp.setBackupDirectory(backupDirectory);
        dirty = true;
    }

    public String getBackupDirectory()
    {
        return mgp.getBackupDirectory();
    }
     */

    public boolean getUseCDATA() {
        return useCDATA;
    }

    public boolean getCanCloseWindow() {
        return canCloseWindow;
    }

    public boolean getSearchWithSLP1() {
        return searchWithSLP1;
    }

    public boolean getShowReviewOnly () {
        return showReviewOnly;
    }

    public void setTextDirectory(String textDirectory) throws Exception {
        mgp.setTextDirectory(textDirectory);
        dirty = true;
    }

    public String getTextDirectory()
    {
        return mgp.getTextDirectory();
    }


    public ArrayList<CommentInfo> getCommentInfo()
    {
        return commentInfo;
    }

    public void toXML(boolean force) throws Exception
    {
        dirty = true;
        toXML();
    }


    public void toXML() throws Exception
    {
        if (!dirty)
            return;

        mgp.toXML();    // write out defaults.xml
        dirty = false;

        /*  This is commented out because anything that changes programmatically is now in machineGenerated.xml

        // Since program changes nothing else in config.xml config.xml can remain as is.

        FileWriter fw = IOAssistance.openUncachedWriter("config.xml");
        fw.write(String.format("<config useCDATA='%s' canCloseWindow='%s' searchWithSLP1='%s' defaultEncoding='%s' useDefaultWidth='%s'>\n",
                            useCDATA ? "y" : "n",
                            canCloseWindow ? "y" : "n",
                            searchWithSLP1 ? "y" : "n",
                            defaultEncoding,
                            useDefaultWidth ? "y" : "n"
        ));

        // Write out virtual files
        fw.write("\t<virtualFiles>\n");
        StringBuilder sb = new StringBuilder();
        for (String key : virtualFileNames)
        {
            fw.write(String.format("\t\t<virtualFile name='%s'>\n", key));
            ArrayList<SegmentInfo> sis = virtualFiles.get(key);
            for (SegmentInfo si : sis) {
                sb.setLength(0);
                sb.append("\t\t\t<segment type='");
                sb.append(si.getType());
                sb.append("'");

                if (si.source != null) {
                    sb.append(String.format(" source='%s'", si.source));
                }

                if (si.matcher != null) {
                    sb.append(String.format(" matcher='%s'", si.matcher));
                }

                if (si.start != null) {
                    sb.append(String.format(" start='%s'", si.start));
                }

                if (si.end != null) {
                    sb.append(String.format(" end='%s'", si.end));
                }

                if (si.content.length() == 0) {
                    sb.append("/>\n");
                } else {
                    sb.append(String.format(">\n%s\n\t\t\t</segment>\n", si.content));
                }
                fw.write(sb.toString());
            }
            fw.write(String.format("\t\t</virtualFile>\n", key));
        }
        fw.write("\t</virtualFiles>\n");

        // Write out comment tags
        fw.write("\t<commentTags>\n");
        for (CommentInfo ci : commentInfo)
        {
            fw.write(String.format("\t\t<commentTag tag='%s'>%s</commentTag>\n", ci.tag, ci.description));
        }
        fw.write("\t</commentTags>\n");

        // Write out defaults
        fw.write("</config>\n");
        fw.close();
*/
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
        elements = new ArrayList<String>();

    }

    @Override
    public void endDocument() throws SAXException
    {
    }


    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        elements.add(qName);
        if (qName.equals("config"))
        {
            String value = atts.getValue("useCDATA");
            this.useCDATA = value == null || value.toLowerCase().equals("y");       // default is to use CDATA

            value = atts.getValue("canCloseWindow");
            this.canCloseWindow = value == null || value.toLowerCase().equals("y"); // default is to close window

            value = atts.getValue("searchWithSLP1");
            this.searchWithSLP1 = value == null || value.toLowerCase().equals("y"); // default is to search with SLP1

            value = atts.getValue("defaultEncoding");
            if (value != null)
                this.defaultEncoding = value;
            else
                this.defaultEncoding = "deva";

            value = atts.getValue("useDefaultWidth");
            if (value == null)
                this.useDefaultWidth = false;                   // default is to not use default width
            else
                this.useDefaultWidth = value.toLowerCase().equals("y");

            value = atts.getValue("showReviewOnly");
            if (value == null) {
                this.showReviewOnly = false;
            } else {
                this.showReviewOnly = value.toLowerCase().equals("y");
            }

        } else if (qName.equals("virtualFiles")) {
            virtualFiles = new Hashtable<String, ArrayList<SegmentInfo>>();
        }
        else if (qName.equals("virtualFile")) {
            virtualFileName = atts.getValue("name");
            if (virtualFileName == null) {
                msg("virtualFile element missing the name attribute");
            }
            virtualFileNames.add(virtualFileName);
            virtualFile = new ArrayList<SegmentInfo>();
        } else if (qName.equals("segment")) {
            segmentContent = "";
            segmentType = atts.getValue("type");
            segmentSource = atts.getValue("source");
            segmentMatcher = atts.getValue("matcher");
            segmentStart = atts.getValue("start");
            segmentEnd = atts.getValue("end");
        } else if (qName.equals("mapMatchedToSLP")) {
            String encoding = atts.getValue("encoding");
            if (encoding == null) {
                msg("mapMatchedToSLP element missing the encoding attribute");
            } else {
                if (segmentMapSpecs == null) {
                    segmentMapSpecs = new Hashtable<String, Boolean>();
                }
                segmentMapSpecs.put(encoding, Boolean.TRUE);
            }
        }
        else if (qName.equals("commentTags"))
            commentInfo = new ArrayList<CommentInfo>();
        else if (qName.equals("commentTag"))
        {
            currentComment = new CommentInfo(atts.getValue("tag"));
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        elements.remove(elements.size()-1);
        if (qName.equals("encoding")) {
            supportedEncodings.add(supportedEncoding);
            supportedEncoding = "";
        } else if (qName.equals("commentTag")) {
            commentInfo.add(currentComment);
            currentComment = null;
        } else if (qName.equals("virtualFile")) {
            virtualFiles.put(virtualFileName, virtualFile);
            virtualFileName = null;
            virtualFile = new ArrayList<SegmentInfo>();         // can't clear, need a new one.
        } else if (qName.equals("segment")) {
            virtualFile.add(new SegmentInfo(segmentType, segmentContent, segmentSource, segmentMatcher, segmentStart, segmentEnd, segmentMapSpecs));
            segmentContent = "";
            segmentType = null;
            segmentSource = null;
            segmentMatcher = null;
            segmentStart = null;
            segmentEnd = null;
            segmentMapSpecs = null;
        }
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        String currentElement = elements.get(elements.size()-1);
        String content = new String(ch, start, length).trim();

        // Note concatenation is used below instead of assignment because the content may come in in chunks.
        if (currentElement.equals("encoding")) {
            supportedEncoding += content;
        }
        else if (currentElement.equals("segment")) {
            segmentContent += content;
        }
        else if (currentElement.equals("commentTag")) {
            if (currentComment.description == null)
                currentComment.description = content;
            else
                currentComment.description += content;
        }
    }

    public void msg(String text) throws SAXException
    {
        throw new SAXException(String.format("%d.%d %s", locator.getLineNumber(), locator.getColumnNumber(), text));
    }
}
