package sita;

import java.util.ArrayList;
import java.io.File;
import java.util.Hashtable;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.sanskritlibrary.CatalogRecord;

//import org.xml.sax.EntityResolver;
//import org.xml.sax.DTDHandler;
//import org.xml.sax.ContentHandler;
//import org.xml.sax.ErrorHandler;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.TabInfo;
import org.sanskritlibrary.Text;
import org.sanskritlibrary.Text.Type;

public class Catalog extends DefaultHandler //implements ErrorHandler, ContentHandler, DTDHandler, EntityResolver
{
    private Locator locator;
    private CatalogRecord currentCatalogRecord;
    private ArrayList<String> elements;
    private Pages pages;
    private String defaultEncoding;     // from config.xml
    private Hashtable<String, StringBuilder> contentHash = new Hashtable<String, StringBuilder>();
    private ArrayList<String> manuscriptTypes = new ArrayList<String>();
    private StringBuilder textTypeBuffer = new StringBuilder();

    private ArrayList<TabInfo> currentTabs = new ArrayList<TabInfo>();
    private TabInfo currentTab;
    private int currentTabNumber;
    private boolean isVirtual;
    public ArrayList<CatalogRecord> catalogs;

    
    public CatalogRecord getCatalogRecordForDirectory(String dirname)
    {
        for (CatalogRecord cr : catalogs)
        {
            if (cr.directoryName.equals(dirname))
                return cr;
        }
        
        return null;
    }

    public ArrayList<String> getManuscriptTypes () {
        return manuscriptTypes;
    }

    public ArrayList<CatalogRecord> getDirectories(String manuscriptType)
    {
        ArrayList<CatalogRecord> a = new ArrayList<CatalogRecord>();

        for (int i=0; i<catalogs.size(); i++)
        {
            CatalogRecord cr = catalogs.get(i);
            if (cr.manuscriptType.equals(manuscriptType)) {
                a.add(cr);
            }
        }

        return a;
    }


    public Catalog(String fileName, Pages pages, String defaultEncoding) throws Exception
    {
        this.pages = pages;
        this.defaultEncoding = defaultEncoding;

        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();

        // Parse the input
        SAXParser saxParser = factory.newSAXParser();
        try {
            saxParser.parse( new File(fileName), this);
        } catch (Exception ex) {
            Logger.writeLog(String.format("Line %d: %s", locator.getLineNumber(), ex.getMessage()));
            throw ex;
        }
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
        elements = new ArrayList<String>();
        catalogs = new ArrayList<CatalogRecord>();
        contentHash.put("directoryName", new StringBuilder());
        contentHash.put("text", new StringBuilder());
        contentHash.put("manuscriptType", new StringBuilder());
        contentHash.put("catalogName", new StringBuilder());
        contentHash.put("numLeaves", new StringBuilder());
        contentHash.put("numFiles", new StringBuilder());
        contentHash.put("title", new StringBuilder());
        contentHash.put("subtitle", new StringBuilder());
        contentHash.put("text", new StringBuilder());
        contentHash.put("textRef", new StringBuilder());
    }

    @Override
    public void endDocument() throws SAXException
    {
    }


    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        elements.add(qName);
        if (qName.equals("manuscript"))
        {
            String manuscriptType = atts.getValue("type");
            if (manuscriptType == null) {
                manuscriptType = "Omitted";
            }
            contentHash.put("manuscriptType", new StringBuilder(manuscriptType));
            if (!manuscriptTypes.contains(manuscriptType)) {
                manuscriptTypes.add(manuscriptType);
            }

            String display = atts.getValue("display");
            if (display == null || display.equals("yes"))
            {
                currentCatalogRecord = new CatalogRecord();
                catalogs.add(currentCatalogRecord);
                currentCatalogRecord.seqno = Integer.parseInt(atts.getValue("id"));
            }
            else
                currentCatalogRecord = null;
        } else if (qName.equals("tabs")) {
            currentTabs.clear();

        } else if (qName.equals("tab")) {
            String name = atts.getValue("name");
            if (name == null) {
                name = String.format("Tab %d", ++currentTabNumber);
            }
            currentTab = createNewTab(name, atts);
        } else if (qName.equals("textRef")) {
            if (atts.getValue("virtual") != null) {
                isVirtual = atts.getValue("virtual").toLowerCase().equals("yes");
            }
            if (currentTab == null) {                   // have textRef not inside a tab element
                currentTab = createNewTab("Tab1", atts);    // put it in a tab that will not be shown as a tab
            }
        }
    }

    private TabInfo createNewTab(String name, Attributes atts) {
        TabInfo ct = new TabInfo(name);
        ct.defaultEncoding = atts.getValue("defaultEncoding");
        if (ct.defaultEncoding == null) {
            ct.defaultEncoding = defaultEncoding;   // from config.xml
        }
        return ct;
    }


    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        elements.remove(elements.size()-1);
        if (currentCatalogRecord != null)
        {
            if (qName.equals("manuscript"))
            {
//                Logger.writeLog(String.format("--%s", contentHash.get("directoryName").toString().trim()));

                currentCatalogRecord.manuscriptType = contentHash.get("manuscriptType").toString().trim();
                currentCatalogRecord.directoryName = contentHash.get("directoryName").toString().trim();
                currentCatalogRecord.catalogName = contentHash.get("catalogName").toString().trim();
                currentCatalogRecord.numLeaves = Integer.parseInt(contentHash.get("numLeaves").toString().trim());
                currentCatalogRecord.numManuscriptFiles = Integer.parseInt(contentHash.get("numFiles").toString().trim());
                currentCatalogRecord.title = contentHash.get("title").toString().trim();
                currentCatalogRecord.subtitle = contentHash.get("subtitle").toString().trim();
                if (currentTabs.size() == 0 && currentTab != null) {
                    currentTabs.add(currentTab);
                }
                currentCatalogRecord.setTabs(currentTabs);
                currentCatalogRecord.setPages(pages.getPages(currentCatalogRecord.directoryName));
                currentCatalogRecord.setTextTypes();
                currentCatalogRecord = null;

                contentHash.get("directoryName").setLength(0);
                contentHash.get("catalogName").setLength(0);
                contentHash.get("numLeaves").setLength(0);
                contentHash.get("numFiles").setLength(0);
                contentHash.get("title").setLength(0);
                contentHash.get("subtitle").setLength(0);

                currentTabNumber = 0;
                currentTabs = new ArrayList<TabInfo>();         // Can't reuse array list, need a new reference
                currentTab = null;
            }
            else if (qName.equals("text")) {
                if (currentTab == null) {
                    System.out.println("text element must appear within tab element");
                }
                else {
                    currentTab.texts.add(new Text(Type.TextRef, contentHash.get("text").toString(), false));
                    contentHash.get("text").setLength(0);
                }
            }
            else if (qName.equals("textRef"))
            {
                currentTab.texts.add(new Text(Type.TextRef, contentHash.get("textRef").toString(), isVirtual));
                contentHash.get("textRef").setLength(0);
                isVirtual = false;
            } else if (qName.equals("tab")) {
                currentTabs.add(currentTab);
                currentTab = null;
            }
        }
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        String element = elements.get(elements.size()-1);
        String content = new String(ch, start, length);
        if (currentCatalogRecord != null)
        {
            StringBuilder sb = contentHash.get(element);
            if (sb != null)
            {
                    contentHash.get(element).append(content);
            }
        }
    }

    private void msg(String text) throws SAXException
    {
        throw new SAXException(String.format("%d.%d %s", locator.getLineNumber(), locator.getColumnNumber(), text));
    }

}
