package sita;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import javax.swing.AbstractAction;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;

public class CommentArea extends RestorableTextArea implements MouseListener, KeyListener, ActionListener
{
    JPopupMenu popupMenu;
    Component _this;

    public CommentArea(ApplicationState appstate, int numRows, int numCols, boolean preserveSelections, Dimension size)
    {
        super(appstate, numRows, numCols, preserveSelections);

        setLineWrap(true);
        setWrapStyleWord(true);
        setEditable(true);
        setSize(size);
        setFocusable(true);
        InputMap im = getInputMap();
        ActionMap am = getActionMap();
        im.put(KeyStroke.getKeyStroke('/'), "popup");
        am.put("popup", new PopupAction());
        appstate.getAnnotater().setupInputActions(im, am);

        addMouseListener(this);
        addKeyListener(this);

        popupMenu = new JPopupMenu();
        for (CommentInfo ci : appstate.getCommentTags())
        {
            JMenuItem item = new JMenuItem(ci.tag);
            item.setActionCommand(ci.tag);
            item.addActionListener(this);
            popupMenu.add(item);
        }

        _this = this;
    }

    public void insertCannedComment()
    {
        popupMenu.show(this, 300, 300);
    }

    public void keyTyped(KeyEvent e)
    {
        if (!e.isActionKey() && !e.isAltDown() && !e.isControlDown() && !e.isMetaDown() && !e.isAltGraphDown())         // to humor the Macintosh
        {
            if (!appstate.getAllowModification())       // if annotation is not modifiable
            {
                if (!appstate.getAnnotater().allowModification())
                    e.consume();
                else
                {
                    grabFocus();       // to humor Ubuntu
                }
                return;
            }
        }
    }

    public void keyPressed(KeyEvent e)
    {
    }

    public void keyReleased(KeyEvent e)
    {
    }


    public void actionPerformed(ActionEvent e)
    {
        CommentInfo ci = appstate.getCommentTag(e.getActionCommand());
        String s = getText();
        if (s.equals("/"))
           setText(ci.description);
        else
        {
            int idx = getCaretPosition();
            StringBuilder sb = new StringBuilder();
            for (int i=0; i<idx; i++)
            {
                sb.append(s.charAt(i));
            }

            sb.append(ci.description);

            for (int i=idx; i< s.length(); i++)
            {
                sb.append(s.charAt(i));
            }

            setText(sb.toString());
        }
    }

    public void mouseClicked(MouseEvent e)
    {
        if (!appstate.getAllowModification())               // if modifications are not allowed
            appstate.getAnnotater().allowModification();
    }

    public void mousePressed(MouseEvent e)
    {
    }

    public void mouseReleased(MouseEvent e)
    {
    }

    public void mouseEntered(MouseEvent e)
    {
    }

    public void mouseExited(MouseEvent e)
    {
    }

    private class PopupAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            if (getText().length() == 0)
                popupMenu.show(_this, 300, 300);
        }
    }
}
