package sita;

import java.awt.BorderLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import javax.swing.JDialog;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.JFileChooser;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.SwingConstants;
import java.awt.Dimension;
import javax.swing.DefaultListModel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;

public class DifferencesDialog extends JDialog implements ListSelectionListener, ActionListener
{
    String annotationsBackupDirectory = null;

    GridBagConstraints c = new GridBagConstraints();
    JLabel backupManuscriptDirectoryLabel;
    ApplicationState appstate;
    JButton okayButton;
    JButton cancelButton;
    JButton diffButton;
    JList resultsList;
    DefaultListModel resultsListModel;
    JScrollPane resultsListScrollPane;
    JLabel messageLabel;
    boolean ignoreValueChanged = false;
    int lastResultSelected = -1;

    String error;
    boolean currentManuscriptOnly;

    public DifferencesDialog(ApplicationState appstate)
    {
        this.appstate = appstate;
        Utils.setEscapeKeyListener(this);
        setLayout(new BorderLayout());
        add(createButtonPanel(), BorderLayout.NORTH);
        add(createListPanel(), BorderLayout.CENTER);
        add(createButtonPanel02(), BorderLayout.SOUTH);
        setModal(false);
        setAlwaysOnTop(true);
        setBounds(0, 0, 800, 250);
        setTitle("Configuration");
        Utils.upperRightScreen(this);
    }


    public void display(boolean currentManuscriptOnly)
    {
        ignoreValueChanged = true;
        lastResultSelected = -1;
        this.currentManuscriptOnly = currentManuscriptOnly;
        resultsListModel.clear();
        resultsList.clearSelection();
        messageLabel.setText("");
        setVisible(true);
        ignoreValueChanged = false;
    }


    private JPanel createListPanel()
    {
        JPanel panel = new JPanel();
        panel.setLayout(new BorderLayout());

        resultsListModel = new DefaultListModel();
        resultsList = new JList(resultsListModel);
        resultsList.addListSelectionListener(this);
        resultsList.setVisibleRowCount(10);
        resultsListScrollPane = new JScrollPane(resultsList);
        resultsListScrollPane.setPreferredSize(new Dimension(140, 200));
        resultsListScrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
        panel.add(resultsListScrollPane, BorderLayout.CENTER);

        messageLabel = new JLabel();
        messageLabel.setPreferredSize(new Dimension(300, 40));
        panel.add(messageLabel, BorderLayout.SOUTH);

        return panel;
    }


    private JPanel createButtonPanel02()
    {
        JPanel panel = new JPanel();
        panel.setLayout(new GridBagLayout());
        GridBagConstraints cs = new GridBagConstraints();

        diffButton = new JButton("Diff");
        diffButton.addActionListener(this);
        diffButton.setActionCommand("diff");
        cs.gridx = 0;
        cs.gridy = 0;
        panel.add(diffButton, cs);

        okayButton = new JButton("OK");
        okayButton.addActionListener(this);
        okayButton.setActionCommand("okay");
        cs.gridx = 1;
        panel.add(okayButton, cs);

        cancelButton = new JButton("Cancel");
        cancelButton.addActionListener(this);
        cancelButton.setActionCommand("cancel");
        cs.gridx = 2;

        getRootPane().setDefaultButton(diffButton);

        panel.add(cancelButton, cs);
        return panel;

    }


    private JPanel createButtonPanel()
    {
        JPanel panel = new JPanel();
        panel.setLayout(new GridBagLayout());

        backupManuscriptDirectoryLabel = addButton(panel, 1, "Backup manuscript directory", appstate.getBackupManuscriptRootDirectory(), "backup");

        return panel;
    }

    private JLabel addButton(JPanel panel, int row, String type, String value, String actionCommand)
    {
        JLabel lblType = new JLabel(type);
        lblType.setHorizontalAlignment(SwingConstants.LEFT);

        JLabel lblValue = new JLabel(value);
        lblValue.setHorizontalAlignment(SwingConstants.LEFT);
        lblValue.setMinimumSize(new Dimension(400, 20));

        JButton b = new JButton("Change");
        b.setMinimumSize(new Dimension(95, 20));
        b.addActionListener(this);
        b.setActionCommand(actionCommand);

        c.gridy = row;
        c.gridx = 0;
        c.fill = GridBagConstraints.NONE;
        c.ipadx = 20;
        c.ipady = 20;
        c.anchor = GridBagConstraints.FIRST_LINE_START;
        c.weightx = 0.0;
        panel.add(lblType, c);              // name of directory

        c.gridx = 1;
        panel.add(lblValue, c);             // value of directory

        c.gridx = 2;
        c.ipadx = 0;
        c.ipady = 0;
        panel.add(b, c);                    // button

        return lblValue;
    }

    public void valueChanged(ListSelectionEvent e)
    {
        if (ignoreValueChanged)
            return;
        int idx = resultsList.getSelectedIndex();

        if (idx == lastResultSelected)
            return;
        lastResultSelected = idx;
        Utils.displayWaitCursor(this);
        ignoreValueChanged = true;
        appstate.getAnnotater().saveState();
        try
        {
            StringBuilder sb = new StringBuilder();
            VerseAnnotationCollection vac = appstate.differencesSummary.get(idx);
            int numAnnotations = vac.annotationNumbers.size();
            boolean annotationDeleted = vac.annotationNumbers.get(0) < 0;

            String rootDirectory = appstate.getManuscriptRootDirectory();
            String lastManuscript = appstate.getLastManuscript();
            String lastPage = (String) resultsList.getSelectedValue();
            if (annotationDeleted)
                lastPage = lastPage.substring(0, lastPage.length()-1);
            String pageName = appstate.getImageFilename(lastManuscript, lastPage + ".jpg");
            String imagefile = String.format("%s/%s/%s", rootDirectory, lastManuscript, pageName);
            appstate.getAnnotater().setCurrentPage(appstate.getLastManuscript(), lastPage);

            if (annotationDeleted)
                numAnnotations--;
            if (numAnnotations == 1)
                sb.append("see annotation ");
            else if (numAnnotations > 1)
                sb.append("see annotations ");
            String delim = "";
            for (int annotationNumber : vac.annotationNumbers)
            {
                if (annotationNumber >= 0)
                {
                    sb.append(String.format("%s%d", delim, annotationNumber + 1));
                    delim = ", ";
                }
            }
            if (annotationDeleted)
                sb.append(String.format(" (%d annotation(s) deleted)", -vac.annotationNumbers.get(0)));

            messageLabel.setText(sb.toString());

            if (numAnnotations > 0)
                appstate.setMoveToAnnotationNumber(vac.annotationNumbers.get(annotationDeleted?1:0));
            appstate.getAnnotater().displayManuscript(imagefile);
            appstate.getAnnotater().bringToFront();
        } catch (Exception ex)
        {
            Utils.displayError(ex);
        }
        Utils.displayDefaultCursor(this);
        ignoreValueChanged = false;
    }



    public void actionPerformed(ActionEvent e)
    {
        if (e.getActionCommand().equals("okay"))
        {
            try
            {
                setVisible(false);
            }
            catch (Exception ex)
            {
                Utils.displayError(ex);
            }
        }
        else if (e.getActionCommand().equals("cancel"))
            setVisible(false);
        else if (e.getActionCommand().equals("diff"))
        {
            try
            {
                ignoreValueChanged = true;
                resultsListModel.clear();
                appstate.generateDifferencesSummary(currentManuscriptOnly);
                if (appstate.differencesSummary != null)
                {
                    int n = 0;
                    for (VerseAnnotationCollection vac : appstate.differencesSummary)
                    {
                        String missing = "";
                        if (vac.annotationNumbers.get(0) < 0)
                            missing = "-";
                        resultsListModel.add(n, vac.filename + missing);
                        n++;
                    }
                    if (n == 0)
                        JOptionPane.showMessageDialog(this, "No differences found");
                    ignoreValueChanged = false;
                }
                else
                {
                    String msg = String.format("%s does not exist. You can set it above.", appstate.getAnnotationsBackupDirectory());
                    JOptionPane.showMessageDialog(this, msg);
                }
            }
            catch (Exception ex)
            {
                Utils.displayError(ex);
            }
        }
        else
        {
            String backupDirectory = null;
            backupDirectory = appstate.getBackupManuscriptRootDirectory();

            JFileChooser chooser = new JFileChooser();
            chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
            chooser.setCurrentDirectory(new File(backupDirectory));
            int returnVal = chooser.showOpenDialog(this);
            if (returnVal == JFileChooser.APPROVE_OPTION)
            {
                String filename = chooser.getSelectedFile().getAbsolutePath();

                try
                {
                    annotationsBackupDirectory = filename;
                    if (annotationsBackupDirectory != null)
                        appstate.setBackupManuscriptRootDirectory(annotationsBackupDirectory);
                    appstate.update();
                    backupManuscriptDirectoryLabel.setText(filename);
                } catch (Exception ex)
                {
                    Utils.displayError(ex);
                }
            }
        }
    }

    public String getError()
    {
        return error;
    }

}
