package sita;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.MediaTracker;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.io.File;

public class ImagePanel extends DrawingPanel
{
    String error;
    Image im;
    MediaTracker imageTracker = new MediaTracker(this);
    int displayWidth;
    int displayHeight;
    //int resolvedImageWidth = -1;         // this is how many pixels the actual image width is resized as.
    private int resolvedImageHeight = -1;        // this is how many pixels the actual image height is resized as.
    ImageScrollPane scroller;
    int x = 0;
    int y = 0;

    public ImagePanel(ApplicationState appstate, int displayWidth, int displayHeight)
    {
        super(appstate);
        
        this.displayWidth = displayWidth;    
        this.displayHeight = displayHeight;

        this.setSize(new Dimension(displayWidth, displayHeight));
        this.setPreferredSize(new Dimension(displayWidth, displayHeight));
    }


    public int getActualWidth()
    {
        return im.getWidth(this);
    }

    public int getActualHeight()
    {
        return im.getHeight(this);
    }

    public String getError()
    {
        return error;
    }

    public void setScroller(ImageScrollPane scroller)
    {
        this.scroller = scroller;
    }

    public void adjustHorizontal(int n)
    {
        x = -n;
        repaint();
    }

    public void adjustVertical(int n)
    {
        y = -n;
        repaint();
    }

    public int getDisplayWidth()
    {
        return displayWidth;
    }

    public void setDisplayWidth(int displayWidth)
    {
        this.displayWidth = displayWidth;
        scroller.setExtentHorizontal(displayWidth);
    }

    public int getDisplayHeight()
    {
        return displayHeight;
    }

    public void setDisplayHeight(int displayHeight)
    {
        this.displayHeight = displayHeight;
        scroller.setExtentVertical(displayHeight);
    }

    public int getResolvedImageWidth()
    {
        return appstate.getResolvedWidth();
    }

    public int getResolvedImageWidth(boolean b)
    {
        return appstate.getResolvedWidth(b);
    }

    public void clearTemporaryMarks()
    {
        rbRect.clearTemporaryMarks();
    }

    
    public void setResolvedImageWidth(int width)
    {
        int old = appstate.getResolvedWidth();           // remember current width
        appstate.setResolvedWidth(width);
        if (old != -1 && old != width)
        {
            appstate.adjustAnnotations(getActualWidth(), old, width);        // so that all annotations appear correctly.
            rbRect.adjustTemporaryMarks(getActualWidth(), old, width);
            if (appstate.getCurrentAnnotation().getLeftBracketExists())     // adjust click area of left bracket if it exists
                rbRect.setLeftBounds(appstate.getPoint1());
            if (appstate.getCurrentAnnotation().getRightBracketExists())    // adjust click area of right bracket if it exists
                rbRect.setRightBounds(appstate.getPoint2());
        }
    }

    public void loadImage(String filename) throws Exception
    {
        if (!new File(filename).exists())
        {
            throw new Exception(String.format("File cannot be opened. Most probable reason is that the external drive is not turned on. If it is, open the configuration menu and check the root manuscript directory.<br/><br/>File name=%s", filename));
        }

        if (im != null)
            im.flush();

        appstate.setImageFilename(filename);

        im = Toolkit.getDefaultToolkit().getImage(filename);  // an Applet has a getImage method
        imageTracker.addImage(im, 0);
        try
        {
            imageTracker.waitForID(0);
            imageTracker.removeImage(im);       // is this necessary? It doesn't appear to do any harm.            
        }
        catch (InterruptedException ex)
        {
            Logger.writeLog(ex);
            error = ex.getMessage();
        }
    }

    public void clearMarkerData()
    {
       rbRect.clearMarkerData();
    }

    public void clearAnnotation()
    {
        rbRect.remove();
        display();
    }

    public void setPointColor(Color c)
    {
        rbRect.setPointColor(c);
        repaint();
    }

    public Color getPointColor()
    {
        return rbRect.getPointColor();
    }

    public void swapBrackets()
    {
        if (appstate.swapBrackets())
            display();
    }


    public void display()
    {
        display(true);
    }

    public void display(boolean adjustScrolling)
    {
        boolean explicitRepaint = true;
        int actualWidth = getActualWidth();
        int actualHeight = getActualHeight();
        int resolvedImageWidth = appstate.getResolvedWidth();
        resolvedImageHeight = (int)(resolvedImageWidth * ((double)actualHeight/(double)actualWidth));
        
        if (resolvedImageWidth <= displayWidth)
        {
            scroller.setHorizontalValue(0);
            scroller.setMinimumHorizontal(0);
            scroller.setMaximumHorizontal(displayWidth);
            x = 0;
            explicitRepaint = true;
        }
        else
            scroller.setMaximumHorizontal(resolvedImageWidth);

        if (resolvedImageHeight <= displayHeight)
        {
            scroller.setVerticalValue(0);
            scroller.setMinimumVertical(0);
            scroller.setMaximumVertical(displayHeight);
            y = 0;
            explicitRepaint = true;
        }
        else
            scroller.setMaximumVertical(resolvedImageHeight);

        if (adjustScrolling && rbRect.exists())
        {
            // Make sure that the mark (left bracket or upper left corner of box) is visible
            Rectangle imageVisible = new Rectangle(
                scroller.getHorizontalValue(), scroller.getVerticalValue(),
                displayWidth, displayHeight);

            Rectangle r = rbRect.getBounds();

            if (r.width > 0 && r.height>0 && !imageVisible.contains(r))      // not all of rectangle is on screen
            {
                int n = (int)r.getX() - 30;
                if (n < 0)
                    n = 0;
                scroller.setHorizontalValue(n);

                n = (int)r.getY() - 30;
                if (n < 0)
                    n = 0;
                scroller.setVerticalValue(n);
            }
        }

        if (explicitRepaint)
            repaint();
    }

    @Override
    public void paint(Graphics g)
    {
        g.clearRect(0, 0, displayWidth, displayHeight);
        g.clipRect(0, 0, displayWidth, displayHeight);
        g.drawImage(im, x, y, appstate.getResolvedWidth(), resolvedImageHeight, this);
        //Utils.saveImage(im, appstate.getResolvedWidth(), resolvedImageHeight);
        rbRect.setXOffset(scroller.getHorizontalValue());
        rbRect.setYOffset(scroller.getVerticalValue());
        rbRect.redraw(g);
    }
}
