/*
 * From Graphic Java 1.2 Volume 1: AWT, Third Edition
 * David Geary
 * Sun Press
 */

package sita;

import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;
abstract public class RubberBand
{
   protected ApplicationState appstate;

   protected int xOffset;
   protected int yOffset;

   protected Point lastPt      = new Point(0,0);
   protected Point endPt       = new Point(0,0);

   protected Point potentialAnchorPt = new Point(0, 0);
   protected long mouseClickTime = 0;
   protected boolean ignoreMouseClicked = false;
   protected boolean ignoreMouseReleased = false;

   static protected Rectangle lbBounds = new Rectangle();
   static protected Rectangle rbBounds = new Rectangle();
   static protected ArrayList<Rectangle> tempMarks = new ArrayList<Rectangle>();

   private Component component;
   private boolean   firstStretch = true;

   abstract public void drawLast(Graphics g);
   abstract public void drawNext(Graphics g);
   abstract public void eraseBracket(Rectangle r1);
   abstract public void drawLeftBracket(Graphics g, Point p);
   abstract public void drawRightBracket(Graphics g, Point p);
   abstract public void setRightBounds(Point p);
   abstract public void setLeftBounds(Point p);
   abstract public void redraw(Graphics g);
   abstract public void drawTemporaryMarks(Graphics g);
   abstract public void clearTemporaryMarks();
   abstract public void adjustTemporaryMarks(int actualWidth, int oldWidth, int newWidth);


   public RubberBand(ApplicationState appstate)
   {
       this.appstate = appstate;
   }

   public RubberBand(ApplicationState appstate, Component c)
   {
      this.appstate = appstate;
      setComponent(c);
   }

   public void clearMarkerData()
   {
       lbBounds.x = lbBounds.y = lbBounds.width = lbBounds.height = 0;
       rbBounds.x = rbBounds.y = rbBounds.width = rbBounds.height = 0;
   }

   public void setXOffset(int xOffset)
   {
       this.xOffset = xOffset;
   }

   public void setYOffset(int yOffset)
   {
       this.yOffset = yOffset;
   }

   public void setPointColor(Color c)
   {
       appstate.setPointColor(c);
   }

   public Color getPointColor()
   {
       return appstate.getPointColor();
   }

   public int getPointSize()
   {
       return appstate.getPointSize();
   }


   public void setComponent(Component c) {
      component = c;

      component.addMouseListener(new MouseAdapter()
      {
         @Override
         public void mousePressed(MouseEvent event)
         {
             if (appstate.getReviewOnly()) {
                 Utils.displayReviewOnlyWarning();
                  if (appstate.isMacOSX())
                    ignoreMouseClicked = true;
                  ignoreMouseReleased = true;
                  return;                           // make user click mouse again
             }
              if (!appstate.getAllowModification())         // remind user that this is a previously existing annotation
              {
                  appstate.getAnnotater().allowModification();
                  
                  if (appstate.isMacOSX())
                    ignoreMouseClicked = true;
                  ignoreMouseReleased = true;
                  return;                           // make user click mouse again
              }

             mouseClickTime = event.getWhen();
             if (appstate.getBoxExists())
             {
                 removeBox();
                 ignoreMouseClicked = true;
             }
             anchor(addOffset(event.getPoint()));
         }

         @Override
          public void mouseClicked(MouseEvent event)        // user is working with brackets
          {
              if (ignoreMouseClicked)               // is true if a box was just removed. Force using to click again to make a mark
              {
                  ignoreMouseClicked = false;
                  return;
              }

              if (event.isShiftDown())              // deal with temporary mark
              {
                  prepareTemporaryMark(addOffset(event.getPoint()));
                  return;
              }

//              if (appstate.getBoxExists())          // get rid of the box
//                  appstate.setBoxExists(false);

              if (lbBounds.contains(event.getPoint()))      // did user click on an existing left bracket?
                  removeBrackets();                         // yes, then remove left bracket and the right bracket if it exists
              else
              {
                  Point mp = addOffset(event.getPoint());
                  if (!appstate.getLeftBracketExists())     // if there is no left bracket then create a left bracket
                  {
                      leftBracket(mp);
                      drawBracket(appstate.getPoint1(), true);
                      appstate.setLeftBracketExists(true);
                  } else                                          // a left bracket is already visible, make create or reposition a right bracket
                  {
                      if (appstate.getRightBracketExists())     // if a right bracket already exists then delete it and reposition it at mouse click
                      {
                          appstate.setRightBracketExists(false);
                          appstate.annotater.refreshManuscript(false);   // get rid of old right bracket and do not change scrolling
                      }
                      appstate.setRightBracketExists(true);
                      rightBracket(mp);
                      drawBracket(appstate.getPoint2(), false);       // draw new right bracket
                  }
              }

              appstate.getAnnotater().updateTitle();
          }

          @Override
          public void mouseReleased(MouseEvent event)
          {
              if (ignoreMouseReleased)                      // is true if user is attempting to modify an existing annotation
              {
                  ignoreMouseReleased = false;
                  return;
              }

              if (appstate.getBoxExists())
                  appstate.getAnnotater().updateTitle();
              appstate.annotater.refreshManuscript(false);           // so that it will be painted in the mark color instead of the XOR color.
          }
      });


      component.addMouseMotionListener(new MouseMotionAdapter()
      {
         @Override
         public void mouseDragged(MouseEvent event)             // user is working with a box
         {
             if (event.getWhen() - mouseClickTime < 128)        // ignore a short hiccup between mouse down and mouse up
                 return;
             else
             {
                if (appstate.getLeftBracketExists())            // if there are brackets visible, remove them
                {
                    removeBrackets();
                    anchor(potentialAnchorPt);
                }
                stretch(addOffset(event.getPoint()));           // set lower right corner of box
                appstate.setBoxExists(true);
             }
         }
      });
   }

   public Point   getAnchor   () { return appstate.getPoint1();    }
   public Point   getStretched() { return appstate.getPoint2(); }
   public Point   getLast     () { return lastPt;      }
   public Point   getEnd      () { return endPt;       }
   public Point   getStart    () { return appstate.getPoint1();     }
   public Point   getFinish   () { return appstate.getPoint2();    }


   /*
    * The logical position of the point is influenced by the scroll position.
    * If xOffset is greater than 0, then user has scrolled to the right, so a point physically displayed at <0, y> is really at <xOffset, y>
    * If yOffset is greater than 0, then user has scrolled to down, so a point physically displayed at <x, 0> is really at <x, yOffset>
   */
   public Point addOffset(Point pt)
   {
       return new Point(pt.x + xOffset, pt.y + yOffset);
   }


    public void removeBrackets()
    {
        eraseBracket(true);     // erase the left bracket
        appstate.setLeftBracketExists(false);
        if (appstate.getRightBracketExists())
        {
            eraseBracket(false);        // erase the right bracket
            appstate.setRightBracketExists(false);
        }
    }

   // Keep track of upper left corner of box
   public void anchor(Point p)
   {
       if (appstate.getLeftBracketExists())
       {
           potentialAnchorPt.x = p.x;           // anchor will be recalled with this on mousedrag event
           potentialAnchorPt.y = p.y;
           return;
       }

       firstStretch = true;

       appstate.setPoint1(p);
       appstate.setPoint2(p);
       lastPt.x = p.x;
       lastPt.y = p.y;
   }

   // Keep track of lower right corner of box.
   public void stretch(Point p)
   {
       if (appstate.getLeftBracketExists())
           return;

       Point p2 = appstate.getPoint2();
       lastPt.x      = p2.x;
       lastPt.y      = p2.y;

       appstate.setPoint2(p);

       Graphics g = component.getGraphics();
       if(g != null) {
         try
         {
           g.setXORMode(component.getBackground());
           if(firstStretch == true)
               firstStretch = false;
           else
               drawLast(g);

           drawNext(g);
         }
         finally
         {
            g.dispose();
         }
       }
   }

   // modify currentAnnotation in application state
   public void leftBracket(Point p)
   {
       appstate.setPoint1(p);
   }

   // modify currentAnnotation in applicatioin state
   public void rightBracket(Point p)
   {
       appstate.setPoint2(p);
   }


   public boolean getLeftBracketExists()
   {
       return appstate.getLeftBracketExists();
   }

   public boolean getRightBracketExists()
   {
       return appstate.getRightBracketExists();
   }

   public boolean getBoxExists()
   {
       return appstate.getBoxExists();
   }

   public void eraseBracket(boolean leftBracket)
   {
        if (leftBracket)
            eraseBracket(lbBounds);
        else
            eraseBracket(rbBounds);
   }

   public void drawBracket(Point p, boolean leftBracket)
   {
       Graphics g = component.getGraphics();
       if(g != null)
       {
         try
         {
           g.setColor(appstate.getPointColor());

           if (leftBracket)
                drawLeftBracket(g, p);
           else
                drawRightBracket(g, p);
         }
         finally
         {
            g.dispose();
         }
       }
   }

   public void removeBox()
   {
       Graphics g = component.getGraphics();
       if(g != null)
       {
         try
         {
           appstate.setBoxExists(false);
           appstate.getAnnotater().refreshManuscript();
         }
         finally
         {
            g.dispose();
         }
       }
   }


   public void remove()         // logically clear all marks from image
   {
       appstate.setLeftBracketExists(false);
       appstate.setRightBracketExists(false);
       appstate.setBoxExists(false);

       appstate.setPoint1(-1, -1);
       appstate.setPoint2(-1, -1);
       appstate.setComment("");
   }

   public boolean exists()
   {
       return (appstate.getLeftBracketExists() || appstate.getBoxExists());
   }



   public Rectangle getBounds()
   {
       Point p1 = appstate.getPoint1();     // note that point1 and point2 are used either for location of bracket(s) or box
       Point p2 = appstate.getPoint2();
       int x = p1.x;
       if (x > p2.x)
           x = p2.x;

       int y = p1.y;
       if (!appstate.getLeftBracketExists() || appstate.getRightBracketExists())
       {
            if (y > p2.y)
                y = p2.y;
       }

       int width = 0;
       if (!appstate.getLeftBracketExists() || appstate.getRightBracketExists())
           width = Math.abs(p2.x - p1.x);

       int height = 0;
       if (!appstate.getLeftBracketExists() || appstate.getRightBracketExists())
           height = Math.abs(p2.y - p1.y);

       return new Rectangle(x, y, width, height);
   }

    public Rectangle lastBounds()
    {
        Point anchorPt = appstate.getPoint1();
        return new Rectangle(
                lastPt.x < anchorPt.x ? lastPt.x : anchorPt.x,
                lastPt.y < anchorPt.y ? lastPt.y : anchorPt.y,
                Math.abs(lastPt.x - anchorPt.x),
                Math.abs(lastPt.y - anchorPt.y));
    }

    private void prepareTemporaryMark(Point p)      // note p is adjusted for whole image
    {
        // See if user wants to get rid of mark (clicked an existing mark)
        boolean removal = false;
        for (int i=0; i<tempMarks.size(); i++)
        {
            Rectangle r = tempMarks.get(i);
            if (r.contains(p))
            {
                tempMarks.remove(i);
                removal = true;
                break;
            }
        }
        if (!removal)
            tempMarks.add(new Rectangle(p.x - 10, p.y - 10, 20, 20));

        appstate.annotater.refreshManuscript();   // redraw marks
    }
}


