package sita;

import java.awt.*;

public class RubberBandRectangle extends RubberBand
{

    private Component component;
    private Rectangle lastEndBracketMarker = null;
    private static int barWidth = 2;

    public RubberBandRectangle(ApplicationState config)
    {
        super(config);
    }

    public RubberBandRectangle(ApplicationState config, Component component)
    {
        super(config, component);
        this.component = component;
    }

    // Redraw the marks on the image (which currently has no marks visible)
    public void redraw(Graphics g)
    {
        if (g != null)
        {
            if (tempMarks.size() != 0)
            {
                g.setColor(Color.yellow);
                drawTemporaryMarks(g);
            }

            if (appstate.getLeftBracketExists())      // brackets exist
            {
                g.setColor(appstate.getPointColor());
                drawLeftBracket(g, appstate.getPoint1());
                if (appstate.getRightBracketExists()) {
                    if (lastEndBracketMarker != null) {
                        tempMarks.remove(lastEndBracketMarker);
                    }
                    lastEndBracketMarker = new Rectangle(
                            appstate.getPoint2().x,
                            appstate.getPoint2().y,
                            10,
                            10);
                    tempMarks.add(lastEndBracketMarker);
                    drawRightBracket(g, appstate.getPoint2());
                }
            } else if (appstate.getBoxExists())   // box exists
            {
                g.setColor(appstate.getPointColor());
                drawNext(g);
            }
            // else no manuscript annotation
        }
    }

    // Erase previous rectangle. Note g will alway be in XORMode here.
    public void drawLast(Graphics g)
    {
        Rectangle rect = lastBounds();
        g.drawRect(rect.x - xOffset, rect.y - yOffset, rect.width, rect.height);
    }

    // Draw rectangle. Note, g will be in XORMode if rectangle is being created by dragging. Other setColor will have been called which
    // means that a call to drawLast will not erase the rectangle
    public void drawNext(Graphics graphics)
    {
        Rectangle rect = getBounds();
        if (rect.getHeight() > 0.0 && rect.getWidth() > 0.0)
            graphics.drawRect(rect.x - xOffset, rect.y - yOffset, rect.width, rect.height);
    }

    public void drawLeftBracket(Graphics g, Point p)
    {
        setLeftBounds(p);

        // draw vertical bar
        g.fillRect(lbBounds.x, lbBounds.y, barWidth, lbBounds.height);
        // draw top horizontal bar
        g.fillRect(lbBounds.x + barWidth, lbBounds.y, barWidth + 1, barWidth);
        // draw bottom horizontal bar
        //g.fillRect(lbBounds.x + barWidth, lbBounds.y + pointSize-barWidth, barWidth, barWidth);
    }

    public void setLeftBounds(Point p)
    {
        int pointSize = appstate.getPointSize();             // this is position of right edge of top of top horizontal bar

        lbBounds.x = p.x - (2 * barWidth + 1) - xOffset;
        lbBounds.y = p.y - yOffset;
        lbBounds.width = 2 * barWidth + 1;
        lbBounds.height = pointSize;
    }

    public void drawRightBracket(Graphics g, Point p)        // this is position of left edge of top of bottom horizontal bar
    {
        setRightBounds(p);

        // draw vertical bar
        g.fillRect(rbBounds.x + barWidth + 1, rbBounds.y, barWidth, rbBounds.height);
        // draw bottom horizontal bar
        int pointSize = appstate.getPointSize();
        g.fillRect(rbBounds.x, rbBounds.y + pointSize - barWidth, 2 * barWidth + 1, barWidth);
    }

    public void setRightBounds(Point p)
    {
        int pointSize = appstate.getPointSize();

        rbBounds.x = p.x - xOffset;
        rbBounds.y = p.y - pointSize - yOffset;
        rbBounds.width = 2 * barWidth + 1;
        rbBounds.height = pointSize;
    }

    public void eraseBracket(Rectangle r1)
    {
        r1.x = r1.y = r1.width = r1.height = 0;      // since there is no more bracket don't need these any more
        appstate.getAnnotater().refreshManuscript(false);
    }

    public void adjustTemporaryMarks(int actualWidth, int oldWidth, int newWidth)
    {
        for (int i=0; i<tempMarks.size(); i++)
        {
            Rectangle r = tempMarks.get(i);
            r.x = Utils.adjustValue(r.x, oldWidth, actualWidth);   // normalize the start point
            r.x = Utils.adjustValue(r.x, actualWidth, newWidth);   // denormalize it to the new width

            r.y = Utils.adjustValue(r.y, oldWidth, actualWidth);   // normalize the start point
            r.y = Utils.adjustValue(r.y, actualWidth, newWidth);   // denormalize it to the new width
        }
    }

    public void drawTemporaryMarks(Graphics g)
    {
        for (int i=0; i<tempMarks.size(); i++)
        {
            Rectangle r = tempMarks.get(i);
            g.fillOval(r.x - xOffset, r.y - yOffset, r.width, r.height);
        }
    }

    public void clearTemporaryMarks()
    {
        tempMarks.clear();
    }
}
