package sita;

import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import javax.swing.AbstractAction;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.KeyStroke;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;


/*
The Sanskritist is accustomed to using grep search and can learn the conventions of the encoded text
 (usually John Smith's MBh files).  If you can supply a search interface that uses regular expressions, it will be great.
 Off the top of my head I think of a few sandhi variations but none of them is used in John Smith's MBh, so
 why trouble yourself now?

visarga versus homorganic spirant
[HS]S
[Hz]z
[Hs]s

anusvAra versus homorganic nasal
[MN][kKgGN]
[MY][cCjJY]
[MR][wWqQR]
[Mn][tTdDn]
[Mm][pPbBm]

doubling of consonant after r or h following a vowel
e.g. vArtika versus vArttika
 */

public class SearchArea extends JPanel implements TextSearch, ChangeListener
{
    private static final String SEARCH_ACTION = "search";

    RestorableTextField query;
    public ArrayList<DigitalTextPage> dtps = new ArrayList<DigitalTextPage>();          // remove public
    //ArrayList<ArrayList<TextMap>> textmap = new ArrayList<ArrayList<TextMap>>();
    ApplicationState appstate;
    Component textPane;
    Annotated annotated;
    boolean ignoreStateChange = false;

    public SearchArea(ApplicationState appstate, Dimension minimumSize, Annotated annotated) throws Exception
    {
//        clippy = Toolkit.getDefaultToolkit().getSystemClipboard();

        this.appstate = appstate;
        this.annotated = annotated;
        query = new RestorableTextField();
        MarkedPanel queryMarkedPanel = new MarkedPanel(query);
        query.setMarkedPanel(queryMarkedPanel);

        InputMap im = query.getInputMap();
        ActionMap am = query.getActionMap();
        setupSearchActions(im, am);
        appstate.getAnnotater().setupInputActions(im, am);

        // If more than 1 tab then use a JTabbedPane, otherwise just use a JPanel
        if (appstate.getNumTabs() > 1) {
            textPane = new JTabbedPane();
        } else {
            textPane = new JPanel();
        }

        setLayout(new BorderLayout());
        add(queryMarkedPanel, BorderLayout.NORTH);
        add(textPane, BorderLayout.CENTER);
    }


    public void selectTab () {
        if (appstate.getNumTabs() > 1) {
            int tabNumber = appstate.getCurrentAnnotationTabNumber();
            if (tabNumber >= 0) {
                ignoreStateChange = true;
                ((JTabbedPane)textPane).setSelectedIndex(tabNumber);
                ignoreStateChange = false;
                appstate.setCurrentTabNumber(tabNumber);
            }
        }
    }


    public void loadTabs() throws Exception {
        dtps.clear();
        int numtabs = appstate.getNumTabs();
        textPane = getTextComponent(numtabs);       // return either a JTabbedPane or a JPanel
        if (numtabs > 1) {
            JTabbedPane tabbedPane = (JTabbedPane)textPane;
            tabbedPane.removeAll();
            int currentTabNumber = appstate.getCurrentTabNumber();
            for (int i=0; i<numtabs; i++) {
                appstate.setCurrentTabNumber(i);
                DigitalTextPage dtp = new DigitalTextPage(appstate, 5, 5, this);
                dtp.setMinimumSize(new Dimension(0, 0));
                dtps.add(dtp);
                /*
                TextArea ta = new TextArea("Some text");
                ta.setMinimumSize(new Dimension(0, 0));
                tabbedPane.addTab(String.format("Tab %d", i), ta);
                */
                tabbedPane.addTab(appstate.getTabName(i), dtp);
                tabbedPane.addChangeListener(this);
                appstate.setCurrentTabNumber(i);
                dtp.loadDigitalText(i, appstate.getEncoding());
            }
            appstate.setCurrentTabNumber(currentTabNumber);
            tabbedPane.setSelectedIndex(currentTabNumber);
            dtps.get(currentTabNumber).giveFocus();
        } else {
            JPanel panel = (JPanel)textPane;
            DigitalTextPage dtp = new DigitalTextPage(appstate, 5, 5, this);
            dtps.add(dtp);
            appstate.setCurrentTabNumber(0);
            dtp.loadDigitalText(0, appstate.getEncoding());
            panel.setLayout(new BorderLayout());
            panel.add(dtp, BorderLayout.CENTER);
            this.validate();
        }
    }

    private Component getTextComponent (int numtabs) {
        Component c = null;
        for (Component cmp : this.getComponents()) {
            if (cmp == textPane) {
                c = cmp;
                break;
            }
        }
        if (numtabs > 1) {
            if (!(c instanceof JTabbedPane)) {
                this.remove(c);
                c = new JTabbedPane();
                this.add(c, BorderLayout.CENTER);
                this.validate();
            }
        } else {
            if (!(c instanceof JPanel)) {
                this.remove(c);
                c = new JPanel();
                this.add(c, BorderLayout.CENTER);
                this.validate();
            }
        }
        return c;
    }

    public void setSearchAreaFocus()
    {
        query.grabFocus();
    }

    public void setTextFocus()
    {
        dtps.get(appstate.getCurrentTabNumber()).setTextFocus();
    }


    public Point getSelected()
    {
        return dtps.get(appstate.getCurrentTabNumber()).getSelected();
    }

    public void increaseFontSize()
    {
        dtps.get(appstate.getCurrentTabNumber()).increaseFontSize();
    }

    public void decreaseFontSize()
    {
        dtps.get(appstate.getCurrentTabNumber()).decreaseFontSize();
    }

    public void clearAnnotation()
    {
        dtps.get(appstate.getCurrentTabNumber()).clearAnnotation();
    }

    public TextPoint getEndOfPreviousAnnotation () {
        return dtps.get(appstate.getCurrentTabNumber()).getEndOfPreviousAnnotation();
    }

    public TextPoint getStartOfCurrentAnnotation () {
        return dtps.get(appstate.getCurrentTabNumber()).getStartOfCurrentAnnotation();
    }

    public void clearEndOfPreviousAnnotation() {
        dtps.get(appstate.getCurrentTabNumber()).clearEndOfPreviousAnnotation();
    }

    public int getTextCols()
    {
        return 0;       // TODO!
    }


    public void setScroller (int row) {
        dtps.get(appstate.getCurrentTabNumber()).setScroller(row);
    }

    public void display()
    {
        dtps.get(appstate.getCurrentTabNumber()).display();
        if (!appstate.isCurrentAnnotationEmpty() && !appstate.getAllowModification())
            query.grabFocus();                  // force user to either do a ctrl T or click so warning dialog will appear!
    }

    public void clearHighlight()
    {

    }


    public void saveSelection()
    {
        dtps.get(appstate.getCurrentTabNumber()).saveSelection();
    }

    public boolean selectionExists()
    {
        return dtps.get(appstate.getCurrentTabNumber()).selectionExists();
    }

    public void setupSearchActions(InputMap im, ActionMap am)
    {
        im.put(KeyStroke.getKeyStroke("ENTER"), SEARCH_ACTION);
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, InputEvent.SHIFT_DOWN_MASK), SEARCH_ACTION);
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, InputEvent.CTRL_DOWN_MASK), SEARCH_ACTION);
        am.put(SEARCH_ACTION, new SearchAction());
    }

    private void debug(String method, KeyEvent e)
    {
        String s = String.format("In %s %s, keyChar=%d, keyCode=%d",
                method,
                e.isControlDown() ? "Control key is down" : "Control key is not down",
                (int)e.getKeyChar(),
                e.getKeyCode());
        Logger.writeLog(s);
    }

    public void stateChanged(ChangeEvent e) {
        if (ignoreStateChange) {                // in case new tab selected programmatically
            return;
        }

        annotated.newAnnotation(true);
        JTabbedPane tp = (JTabbedPane)textPane;
        appstate.setCurrentTabNumber(tp.getSelectedIndex());

        if (tp.getSelectedIndex() >= 0) {
            appstate.setCurrentTabNumber(tp.getSelectedIndex());
            appstate.getCurrentAnnotation().setTabNumber(appstate.getCurrentTabNumber());
            appstate.getCurrentAnnotation().setEncoding(appstate.getEncoding());
        }
    }


    private class SearchAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            String text = query.getText();
            try
            {
                int searchType = 0;
                int modifiers = ev.getModifiers();
                if ((modifiers & ActionEvent.CTRL_MASK) != 0 )
                    searchType = 1;     // go backwards

                dtps.get(appstate.getCurrentTabNumber()).search(text, searchType);
            }
            catch (Exception ex)
            {
                Logger.writeLog(ex);
            }
        }
    }

}
