package sita;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import javax.swing.DefaultListModel;
import javax.swing.JDialog;
import javax.swing.ButtonGroup;
import javax.swing.JRadioButton;
import javax.swing.JLabel;

import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import org.sanskritlibrary.Regex;

public class SummaryDialog extends JDialog implements ListSelectionListener, ActionListener
{
    boolean ignoreValueChanged = false;

    JList resultsList;
    DefaultListModel resultsListModel;
    JScrollPane resultsListScrollPane;
    JLabel messageLabel;
    ArrayList<JRadioButton> buttons = new ArrayList<JRadioButton>();
    int currentDisplayType = -1;
    boolean currentManuscriptOnly;
    ButtonGroup bg;
    ApplicationState appstate;

    public SummaryDialog(ApplicationState appstate)
    {
        this.appstate = appstate;

        setLayout(new BorderLayout());

        JPanel topPanel = new JPanel(new BorderLayout());

        bg = new ButtonGroup();
        JRadioButton rb = null;
        JPanel radioPanel = new JPanel(new GridLayout(0, 1));

        int n = 0;
        for (String s : TextAnalyzer.types)
        {
            rb = new JRadioButton(s);
            buttons.add(rb);
            rb.setActionCommand(s);
            rb.setSelected(false);
            rb.addActionListener(this);
            rb.setEnabled(appstate.annotationsSummary.get(n).size() > 0);

            bg.add(rb);
            radioPanel.add(rb);

            n++;
        }


        resultsListModel = new DefaultListModel();
        resultsList = new JList(resultsListModel);
        resultsList.addListSelectionListener(this);
        resultsList.setVisibleRowCount(10);
        resultsListScrollPane = new JScrollPane(resultsList);
        resultsListScrollPane.setPreferredSize(new Dimension(140, 200));
        resultsListScrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);

        topPanel.add(resultsListScrollPane, BorderLayout.CENTER);
        topPanel.add(radioPanel, BorderLayout.EAST);

        add(topPanel, BorderLayout.CENTER);
        messageLabel = new JLabel();
        messageLabel.setPreferredSize(new Dimension(300, 40));
        messageLabel.setText("");
        add(messageLabel, BorderLayout.SOUTH);

        setBounds(0, 0, 350, 250);
        Utils.upperRightScreen(this);
        setModal(false);
        setAlwaysOnTop(true);
    }

    public void display(boolean currentManuscriptOnly)
    {
        ignoreValueChanged = true;
        this.currentManuscriptOnly = currentManuscriptOnly;
        bg.clearSelection();
        for (int i=0; i<TextAnalyzer.types.length; i++)
        {
            JRadioButton rb = buttons.get(i);
            rb.setEnabled(appstate.annotationsSummary.get(i).size() > 0);
        }

        resultsListModel.clear();
        resultsList.clearSelection();
        messageLabel.setText("");
        setVisible(true);
        ignoreValueChanged = false;
    }

    public void valueChanged(ListSelectionEvent e)
    {
        if (ignoreValueChanged)
            return;

        JList src = (JList)e.getSource();
        if (src == resultsList)
            displaySelectedImage();
    }

    private void displaySelectedImage()
    {
        Utils.displayWaitCursor(this);
        ignoreValueChanged = true;
        appstate.getAnnotater().saveState();
        try
        {
            String rootDirectory = appstate.getManuscriptRootDirectory();
            String manuscriptDirectory = appstate.getLastManuscript();
            String lastPage = (String) resultsList.getSelectedValue();
            String extension = "";
            if (!lastPage.toLowerCase().endsWith(".jpg"))
                extension = ".jpg";

            String pageName = appstate.getImageFilename(manuscriptDirectory, lastPage + extension);
            String imagefile = String.format("%s/%s/%s", rootDirectory, manuscriptDirectory, pageName);

            appstate.getAnnotater().setCurrentPage(manuscriptDirectory, lastPage);
            Hashtable<String, String> h = appstate.annotationsSummary.get(currentDisplayType);
            String info = h.get(lastPage);
            int n = 0;
            Regex re = new Regex(info, "(\\d+)");
            if (re.find())
                n = Integer.parseInt(re.getGroup(1));

            messageLabel.setText(info);

            appstate.setMoveToAnnotationNumber(n - 1);
            appstate.getAnnotater().displayManuscript(imagefile);
            appstate.getAnnotater().bringToFront();
        } catch (Exception ex)
        {
            Utils.displayError(ex);
        }
        ignoreValueChanged = false;
        Utils.displayDefaultCursor(this);
    }

    public void actionPerformed(ActionEvent e)
    {
        ignoreValueChanged = true;
        String action = e.getActionCommand();

        currentDisplayType = -1;
        for (String s : TextAnalyzer.types)
        {
            currentDisplayType++;
            if (s.equals(action))
                break;
        }

        Hashtable<String, String> h = appstate.annotationsSummary.get(currentDisplayType);
        ArrayList<String> keys = new ArrayList<String>();
        Enumeration<String> en = h.keys();
        while (en.hasMoreElements())
        {
            keys.add(en.nextElement());
        }

        Collections.sort(keys);

        resultsListModel.clear();
        int i = 0;
        for (String key : keys)
        {
            resultsListModel.add(i, key);
            i++;
        }
        messageLabel.setText("");
        ignoreValueChanged = false;
    }

}
