package sita;

import java.util.ArrayList;
import java.util.Hashtable;
import javax.swing.DefaultListModel;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.TextFormatter;

public class VerseIndexDialog extends JDialog implements ListSelectionListener
{
    boolean ignoreValueChanged = false;

    ArrayList<VerseAnnotation> currentImages;          // contains the image(s) that the selected verse number are along with the annotation with the longest highlight of the selected verse
    Hashtable<Integer, Integer> verses = new Hashtable<Integer, Integer>();
    JLabel label06;
    JList chapterList;
    DefaultListModel chapterListModel;
    JScrollPane chapterListScrollPane;

    JList verseList;
    DefaultListModel verseListModel;
    JScrollPane verseListScrollPane;

    JList resultsList;
    DefaultListModel resultsListModel;
    JScrollPane resultsListScrollPane;

    ApplicationState appstate;

    int lastChapterSelected = -1;
    int lastVerseSelected = -1;
    int lastResultSelected = -1;
    boolean useCurrentManuscript;

    public VerseIndexDialog(ApplicationState appstate)
    {
        this.appstate = appstate;
        
        verses.put(1, 47);
        verses.put(2, 72);
        verses.put(3, 43);
        verses.put(4, 42);
        verses.put(5, 29);
        verses.put(6, 47);
        verses.put(7, 30);
        verses.put(8, 28);
        verses.put(9, 34);
        verses.put(10, 42);
        verses.put(11, 55);
        verses.put(12, 20);
        verses.put(13, 34);
        verses.put(14, 27);
        verses.put(15, 20);
        verses.put(16, 24);
        verses.put(17, 28);
        verses.put(18, 78);

        setLayout(null);

        label06 = new JLabel("6.");
        label06.setBounds(10, 10, 40, 20);
        add(label06);

        chapterListModel = new DefaultListModel();

        chapterList = new JList(chapterListModel);
        chapterList.addListSelectionListener(this);
        chapterList.setVisibleRowCount(10);
        chapterList.setBounds(20, 10, 40, 100);
        chapterListScrollPane = new JScrollPane(chapterList);
        chapterListScrollPane.setBounds(20, 10, 40, 100);

        add(chapterListScrollPane);

        verseListModel = new DefaultListModel();
        verseList = new JList(verseListModel);
        verseList.addListSelectionListener(this);
        verseList.setVisibleRowCount(10);
        verseList.setBounds(65, 10, 40, 100);
        verseListScrollPane = new JScrollPane(verseList);
        verseListScrollPane.setBounds(65, 10, 40, 100);

        add(verseListScrollPane);

        resultsListModel = new DefaultListModel();
        resultsList = new JList(resultsListModel);
        resultsList.addListSelectionListener(this);
        resultsList.setVisibleRowCount(10);
        //resultsList.setBounds(10, 140, 300, 100);
        resultsListScrollPane = new JScrollPane(resultsList);
        resultsListScrollPane.setBounds(10, 140, 300, 100);
        resultsListScrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);

        add(resultsListScrollPane);

        setBounds(0, 0, 320, 280);
        Utils.upperRightScreen(this);
        setModal(false);
        setAlwaysOnTop(true);
    }

    public void display(boolean useCurrentManuscript)
    {
        ignoreValueChanged = true;
        this.useCurrentManuscript = useCurrentManuscript;
        lastResultSelected = -1;

        chapterListModel.clear();
        for (int chap = 1; chap<=18; chap++)
        {
            if (hasOmittedVerses(chap))
            {
                String cs = String.format("%d?", chap+22);
                chapterListModel.add(chap-1, cs);
            }
            else
                chapterListModel.add(chap-1, chap+22);
        }
        chapterList.clearSelection();
        chapterListScrollPane.getVerticalScrollBar().setValue(0);



        verseListModel.clear();
        verseList.clearSelection();

        resultsListModel.clear();
        resultsList.clearSelection();

        ignoreValueChanged = false;
        
        setVisible(true);

    }

    public void valueChanged(ListSelectionEvent e)
    {
        if (ignoreValueChanged)
            return;

        JList src = (JList)e.getSource();
        if (src == chapterList)
        {
            ignoreValueChanged = true;
            int n = chapterList.getSelectedIndex() + 1;
            int numVerses = verses.get(n);
            verseListModel.clear();
            for (int i=1; i<=numVerses; i++)
            {
                String key = String.format("060%d%s", n+22, TextFormatter.intToString(i, 3));
                ArrayList<VerseAnnotation> a = appstate.verseIndex.get(key);
                if (a == null)
                    verseListModel.add(i-1, String.format("%d?", i));
                else if (a.size() > 1)
                    verseListModel.add(i-1, String.format("%d*", i));
                else if (a.size() == 1)
                    verseListModel.add(i-1, Integer.toString(i));
            }
            verseList.clearSelection();
            verseListScrollPane.getVerticalScrollBar().setValue(0);
            ignoreValueChanged = false;
        }
        else if (src == verseList)
        {
            ignoreValueChanged = true;
            int chap = chapterList.getSelectedIndex() + 1;
            int verse = verseList.getSelectedIndex() + 1;
            String key = String.format("060%d%s", chap+22, TextFormatter.intToString(verse, 3));
            currentImages = appstate.verseIndex.get(key);
            resultsListModel.clear();
            if (currentImages == null)
                resultsListModel.add(0, "No annotations for this verse");
            else
            {
                int n = 0;
                for (VerseAnnotation va : currentImages)
                {
                    Regex re = new Regex(va.filename, "(\\p{Nd}+)\\.xml$");
                    String name = va.filename;
                    if (re.find())
                        name = re.getGroup(1) + ".jpg";

                    if (useCurrentManuscript)           // don't bother displaying manuscript name.
                        resultsListModel.add(n, String.format("%s\n", name));
                    else
                        resultsListModel.add(n, String.format("%s/%s\n", va.manuscriptDirectory, name));

                    n++;
                }
            }
            ignoreValueChanged = false;
        }
        else if (src == resultsList)
            displaySelectedImage();
    }


    private boolean hasOmittedVerses(int n)
    {
        int numVerses = verses.get(n);
        for (int i = 1; i <= numVerses; i++)
        {
            String key = String.format("060%d%s", n + 22, TextFormatter.intToString(i, 3));
            ArrayList<VerseAnnotation> a = appstate.verseIndex.get(key);
            if (a == null)
                return true;
        }

        return false;

    }


    private void displaySelectedImage()
    {
        int chapter = chapterList.getSelectedIndex();
        int verse = verseList.getSelectedIndex();
        int n = resultsList.getSelectedIndex();             // to avoid this getting called twice.
        if (chapter == lastChapterSelected && verse == lastVerseSelected && n == lastResultSelected)
            return;
        lastChapterSelected = chapter;
        lastVerseSelected = verse;
        lastResultSelected = n;
        Utils.displayWaitCursor(this);
        ignoreValueChanged = true;
        appstate.getAnnotater().saveState();
        try
        {
            String rootDirectory = appstate.getManuscriptRootDirectory();
            
            String manuscriptDirectory = currentImages.get(n).manuscriptDirectory;      // get the directory that contains the desired page
            String lastPage = (String) resultsList.getSelectedValue();
            Regex re = new Regex(lastPage, "([^\\\\/]+\\.jpg)");
            re.find();
            String abbrev = re.getGroup(1);
            String pageName = appstate.getImageFilename(manuscriptDirectory, abbrev);
            String imagefile = String.format("%s/%s/%s", rootDirectory, manuscriptDirectory, pageName);
            int chapIdx = chapterList.getSelectedIndex() + 1;
            int verseIdx = verseList.getSelectedIndex() + 1;
            String key = String.format("060%d%s", chapIdx+22, TextFormatter.intToString(verseIdx, 3));
            currentImages = appstate.verseIndex.get(key);

            re = new Regex(abbrev, "^(.*)\\.[^.]+$");
            re.find();
            appstate.getAnnotater().setCurrentPage(manuscriptDirectory, re.getGroup(1));
            appstate.setMoveToAnnotationNumber(this.currentImages.get(n).annotationNumber);
            appstate.getAnnotater().displayManuscript(imagefile);
            appstate.getAnnotater().bringToFront();
        } catch (Exception ex)
        {
            Utils.displayError(ex);
        }
        ignoreValueChanged = false;
        Utils.displayDefaultCursor(this);
    }

}
