package org.sanskritlibrary;

import java.util.ArrayList;
import java.util.Hashtable;
import org.w3c.dom.NodeList;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

public class Buffer
{
    public String name;
    public String init;
    public FSMEntry initfse;
    public String card;
    public Merge merge;
    public String existsString;
    public String[] exists;
    public Buffers buffers;
    public boolean or = true;
    private int countAdjustment = 0;
    StringBuilder sb = new StringBuilder();
    private String accumulatedExistsString = null;

    public ActionType[] actions;
    public ArrayList<Member> members = new ArrayList<Member>(); // default actions for all members
    public Hashtable<String, ErrorMessage> errorMessages = new Hashtable<String, ErrorMessage>();

//    public Buffer(String name, String action, String card, Merge merge, String existsString, String init, Buffers buffers) throws Exception
    public Buffer(Buffers buffers, XML doc) throws Exception
    {
        this.buffers = buffers;
        this.name= doc.getAttribute("name");
        this.init = doc.getOptionalAttribute("init", null);
        if (init != null)
        {
            initfse = buffers.getFSMEntry(init);
            if (initfse == null)
                throw new Exception(String.format("buffer.init (%s) not found", init));
        }
        this.card = doc.getOptionalAttribute("card", null);

        String mergeValue = doc.getOptionalAttribute("merge", null);
        if (mergeValue != null)
            merge = buffers.getMerge(mergeValue);

        String actionValue = doc.getOptionalAttribute("action", null);
        if (actionValue != null)
            this.actions = Member.parseActions(actionValue);

        this.existsString = doc.getOptionalValue("exists", null);
        parseExistsString(existsString);

        NodeList memberList = doc.getRelativeNodeList("member");
        for (int i=0; i<memberList.getLength(); i++)
        {
            doc.pushNode(memberList.item(i));
            members.add(new Member(this, doc));
            doc.popNode();
        }

        NodeList errorMessageList = doc.getRelativeNodeList("errorMessage");
        for (int i=0; i<errorMessageList.getLength(); i++)
        {
            doc.pushNode(errorMessageList.item(i));
            addErrorMessage(new ErrorMessage(doc));
            doc.popNode();
        }
    }

    public void accumulateContent(String content)
    {
        if (accumulatedExistsString == null)
            accumulatedExistsString = content;
        else
            accumulatedExistsString += content;
    }

    public void parseAccumulatedExistsString()
    {
        parseExistsString(accumulatedExistsString);
    }

    public void parseExistsString(String existsString)
    {
        if (existsString != null)
        {
            if (existsString.contains("|"))
            {
                this.exists = existsString.split("\\|");
                or = true;
            }
            else if (existsString.contains("&"))
            {
                this.exists = existsString.split("\\&");
                or = false;
            }
            else
            {
                exists = new String[1];
                exists[0] = existsString;
                or = true;
            }
        }
    }

    public Buffer(Buffers buffers, Attributes atts) throws SAXException
    {
        this.buffers = buffers;
        this.name= atts.getValue("name");
        this.init = atts.getValue("init");
        if (init != null)
        {
            initfse = buffers.getFSMEntry(init);
            if (initfse == null)
                throw new SAXException(String.format("buffer.init (%s) not found", init));
        }
        this.card = atts.getValue("card");

        String mergeValue = atts.getValue("merge");
        if (mergeValue != null)
            merge = buffers.getMerge(mergeValue);

        String actionValue = atts.getValue("action");
        if (actionValue != null)
        {
            try
            {
                this.actions = Member.parseActions(actionValue);
            }
            catch (Exception ex)
            {
                throw new SAXException(ex.getMessage());
            }
        }
        this.existsString = atts.getValue("exists");
        if (existsString != null)
        {
            if (existsString.contains("|"))
            {
                this.exists = existsString.split("\\|");
                or = true;
            }
            else if (existsString.contains("&"))
            {
                this.exists = existsString.split("\\&");
                or = false;
            }
            else
            {
                exists = new String[1];
                exists[0] = existsString;
                or = true;
            }
        }
    }

    public void addMember(Member member)
    {
        members.add(member);
    }

    public void addErrorMessage(ErrorMessage errorMessage)
    {
        errorMessages.put(errorMessage.type, errorMessage);
    }

    public void clearEntries()
    {
        for (Member member : members)
        {
            member.clearEntries();
        }
    }


    public String makeString()
    {
        sb.setLength(0);
        for (Member member : members)
        {
            sb.append(member.makeString());
        }
        
        if (merge != null)
        {
            if (sb.length() > 1)
            {
                String out = merge.getMergedValue(sb.toString());
                if (out != null)
                {
                    countAdjustment = out.length() - sb.length();
                    if (countAdjustment != 0)
                        buffers.setCheckForErrors(false);     // too complicated to figure out what goes where
                    return out;
                }
            }
        }

        return sb.toString();
    }


    public void register(Hashtable<String, Member> h)
    {
        for (Member member : members)
        {
            member.register(h);
        }
    }

    // Note card and exists are both based on count which is not reset when member is cleared!
    public String getError(Hashtable<String, Member>h)
    {
        if (buffers.getCheckForErrors())
        {
            boolean requiredSatisfied = true;

            int count = countAdjustment;
            countAdjustment = 0;
            for (Member member : members)
            {
                count += member.count;
                if (member.required && member.count == 0)
                    requiredSatisfied = false;
            }

            // if count has been adjusted assume for now that result is what we are looking for. This is getting too complicated
            if (!requiredSatisfied && count > 0)
                return String.format("no required member");

            if (card != null)
            {
                if (card.equals("?") && count > 1)
                    return String.format("name buffer has %d characters. Must have exactly 1.", count);
            }

            if (exists != null && count > 0)
            {
                boolean okay = false;
                for (String s : exists)
                {
                    Member m = h.get(s);
                    if (m == null)
                        m = null;
                    if (m.count>0 || m.fsmEntry.size()>0)       // 2nd test for mapped members
                    {
                        if (or)
                        {
                            okay = true;
                            break;
                        }
                    } else if (!or)
                    {
                        okay = false;
                        break;
                    }
                }

                if (!okay)
                    return String.format("buffer %s does not satisfy %s", name, existsString);
            }
        }

        return null;
    }

    public void initialize() throws Exception
    {
        int count = 0;

        for (Member m : members)
        {
            count += m.getEntryCount();
        }

        if (count == 0)
        {
            // using the removed field is a hack. It was added because of a java.util.ConcurrentModificationException
            // that occured when I tried to physically remove it. Evidently fsmEntry was already being iterated.
            Member member = buffers.getMember(initfse.klass);

            for (FSMEntry fe : member.fsmEntry)         // if it is already there reset removed
            {
                if (fe == initfse)
                {
                    fe.removed = false;
                    return;
                }
            }

            member.addFSMEntry(initfse);   // otherwise add it.
            initfse.removed = false;
        }
    }

    public void clearDefaultMember() throws Exception
    {
        if (initfse != null)
        {
            Member m = buffers.getMembers().get(initfse.klass);
            if (m != null)
            {
                for (FSMEntry fe : m.fsmEntry)
                {
                    if (fe != initfse)
                    {
                        buffers.setError(true);       // virama used when a vowel sign already in place
                        break;
                    }
                }
            }
            initfse.removed = true;
        }
        else
            throw new Exception("Buffer has no default member");
    }
}
