package org.sanskritlibrary;

import java.util.ArrayList;

public class BufferChars
{
	public BufferChar[] bcs;
	public FSMEntry defaultEntry;
	//public int duplicateCount = 0;


	public BufferChars(ArrayList<FSMEntry> fes, FSMEntry defaultEntry) throws Exception
	{
		int len = fes.size();
		bcs = new BufferChar[len];
		for (int i=0; i<len; i++)
		{
			FSMEntry fe = fes.get(i);
			bcs[i] = new BufferChar(fe, false);
		}
		this.defaultEntry = defaultEntry;
	}

	public int updateBufferChars(FSMEntry fe)
			  throws Exception
	{
		int numUpdated = 0;

		if (fe == defaultEntry)	// don't add default entry if something already in buffer
		{
			for (BufferChar bc : bcs)
			{
				if (bc.present)
					return 0;		// don't need default entry
			}
		}

		for (BufferChar bc : bcs)
		{
			if (bc.fe == fe)
			{
				bc.present = true;
				numUpdated++;
				break;
			}
		}

/*
		if (numUpdated > 0)
		{
			removeDefault(fe);
		}
*/
		return numUpdated;
	}

/*
	private void removeDefault(FSMEntry fe)
	{
		if (fe != defaultEntry)
		{
			for (BufferChar bc : bcs)
			{
				if (bc.fe == defaultEntry)
				{
					bc.present = false;
					break;
				}
			}
		}
	}
*/

	public void remove(FSMEntry fe)
	{
		for (BufferChar bc : bcs)
		{
			if (bc.fe == fe)
			{
				bc.present = false;
				break;
			}
		}
	}

	public void clear()
	{
		for (BufferChar bc : bcs)
		{
			bc.present = false;
		}
	}

	public boolean clear(FSMEntry fe)
	{
		for (BufferChar bc : bcs)
		{
			if (bc.fe == fe)
			{
				if (bc.present)
				{
					bc.present = false;
					return true;
				}
				else
					return false;
			}
		}
		return false;
	}

	public int resetPresent()
	{
		int count = 0;
		for (BufferChar bc : bcs)
		{
			if (bc.present)
			{
				count++;
				bc.present = false;
			}
		}
		return count;
	}

	// This can't be toString() because it changes state!
	public String makeString()
	{
		StringBuilder sb = new StringBuilder();
		boolean found = false;
		for (BufferChar bc : bcs)
		{
			if (bc.present)
			{
				sb.append(bc.fe.target);
				found = true;
			}
		}

        String result = sb.toString();

        if (found)
		{
			if (resetPresent() != 1)	// mark duplicate prefixes or suffixes
            {
                if (result.equals("AE") || result.equals("EA"))
                    result = "O";       // danda plus E is O
                else if (result.equals("Ae") || result.equals("eA"))
                    result = "o";        // danda plus e is o
                else
                {
                    result += "!";       // probably an error, mark the second suffix with a !
                    System.out.format("!!!!!%d:%s\n", RTF2.debug, result);
                }
            }
            else
                result = sb.toString();
		}

		return result;
	}
}
