package org.sanskritlibrary;

import java.util.ArrayList;
import java.util.Hashtable;

public class Buffers
{
    StateInfo si;
    StringBuilder sb = new StringBuilder();
    SymbolTable symtab;
    private Hashtable<String, Member> members = new Hashtable<String, Member>();
    private ArrayList<Buffer> buffers = new ArrayList<Buffer>();
    private Hashtable<String, Merge> merges = new Hashtable<String, Merge>();
    private boolean bufferReady = false;
    private boolean checkForErrors = true;
    private boolean baseFound = false;          // set to true when a non-dependant entry is added to a buffer.
    private boolean error = false;               // an error was detected before buffers concatenated, e.g. dependent entry added before baseFound=true
    private Buffers b;                          // This may point to a shared buffer

    public Buffers(SymbolTable symtab, StateInfo si)
    {
        this.symtab = symtab;
        this.si = si;
        this.b = this;
    }

    public Buffers b()
    {
        return b;
    }

    // use members and buffer elements from another buffer
    public void initialize(Buffers sharedBuffers)
    {
        this.b = sharedBuffers;
    }

    // Getters and setters

    public Hashtable<String, Member> getMembers()
    {
        return b.members;
    }

    public void setCheckForErrors(boolean checkForErrors)
    {
        b.checkForErrors = checkForErrors;
    }

    public void setError(boolean error)
    {
        b.error = error;
    }

    public boolean getBaseFound()
    {
        return b.baseFound;
    }

    public void setBaseFound(boolean baseFound)
    {
        b.baseFound = baseFound;
    }

    public boolean getCheckForErrors()
    {
        return b.checkForErrors;
    }

    public Member getMember(String name)
    {
        return b.members.get(name);
    }

    public FSMEntry getFSMEntry(String name)
    {
        return symtab.getFSMEntry(name);
    }

    public void setErrorMark()
    {
        si.getResult().append(errorMark());
    }

    public void setBufferReady(boolean bufferReady)
    {
        b.bufferReady = bufferReady;
    }

    public String makeString(boolean clear)
    {
        b.sb.setLength(0);
        for (Buffer buffer : b.buffers)
        {
            b.sb.append(buffer.makeString());
        }

        if (b.error)
            b.sb.append(errorMark());
        else if (b.checkForErrors)     // will be false if either a merge took place or an action='x' occured
        {
            // Now check for errors
            boolean errorExists = false;
            for (Buffer buffer : b.buffers)
            {
                if (buffer.getError(b.members) != null)
                    errorExists = true;
            }
            if (errorExists)
                b.sb.append(errorMark());
        }


        if (clear)
        {
            b.checkForErrors = true;
            b.bufferReady = false;        // force this to be reset to true
            b.baseFound = false;
            b.error = false;

            for (Buffer buffer : b.buffers)
            {
                buffer.clearEntries();
            }
        }

        return b.sb.toString();
    }

    String errorMark()
    {
        return String.format("[!%d]", RTF2.debug);
    }

    // adds fe to member after performing any actions
    private void performActions(FSMEntry fe) throws Exception
    {
        Member member = b.members.get(fe.key);       // try specific first
        if (member == null)
        {
            if (fe.klass == null)
                throw new Exception(String.format("Entry %s has no <cl> child", fe.id));
            member = b.members.get(fe.klass);
        }

        if (member.actions!=null)
            member.performActions(fe);
        else if (member.buffer.actions != null)     // member inherits actions from buffer
        {
            member.actions = member.buffer.actions;
            member.performActions(fe);
        }
        else
        {
            if (member.buffer.init != null)  // remove default entry, if any
                member.removeFSMEntry(member.buffer.initfse);
            member.addFSMEntry(fe);
        }
    }


    public void appendBuffers(boolean conditional)
    {
        if (conditional && !b.bufferReady)
            return;

        si.getResult().append(makeString(true));
    }

    public Merge getMerge(String name)
    {
        return b.merges.get(name);
    }

    public Buffer getBuffer(String name)
    {
        for (Buffer bfr : b.buffers)
        {
            if (bfr.name.equals(name))
                return bfr;
        }
        
        return null;
    }


    // No explicit sharing done in these methods
    public void addBuffer(Buffer buffer) throws Exception
    {
        if (b != this)
            throw new Exception("Adding a buffer to a shared buffer is not permitted.");
        buffers.add(buffer);
        buffer.register(members);
    }

    public void addMerge(Merge merge) throws Exception
    {
        if (b!=this)
            throw new Exception("Adding a merge to a shared buffer is not permitted.");

        merges.put(merge.name, merge);
    }

    public void update(FSMEntry fe) throws Exception
    {
        if (fe.targets == null)
            performActions(fe);
        else
        {
            for (FSMEntry fse : fe.targets)
            {
                performActions(fse);
            }
        }
    }
}

