package org.sanskritlibrary;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.nio.charset.Charset;
import java.util.Hashtable;
import org.sanskritlibrary.admin.VerseInfoEx;
import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;
import org.w3c.dom.Node;

public class CardonaFile
{
    private Hashtable<String, FSM[]> fsmh;
    private ResourceInputStream ris;
    private TransformMap tm;

    public CardonaFile(Hashtable<String, FSM[]> fsmh, ResourceInputStream ris) throws Exception
    {
        this.fsmh = fsmh;
        this.ris = ris;
        tm = new TransformMap("*:sat->slp1", ris, fsmh);
    }


    public void convertToXml(String indir, String outdir, Node textnode) throws Exception
    {
        String line = null;
        Node versePattern = XML.getRelativeNode(textnode, "versePattern");
        String filename = XML.getRelativeNode(textnode, "fileName").getTextContent().trim();
        String encoding = XML.getRelativeNode(textnode, "originalEncoding/encoding").getTextContent().trim();

        String infile = String.format("%s/%s.%s", indir, filename, encoding);
        String xmlfile = String.format("%s/%s.xml", outdir, filename);
        BufferedWriter bw = null;
        BufferedReader br = null;
        if (encoding.toLowerCase().equals("sat"))
        {
            br = IOAssistance.openReader(infile, new RawDecoder(Charset.forName("UTF-8"), 1F, 1F));
            bw = IOAssistance.openWriter(xmlfile, new RawEncoder());
        }
        else
        {
            br = IOAssistance.openReader(infile);
            bw = IOAssistance.openWriter(xmlfile);
        }

        bw.write("<lines>\n");
        while ((line = br.readLine()) != null)
        {
            try
            {
                VerseInfoEx vi = new VerseInfoEx(line, versePattern);
                if (vi.isError)
                    bw.write(String.format("\t<line matched='f'>%s</line>\n", line));
                else
                {
                    bw.write("\t<line>\n");
                    int numtags = vi.names.length;
                    for (int i=0; i<numtags; i++)
                    {
                        String value = vi.values[i];
                        if (vi.transcode[i])
                            value = WebServices.transformString(value, tm);
                        bw.write(String.format("\t\t<%s>%s</%s>\n", vi.names[i], value, vi.names[i]));
                    }
                    bw.write("\t</line>\n");
                }
            }
            catch (Exception ex)
            {
                System.out.println(ex.getMessage());
                bw.write(String.format("\t<line matched='f'>%s</line>\n", line));
            }
        }
        bw.write("</lines>");
        br.close();
        bw.close();
    }

	public String removeRedundantCharacters(String line)
	{
		if (line == null)
		{
			return line;
		}

		char satlongudatta = '\u00DA';	// {shift option 1: dead key preceding the vowel it accents}
		char satlongsvarita = '\u00D9';
		char satmacron = '\u00CF';
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < line.length(); i++)
		{
			char current = line.charAt(i);
			if (skip(line, i, current, satlongudatta) ||
					  skip(line, i, current, satlongsvarita) ||
					  skip(line, i, current, satmacron))
			{
				continue;
			}
			sb.append(current);
		}
		return sb.toString();
	}

	private boolean skip(String s, int i, char current, char ch)
	{
		if (i == s.length() - 1)
		{
			return current == ch;
		} else
		{
			return current == ch && s.charAt(i + 1) == ch;
		}
	}
    
}
