package org.sanskritlibrary;

import java.io.BufferedWriter;
import java.util.Hashtable;

/*
<fonttbl>	'{' \fonttbl (<fontinfo> | ('{' <fontinfo> '}'))+ '}'
<fontinfo>	<themefont>? \fN <fontfamily> \fcharsetN? \fprq? <panose>? <nontaggedname>? <fontemb>? \cpgN? <fontname> <fontaltname>? ';'
<themefont>	\flomajor | \fhimajor | \fdbmajor | \fbimajor | \flominor | \fhiminor |
\fdbminor | \fbiminor
<fontfamily>	\fnil | \froman | \fswiss | \fmodern | \fscript | \fdecor | \ftech | \fbidi
<panose>	'{\*' \panose <data> '}'
<nontaggedname>	'{\*' \fname #PCDATA ';}'
<fontname>	#PCDATA
<fontaltname>	'{\*' \falt #PCDATA '}'
<fontemb>	'{\*' \fontemb <fonttype> <fontfname>? <data>? '}'
<fonttype>	\ftnil | \fttruetype
<fontfname>	'{\*' \fontfile \cpgN? #PCDATA '}'
 */
public class FonttblParser extends Parser
{

    static public Hashtable<String, FontInfo> fonts;
    static public Hashtable<String, Boolean> themeFont;

    static
    {
        themeFont = new Hashtable<String, Boolean>();
        themeFont.put("flomajor", Boolean.TRUE);
        themeFont.put("fhimajor", Boolean.TRUE);
        themeFont.put("fdbmajor", Boolean.TRUE);
        themeFont.put("fbimajor", Boolean.TRUE);
        themeFont.put("flominor", Boolean.TRUE);
        themeFont.put("fhiminor", Boolean.TRUE);
        themeFont.put("fdbminor", Boolean.TRUE);
        themeFont.put("fbiminor", Boolean.TRUE);

        fonts = new Hashtable<String, FontInfo>();
    }
    BufferedWriter out;
    String unicodeFontTemplate;

    public FonttblParser()
    {
        unicodeFontTemplate = "{\\f%d\\fswiss\\fcharset128\\fprq2{\\*\\panose 020b0604020202020204}Arial Unicode MS;}";
    }


    public void setUnicodeFontTemplate(String unicodeFontTemplate)
    {
        this.unicodeFontTemplate = unicodeFontTemplate;
    }

    public boolean canHandle(RTF2 rtf, int n) throws Exception
    {
        if (rtf.getRTF().charAt(n) != '{')
            throw new Exception("FonttblParser expecting a '{'");

        n = rtf.skipEOLs(n);

        if (n == 8998) {
            n = 8998;
        }

        Token t = rtf.nextToken(n + 1);
        if (t.controlWord == null) {     // probably have a \r immediately following a {
            return false;
        }
        else {
            return t.controlWord.equals("fonttbl");
        }
    }

    public int parse(RTF2 rtf, int n, ParseState ps) throws Exception
    {
        String s = rtf.getRTF();
        out = rtf.getOutput();
        n = startup(rtf, n, "FonttblParser");       // writes out starting { plus white space
        Token t = rtf.nextToken(n);
        if (!t.controlWord.equals("fonttbl"))
            return n;                               // this will be handled later by handleToken (see below)
        fonts.clear();
        t.dump(out);
        n = t.end;
        boolean inBrackets = false;
        String fontNumber = null;
        do
        {
            // <fontinfo>	<themefont>? \fN <fontfamily> \fcharsetN? \fprq? <panose>? <nontaggedname>? <fontemb>? \cpgN? <fontname> <fontaltname>? ';'
            n = rtf.writeEOLs(n);
            inBrackets = s.charAt(n) == '{';

            if (inBrackets)
            {
                rtf.write('{');
                n++;
            }

            boolean more = true;
            do
            {
                if (s.charAt(n) == '{')
                {
                    n = dumpGroup(s, n);
                    more = true;
                } else
                {
                    t = rtf.nextToken(n);
                    more = t.controlWord != null;
                    if (more)
                    {
                        if (t.controlWord.equals("f"))
                            fontNumber = t.value;

                        t.dump(out);
                        n = t.end;
                    }
                }
            } while (more);	// loop until we get to the font name


            // should be at fontname now
            if (t.value == null)
                throw new Exception("Why is fontname null?");
            String fontName = t.value.toLowerCase();
            if (fontName.endsWith(";"))
                fontName = fontName.substring(0, fontName.length() - 1);
            if (fontNumber == null)
                throw new Exception("Why is font number null in FonttblParser?");

            fonts.put(fontNumber, new FontInfo(fontNumber, fontName, false));
            t.dump(out);
            n = t.end;

            n = rtf.writeEOLs(n);
            if (s.charAt(n) == '{')	// {fontaltname}?
                n = dumpGroup(s, n);

            n = rtf.writeEOLs(n);
            if (s.charAt(n) == ';')
            {
                n++;
                rtf.write(";");
            }

            if (inBrackets)
            {
                rtf.write('}');
                n++;
            }

            fontNumber = null;
            n = rtf.writeEOLs(n);
        } while (s.charAt(n) != '}');

        // see if there are any default fonts that need resolving
        int handle = ps.getCurrentHandle();
        for (int i = handle; i >= 0; i--)
        {
            String dfn = ps.get("_defaultFont", i);
            if (dfn != null)		// this will be a font number of the default font.
            {
                FontInfo fi = fonts.get(dfn);
                if (fi != null)
                    ps.update("_defaultFont", fi.fontName, i);
            }
        }

        if (RTF2.convertToUnicode)
        {
            // add entry for Arial Unicode. Don't care if one already exists
            for (int i = 1; i < 999; i++)
            {
                if (fonts.get(Integer.toString(i)) == null)
                {
                    String entry = String.format(unicodeFontTemplate, i);
                    rtf.write(entry);
                    handle = ps.getCurrentHandle();	// probably will be 0, i.e., the only other state is \rtf
                    for (int j = handle; j >= 0; j--)
                    {
                        ps.update("_unicodeFontPrefix", String.format("\\loch\\f%d \\hich\\f%d ", i, i), j);
                    }
                    break;
                }
            }
        }

        rtf.write(s.charAt(n++));       // write out final brace
        return n;  // points to character immediately after ending brace
    }

    private int dumpGroup(String s, int n) throws Exception
    {
        int start = n;
        while (s.charAt(n) != '}')
        {
            n++;
        }
        n++;	// move past '}'
        if (out != null)
            out.write(s.substring(start, n));

        return n;
    }

    // this is called whenever a \fN control word is encounter (outside of a fonttbl group)
    public void handleToken(Token t, ParseState ps, StringBuffer accumulateBuffer) throws Exception
    {
        if (t.controlWord == null)
            throw new Exception("No control word");
        String controlWord = t.controlWord;
        if (controlWord.equals("dn"))       // do not move characters down
        {
            t.value = "0";
            t.actual = "0";
        }
        else
        {
            boolean regularFont = controlWord.equals("f");
            boolean associatedFont = controlWord.equals("af");
            boolean defaultFont = controlWord.equals("deff") || controlWord.equals("adeff");

            FontInfo fi = fonts.get(t.value);
            if (regularFont || associatedFont)
            {
                if (fi == null)
                    throw new Exception(String.format("font number %s not found", t.value));
                else	// remember the fontName
                {
                    fi.used = true;
                    String key = null;
                    if (regularFont)
                        key = "_font";
                    else
                        key = "_afont";
                    if (fi.fontName.equals("times"))
                        fi.fontName = "times";
                    ps.addControlWord(key, fi.fontName);
                }
            } else if (defaultFont)
            {
                if (fi != null)
                    ps.addControlWord("_defaultFont", fi.fontName);
                else		// probably fonttbl has not been encountered yet.
                    ps.addControlWord("_defaultFont", t.value);
            }
        }
    }

    // this is called whenever a \fN control word is encounter (outside of a fonttbl group)
    public void handleToken(Token t, ParseState ps) throws Exception
    {
        handleToken(t, ps, null);
    }
}
