package org.sanskritlibrary;
import java.io.BufferedReader;
import java.io.BufferedInputStream;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.BufferedWriter;
import java.io.BufferedOutputStream;
import java.io.OutputStreamWriter;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.nio.CharBuffer;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.Charset;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.File;
import java.io.FileDescriptor;
import java.io.FileWriter;

public class IOAssistance
{
	static public BufferedReader openReader(String inputFile, String... encoding)
			  throws IOException
	{
		BufferedReader reader = null;
		String enc = "UTF-8";
		if (encoding.length > 0)
		{
			enc = encoding[0];
		}
		BufferedInputStream bis = null;
		InputStreamReader isr = null;
		FileInputStream fis = new FileInputStream(inputFile);
		bis = new BufferedInputStream(fis);

		isr = new InputStreamReader(bis, enc);
		reader = new BufferedReader(isr);
		return reader;
	}

	static public BufferedReader openReader(InputStream is, String... encoding)
			  throws IOException
	{
		BufferedReader reader = null;
		String enc = "UTF-8";
		if (encoding.length > 0)
		{
			enc = encoding[0];
		}
		BufferedInputStream bis = new BufferedInputStream(is);
		InputStreamReader isr = new InputStreamReader(bis, enc);
		reader = new BufferedReader(isr);

		return reader;
	}

	static public BufferedReader openReader(String inputFile, CharsetDecoder encoding)
			  throws IOException
	{
		if (encoding == null)
			return openReader(inputFile, "UTF-8");
		BufferedReader reader = null;
		BufferedInputStream bis = null;
		InputStreamReader isr = null;
		FileInputStream fis = new FileInputStream(inputFile);
		bis = new BufferedInputStream(fis);

		isr = new InputStreamReader(bis, encoding);
		reader = new BufferedReader(isr);
		return reader;
	}

	static public BufferedReader openReader(InputStream is, CharsetDecoder encoding)
			  throws IOException
	{
		if (encoding == null)
			return openReader(is, "UTF-8");
		BufferedReader reader = null;
		BufferedInputStream bis = new BufferedInputStream(is);
		InputStreamReader isr = new InputStreamReader(bis, encoding);
		reader = new BufferedReader(isr);

		return reader;
	}

	static public String readFile(BufferedReader br) throws Exception
	{
		StringBuffer sb = new StringBuffer();
		try
		{
			String line = null;
			while ((line = br.readLine()) != null)
			{
				sb.append(line);
				sb.append("\n");
			}
		}
		catch (Exception ex)
		{
			throw ex;
		}
		finally
		{
			if (br != null)
				br.close();
		}
		return sb.toString();
	}

	static public String readFile(String fileName) throws Exception
	{
		return readFile(openReader(fileName));
	}

	static public String readFile(String fileName, String encoding) throws Exception
	{
		return readFile(openReader(fileName, encoding));
	}

	static public String readFile(String fileName, CharsetDecoder encoding) throws Exception
	{
		return readFile(openReader(fileName, encoding));
	}

	static public FileWriter openUncachedWriter(String outputFile) throws IOException
	{
		FileWriter fw = new FileWriter(outputFile);
		return fw;
	}



    static public BufferedWriter openWriter(String outputFile, Charset... encodingName)
			  throws IOException
	{
		Charset encoding = null;
		if (encodingName.length == 0)
			encoding = Charset.forName("UTF-8");
		else
			encoding = encodingName[0];
		
		BufferedWriter writer = null;
		BufferedOutputStream bos = null;
		OutputStreamWriter osw = null;
		FileOutputStream fos = new FileOutputStream(outputFile);
		bos = new BufferedOutputStream(fos);

		osw = new OutputStreamWriter(bos, encoding);
		writer = new BufferedWriter(osw);

		return writer;
	}

	static public BufferedWriter openWriter(String outputFile, CharsetEncoder encoding)
			  throws IOException
	{
		if (encoding == null)
			return openWriter(outputFile, Charset.forName("UTF-8"));
		BufferedWriter writer = null;
		BufferedOutputStream bos = null;
		OutputStreamWriter osw = null;
		FileOutputStream fos = new FileOutputStream(outputFile);
		bos = new BufferedOutputStream(fos);

		osw = new OutputStreamWriter(bos, encoding);
		writer = new BufferedWriter(osw);

		return writer;
	}

	public static String readEntireFile(String fileName, String... charset) throws Exception
	{
		String cs = null;
		if (charset.length == 0)
			cs = "UTF-8";
		else
			cs = charset[0];

		FileInputStream fis = new FileInputStream(fileName);
		int len = fis.available();
		byte[] b= new byte[len];
		fis.read(b);
		fis.close();
		return new String(b, 0, b.length, cs);
	}

	public static String readEntireFile(String inputFile, CharsetDecoder encoding) throws Exception
	{
		if (encoding == null)
			return readEntireFile(inputFile);
		BufferedReader reader = null;
		BufferedInputStream bis = null;
		InputStreamReader isr = null;
		FileInputStream fis = new FileInputStream(inputFile);
		bis = new BufferedInputStream(fis);

		isr = new InputStreamReader(bis, encoding);
        BufferedReader br = new BufferedReader(isr);
        StringBuilder sb = new StringBuilder();
        int ch = 0;
        while ((ch=br.read())!=-1)
        {
            sb.append((char)ch);
        }
        br.close();
		return new String(sb.toString());
	}

    public static Object readObject(String fileName) throws Exception
	{
		FileInputStream fis = new FileInputStream(fileName);
		ObjectInputStream ois = new ObjectInputStream(fis);
		Object obj = ois.readObject();
		ois.close();
		return obj;
	}

	public static void writeObject(String fileName, Object obj) throws Exception
	{
		FileOutputStream fos = new FileOutputStream(fileName);
		ObjectOutputStream oos = new ObjectOutputStream(fos);
		oos.writeObject(obj);
		oos.close();
	}

	public static void append(String directory, String fileName, String text, String delim) throws IOException
	{
		String pathName = String.format("%s/%s", directory, fileName);
        if (delim != null)
            text += delim;
        append(pathName, text);
	}

	public static void append(String pathName, String text) throws IOException
	{
		File f = new File(pathName);
		if (!f.exists())
			f.createNewFile();
		FileOutputStream fos = new FileOutputStream(pathName, true);
		BufferedOutputStream bos = new BufferedOutputStream(fos);

		OutputStreamWriter osw = new OutputStreamWriter(bos, Charset.forName("UTF-8"));
		osw.append(text);
		osw.close();
	}

    public static boolean fileExists(String name)
	{
		File f = new File(name);
		return f.exists();
	}
}
