package org.sanskritlibrary;

import java.util.Hashtable;
import org.xml.sax.Attributes;

public class MatchItem
{
	public String _classCollection;
	public int maxEdgeLength = 0;
	public Hashtable<String, Boolean> classCollection;
	public String option;
	public boolean leftToRight;
    public boolean exists;
    public boolean eolOverride;

    public MatchItem()
    {

    }

    public MatchItem(XML doc) throws Exception
    {
        _classCollection = doc.getTextContent();
        // card attribute "*" zero or more, "+" 1 or more, "?" 0 or 1, "1" exactly 1
        option = doc.getOptionalAttribute("card", "+");
        leftToRight = doc.getOptionalAttribute("dir", "after").toLowerCase().equals("after");
        String x = doc.getOptionalAttribute("x", null);
        exists = x != null;
        if (exists)
            parseXAttribute(x);
    }


    public MatchItem(Attributes atts)
    {
        option = atts.getValue("card");
        if (option == null)
            option = "+";
        
        String value = atts.getValue("dir");
        if (value == null)
                leftToRight = true;
        else
                leftToRight = value.toLowerCase().equals("after");

        String x = atts.getValue("x");
        exists =   x != null;
        if (exists)
            parseXAttribute(x);
    }
    
    private void parseXAttribute(String existance)
    {
        String[] flds = existance.toLowerCase().split("\\|");
        for (String fld : flds)
        {
            if (fld.equals("y"))
               this.exists = true;
            else if (fld.equals("n"))
                this.exists = false;
            else if (fld.equals("eol"))
                this.eolOverride = true;
        }
    }

	public String matchRight(String s, int idx)
	{
        // try longest edges first
		for (int len = maxEdgeLength; len>=1; len--)
		{
			if (idx + len > s.length())   // len is too long, continue loop to get shorter length
				continue;
            String key = s.substring(idx, idx+len);
			if (classCollection.get(key) != null)
                return key;
		}

		return null;
	}

	public String matchLeft(String s, int idx)
	{
        // try longest edges first
		for (int len = maxEdgeLength; len>=1; len--)
		{
			if (idx - len < 0)   // len is too long, continue loop to get a shorter length
				continue;
            String key = s.substring(idx-len, idx);
			if (classCollection.get(key) != null)
                return key;
		}

		return null;
	}

    public void resolveclassCollection(SymbolTable symtab) throws Exception
    {
        if (_classCollection != null)
        {
            classCollection = symtab.getClassCollection(_classCollection);
            maxEdgeLength = symtab.getMaxKeyLength(_classCollection);
        }
    }

}
