// http://latex2rtf.sourceforge.net/RTF-Spec-1.0.txt (show values of control words)
package org.sanskritlibrary;

import java.io.BufferedWriter;
import java.io.File;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.ArrayList;
import org.sanskritlibrary.webservice.WebServices;
import org.sanskritlibrary.webservice.TransformMap;
public class RTF2
{
    public static Hashtable<String, Hashtable<String, Integer>> usedChars = new Hashtable<String, Hashtable<String, Integer>>();
    static private Hashtable<String, Boolean> charCodes = new Hashtable<String, Boolean>();
	static public boolean convertToUnicode = true;
    static public boolean ignoreUnicodeCharacters = true;   // this should be true only for fonts that we are converting, otherwise leave unicode characters alone.
	static public int debug = 0;
    static public boolean doDump = false;
	static private int[] hexDigit = new int[256];
	public static Hashtable<String, FSM[]> fsmh = new Hashtable<String, FSM[]>();
	static
	{
		for (int i=0; i<hexDigit.length; i++)
		{
			hexDigit[i]= -1;
		}
		hexDigit['0'] = 0;
		hexDigit['1'] = 1;
		hexDigit['2'] = 2;
		hexDigit['3'] = 3;
		hexDigit['4'] = 4;
		hexDigit['5'] = 5;
		hexDigit['6'] = 6;
		hexDigit['7'] = 7;
		hexDigit['8'] = 8;
		hexDigit['9'] = 9;
		hexDigit['a'] = 10;
		hexDigit['A'] = 10;
		hexDigit['b'] = 11;
		hexDigit['B'] = 11;
		hexDigit['c'] = 12;
		hexDigit['C'] = 12;
		hexDigit['d'] = 13;
		hexDigit['D'] = 13;
		hexDigit['e'] = 14;
		hexDigit['E'] = 14;
		hexDigit['f'] = 15;
		hexDigit['F'] = 15;

        charCodes.put("u", true);
        charCodes.put("bullet", true);
        charCodes.put("lquote", true);
        charCodes.put("rquote", true);
        charCodes.put("emdash", true);
        charCodes.put("endash", true);
        charCodes.put("ldblquote", true);
        charCodes.put("rdblquote", true);
	}
    private boolean skipNonStandard = true;
	private BufferedWriter out;
    private BufferedWriter log;
    private BufferedWriter extractionRaw;
    private BufferedWriter extractionProcessed;
	private String rtf;
	private int len;
	private Rules rules;
	private File file;
	private String platform;
    private boolean clearStateSet = false;
    private boolean clearState = true;
    private boolean accumulateToEOL = false;
    private StringBuilder accumulationBufferRaw = new StringBuilder();
    private StringBuilder accumulationBufferSLP1 = new StringBuilder();
    private FonttblParser fp;
    private String defaultFont;
	Hashtable<String, TransformMap> map;
    Hashtable<String, Boolean> dependantText = new Hashtable<String, Boolean>();

    private Parser passThroughParser;
    private ArrayList<Integer> textNumber = new ArrayList<Integer>();             // number of current chunk of text.

	public RTF2(String fileName, Rules rules, Hashtable<String, TransformMap> map,
			  BufferedWriter out, String platform, String defaultFont) throws Exception
	{
		if (fileName != null)
		{
			rtf = IOAssistance.readEntireFile(fileName);
			this.file = new File(fileName);
			len = rtf.length();
		}

		this.out = out;
		this.map = map;
		this.rules = rules;
		this.platform = platform;
        this.defaultFont = defaultFont;
        this.passThroughParser = new PassThroughParser();
	}

    // use a lot of defaults
    public RTF2(String fileName, String platform) throws Exception
    {
		if (fileName != null)
		{
			rtf = IOAssistance.readEntireFile(fileName);
			this.file = new File(fileName);
			len = rtf.length();
		}
        this.map = new Hashtable<String, TransformMap>();

        fp = new FonttblParser();
        
        PassThroughParser pt = new PassThroughParser();

        this.rules = new Rules();
        this.rules.add("fonttbl", fp);
        this.rules.add("stylesheet", pt);
        this.rules.add("colortbl", pt);
        this.rules.add("themedata", pt);
        this.rules.add("colorschememapping", pt);
        this.rules.add("latentstyles", pt);
        this.rules.add("datastyles", pt);
        this.rules.add("f", fp);
        this.rules.add("af", fp);
        this.rules.add("deff", fp);
        this.rules.add("adeff", fp);
        this.rules.add("dn", fp);

        this.platform = platform;
    }


    public void setUnicodeFontFormat(String unicodeFontTemplate)
    {
        fp.setUnicodeFontTemplate(unicodeFontTemplate);
    }

    public RTF2(String platform) throws Exception
    {
        this(null, platform);       // assumes that setRTF will be called.
    }

    public void addMap(String fontname, TransformMap tm)
    {
        map.put(fontname, tm);
    }

    public void accumulateToEOL(boolean b)
    {
        accumulateToEOL = b;
    }

    public void setSkipNonStandard(boolean b)
    {
        this.skipNonStandard = b;
    }

    public void setLog(BufferedWriter log)
    {
        this.log = log;
    }

    public void setOutputFile(BufferedWriter out)
    {
        this.out = out;
    }

	public void setRtf(String test)
	{
		rtf = test;
		len = rtf.length();
	}

    public void setExtractionRaw(BufferedWriter extractionRaw)
    {
        this.extractionRaw = extractionRaw;
    }

    public void setExtractionProcessed(BufferedWriter extractionProcessed)
    {
        this.extractionProcessed = extractionProcessed;
    }

	public String getRTF()
	{
		return rtf;
	}

	public File getFile()
	{
		return file;
	}

	public BufferedWriter getOutput()
	{
		return out;
	}

    public int removeSymbols() throws Exception
    {
        return removeSymbols(0, new ParseState(), 1);
    }

    public int removeSymbols(int n, ParseState ps, int level) throws Exception
	{
        int start = n;
		char ch = rtf.charAt(n++);
		if (ch != '{')
			throw new Exception("Must begin with {");
		n = skipEOLs(n);
		Token t = nextToken(n);
		Parser p = null;
		if (t.controlWord != null)
		{
            p = rules.lookup(t.controlWord);
            if (p == null && t.nonStandard != null && skipNonStandard)
                p = passThroughParser;

			if (p != null)
			{
				n = p.parse(this, start, ps);		// let custom parser do this group
				return writeEOLs(n);	// move past any whitespace
			}
		}

        write(rtf.substring(start, n));     // write out initial brace and any whitespace that follows it.

		ps.pushState(t.controlWord);

        do
		{
            ch = rtf.charAt(n);
			if (ch == '{')
				n = removeSymbols(n, ps, level + 1);
            else
            {
                t.dump(out);
                n = t.end;
            }

            n = writeEOLs(n);
            if (n >= len)
                throw new Exception("missing }");

            t = nextToken(n);
		} while (rtf.charAt(n) !='}');

        write("}");
		ps.popState();
		return writeEOLs(n + 1);
	}

	public int parse() throws Exception
	{
		if (rtf.charAt(0) != '{')
			throw new Exception(String.format("*****%s not an rtf file\n", file.getAbsolutePath()));
		else
		{
            if (convertToUnicode && !accumulateToEOL)   // make initial pass to find character dependencies
            {
                BufferedWriter saveOut = out;
                out = null;
                textNumber.clear();
                parse(0, new ParseState(), 0);          // first pass to get character dependencies
                out = saveOut;
            }

            textNumber.clear();
            int n = parse(0, new ParseState(), 0);  // second pass to do transcoding;

            if (file != null)
                System.out.println(getFile().getName());
            Enumeration<String> en = FonttblParser.fonts.keys();
            while(en.hasMoreElements())
            {
                FontInfo fi = FonttblParser.fonts.get(en.nextElement());
                if (fi.used)
                    System.out.format("\t%s\n", fi.fontName);
            }

			return n;
		}
	}



	// assume it starts with a '{'
	public int parse(int n, ParseState ps, int level) throws Exception
	{
        if (textNumber.size() == level)
            textNumber.add(0);

        int start = n;
		char ch = rtf.charAt(n++);
		if (ch != '{')
			throw new Exception("Must begin with {");
		n = skipEOLs(n);
		Token t = nextToken(n);
		Parser p = null;
		if (t.controlWord != null)
		{
            if (t.controlWord.equals("clearState") && t.nonStandard != null)
            {
                clearStateSet = true;
                clearState = t.value.equals("1");
                return PassThroughParser.skipGroup(rtf, n);
            }
            else
            {
                p = rules.lookup(t.controlWord);
                if (p == null && t.nonStandard != null && skipNonStandard)
                    p = passThroughParser;

                if (p != null && p.canHandle(this, start))
                {
                    n = p.parse(this, start, ps);		// let custom parser do this group
                        return writeEOLs(n);	// move past any whitespace
                }
            }
		}

        write(rtf.substring(start, n));     // write out initial brace and any whitespace that follows it.

		ps.pushState(t.controlWord);

        do
		{
            if (out != null)
            {
                debug++;
                if (debug == -1)
                    doDump = true;
            }

            ch = rtf.charAt(n);
			if (ch == '{')
				n = parse(n, ps, level + 1);
            else
            {
                String fontName = ps.get("_font");
				if (fontName == null)
                    fontName = ps.get("_afont");
				if (fontName == null)
					fontName = ps.get("_defaultFont");
                TransformMap tm = null;
                if (fontName != null)
                    tm = map.get(fontName.toLowerCase());
                else if (defaultFont != null)
                    tm = map.get(defaultFont.toLowerCase());

                if (t.controlWord != null)
                {
                    p = rules.lookup(t.controlWord);
                    if (p != null)
                        p.handleToken(t, ps);		// handleToken will update state.
                    else
                        ps.addControlWord(t.controlWord, t.value);
                    if (accumulateToEOL)
                    {
                        if (!t.controlWord.equals("f"))
                            dumpAccumulationBuffer(tm, ps);
                        if (t.controlWord.equals("par"))
                        {
                            accumulationBufferRaw.append("\n");
                            accumulationBufferSLP1.append("\n");                            
                        }
                    }
                    
                    t.dump(out);
                }
                else if (t.value.length() > 0)	// have text? (will have length 0 if \rs or \ns encountered.)
                {
                    if (doDump)
                    {
                        String ds = tm.dumpText(t.value);
                        BufferedWriter bw = IOAssistance.openWriter("doDump.txt");
                        bw.write(ds);
                        bw.close();
                        doDump = false;
                    }
                    if (fontName == null && defaultFont != null)
                        fontName = defaultFont;

                    if (fontName == null) {
                        fontName = null;
                    }
                    if (fontName != null)   // will be null only if text appears before \fonttbl group which is pathological
                    {
                        if (tm != null)
                        {
                            int tn = textNumber.get(level) + 1;
                            textNumber.set(level, tn);
                            if (out == null)
                            {
                                checkFirstCharacter(tm, t.value, level, tn);         // remember if first character cannot begin a syllable.
                                computeCharFrequencies(fontName, t.value);
                            }
                            else
                            {
                                if (n == 8432) {
                                    n = 8432;
                                }
                                if (convertToUnicode)
                                    outputText(tm, ps, fontName, t, level, tn);
                                else
                                    write(t.actual);
                            }
                        }
                        else
                            write(t.actual);    // send it through unmodified.
                    }
                    else
                        write(t.actual);        // should never reach here, but this seems to be a safe thing to do
                }

                n = t.end;
            }

            n = writeEOLs(n);
            if (n >= len)
                throw new Exception("missing }");

            t = nextToken(n);
		} while (rtf.charAt(n) !='}');

        write("}");
		ps.popState();
		return writeEOLs(n + 1);
	}

    private void checkFirstCharacter(TransformMap tm, String text, int level, int tn) throws Exception
    {
        if (tm.isDependantCharacter(text.charAt(0)))
            dependantText.put(String.format("%s.%d.%d", tm.getVirtualFontname(), level, tn), true);
    }

    private void outputText(TransformMap tm, ParseState ps, String fontName, Token t, int level, int tn) throws Exception
    {
		if (out != null)
        {
            if (tm != null)
            {
                if (this.accumulateToEOL)
                    outputTextAccumulate(tm, fontName, t);
                else
                    outputTextByFont(tm, ps, fontName, t, level, tn);
            }
            else
                write(t.actual);        // no conversion necessary
        }
    }

    private void dumpAccumulationBuffer(TransformMap tm, ParseState ps) throws Exception
    {
        if (tm != null && accumulationBufferSLP1.length() > 0)
        {
            // force a buffer update by sending an empty string.
            tm.setClearState(true);
            WebServices.transformString("", tm, 0, accumulationBufferSLP1);  // convert to slp1 only and accumulate result in accumulateBuffer;

            String unitext = WebServices.transformString(accumulationBufferSLP1.toString(), tm, 1, null);  // e.g., convert Vedic 97 to unicode via SLP1
            out.write(createUnicode(unitext, ps));
            if (extractionRaw != null)
            {
                String s = accumulationBufferRaw.toString();
                extractionRaw.write(s);
                if (!s.endsWith("\n"))
                    extractionRaw.write("\n");
            }
        
            accumulationBufferRaw.setLength(0);
            accumulationBufferSLP1.setLength(0);
        }
    }

    private void outputTextByFont(TransformMap tm, ParseState ps, String fontName, Token t, int level, int tn) throws Exception
    {
        computeCharFrequencies(fontName, t.value);
        boolean _clearState = true;

        if (clearStateSet)      // preprocessor inserted {\*\clearState} group
        {
            clearStateSet = false;  // force it to be turned on again.
            _clearState = clearState;
        } else
            _clearState = this.dependantText.get(String.format("%s.%d.%d", tm.getVirtualFontname(), level, tn + 1)) == null;

        tm.setClearState(_clearState);     // if next text begins with dependant character, then don't clear state.
        String unitext = WebServices.transformString(t.value, tm);  // e.g., convert Vedic 97 to unicode via SLP1
        out.write(createUnicode(unitext, ps));
        doExtractions(fontName, t.value, _clearState);

    }

    private void outputTextAccumulate(TransformMap tm, String fontName, Token t) throws Exception
    {
        computeCharFrequencies(fontName, t.value);
        String key = String.format("%s$%s", t.value);
        if (!tm.isOverriddenCharacter(key))
            accumulationBufferRaw.append(t.value);
        else
            accumulationBufferRaw.append("k");      // replace with something that dependent entities can attach to.
        tm.setClearState(false);   // in case there is some more devanagari in another font.
        WebServices.transformString(t.value, tm, 0, accumulationBufferSLP1);  // convert to slp1 only and accumulate result in accumulateBuffer;
    }

    private String createUnicode(String text, ParseState ps) throws Exception
	{
        StringBuilder sb = new StringBuilder();

        String unicodeFontPrefix = ps.get("_unicodeFontPrefix");  // if not there try all current states.
        if (unicodeFontPrefix == null)
                throw new Exception("Why is unicodeFontPrefix null?");
        sb.append(unicodeFontPrefix);

        int textlen = text.length();
		String uc = ps.get("uc");
        if (uc == null)
            uc = "1";

		for (int i=0; i<textlen; i++)
		{
            char ch = text.charAt(i);
            if (ch < 0x80)
            {
                if (ch == '{')
                    sb.append("\\'7b");
                else if (ch == '}')
                    sb.append("\\'7d");
                else
                    sb.append(ch);
            }
            else
            {
                sb.append(TextFormatter.charToSlashUDecimal(text.charAt(i)));
                if (uc != null && uc.equals("1"))
                    sb.append("\\'3F");
            }
		}

        return sb.toString();
	}


    public void write(String s) throws Exception
    {
        if (out != null)
            out.write(s);
    }


    public void write(char ch) throws Exception
    {
        if (out != null)
            out.write(ch);
    }


    public int writeEOLs(int n) throws Exception
	{
        int start = n;
        n = skipEOLs(n);
        if (out != null && start != n)
            out.write(rtf.substring(start, n));

        return n;
	}


	public int skipEOLs(int n) throws Exception
	{
		while (true)
		{
			if (n < len)
			{
				char ch = rtf.charAt(n);
				if (ch == '\r' || ch == '\n')
					n++;
				else
					break;
			}
			else
				break;
		}

		return n;
	}

    private char getChar(int n)
	{
		if (n >= len)
			return ' ';
		else
			return rtf.charAt(n);
	}

	public Token nextToken(int start) throws Exception
	{
        Token returnToken =  null;
		StringBuffer sb = new StringBuffer();
		StringBuffer sbActual = new StringBuffer();
		String whiteSpaceRemoved = null;
		int end = start;
		int beginValue = start;
		String controlWord =  null;
		String value = null;
		String nonStandard = null;
		int inc = 1;
		if (getChar(start) == '\\')		// assume that \ is not escaped
		{
			end = skipEOLs(start + 1);
			if (getChar(end) == '*')
			{
				end = skipEOLs(end+1) + 1;  // In case eols follow the \*
				nonStandard = rtf.substring(start, end-1);	// remember what happened with \* so it can be dumped
			}
			inc = end - start;		// will be > 1 if there are eols
            if (Character.isLetter(getChar(end)))		// if it is a bonafide control word instead of a hex character
			{
				while (end < len)		// get control word
				{
					char ch = getChar(end);
					if (Character.isLetter(ch))
						end++;
					else
						break;
				}
				controlWord = rtf.substring(start+inc, end);
                if (charCodes.get(controlWord) == null) // if control word is not a char code, e.g, \bullet, \lquote, \rquote, \endash, \emdash
                {
                    if (end < len && getChar(end) == ' ')
                    {
                        whiteSpaceRemoved = " ";
                        end++;		// blank belongs to control word
                    }
                }
                else
                {
                    end = start;        // process char code as text instead of a control word.
                    controlWord = null;
                }
			}
            else if (isSpecialCharacter(getChar(end)))      // is it optional hyphen, non-breaking hyphen or non-breaking space?
            {
                end++;
                controlWord = rtf.substring(start+inc, end);    // assume that a following blank is text and does not belong to special character
            }
			else
				end = start;			// point back to beginning of hex character (note: whitespace may be  between hex character and its \?

			beginValue = end;
		}

		if (whiteSpaceRemoved == null)		// if either no control word or there is one but it does not end in a blank
		{
			if (controlWord != null)		// have a control word so we are getting value (maybe), assume no hex
			{
				// assume that value is either a positive or negative integer?
				int cnt = 0;
				while (end < len)
				{
					char ch = getChar(end);
					if (cnt == 0 && ch == '-' && Character.isDigit(getChar(end+1)))
						end++;
					else if (!Character.isDigit(ch))		// assume no hex notation
						break;	// have found next control word or end of bracketed expression
					else
						end++;
					cnt++;			// to detect leading minus
				}

				if (beginValue != end)		// if we have a value
				{
					value = rtf.substring(beginValue, end);
					char ch = getChar(end);
					if (controlWord != null && Character.isWhitespace(ch))
					{
						whiteSpaceRemoved = new String(new char[] {ch});
						end++;
					}
				}
				returnToken = new Token(controlWord, value, value, start, end, whiteSpaceRemoved, nonStandard);
			}
			else		// have some text
			{
				while (true)
				{
                    int temp = end;
                    end = skipEOLs(end);
                    if (temp != end)
                        sbActual.append(rtf.substring(temp, end));      // ?? Do I want to append EOLs to sb also?
					char ch1 = getChar(end);
					char ch2 = getChar(end+1);
					if (ch1 == '\\')		// have a hex value, an escaped \,{,} or a control word
					{
						if (ch2 == '\'')		// have \'xx
						{
							sbActual.append(getChar(end));
							sbActual.append(getChar(end+1));
							end += 2;
							ch1 = getChar(end);
							ch2 = getChar(end+1);
							sbActual.append(getChar(end));
							sbActual.append(getChar(end+1));
							end += 2;
							sb.append(makeHex(ch1, ch2));
						}
                        else if (isSpecialCharacter(ch2))
                        {
                            break;      // treat this the same as another other control word.
                        }
						else if (!Character.isLetter(ch2))	// have escaped \\ { or }, or some other character (-)
						{
							sb.append(ch2);
							sbActual.append(ch1);
							sbActual.append(ch2);
							end += 2;
						}
						else
						{
							Token tt = resolveEntity(end);

							if (tt == null)
								break;		// must have hit a control word
							else
							{
                                if (tt.value.charAt(0) != 0xffff)       // ignore unicode character
                                {
                                    sb.append(tt.value);
                                    sbActual.append(tt.actual);
                                }
								end = tt.end;
							}
						}
					}
					else if (ch1 == '{' || ch1 == '}')
						break;
					else	// have a regular character
					{
						if (ch1 != 0x0D && ch1 !=0x0A)		// ignore eol characters
							sb.append(ch1);
						sbActual.append(ch1);
						end++;
					}
				}
				returnToken = new Token(null, sb.toString(), sbActual.toString(), start, end, whiteSpaceRemoved,
										nonStandard);
			}
		}
		else		// return controlWord with no value
			returnToken = new Token(controlWord, null, null, start, end, whiteSpaceRemoved,
									nonStandard);

        if (returnToken == null) {      // should never happen
            return null;
        } else {
            if (returnToken.controlWord != null && returnToken.controlWord.equals("up") &&
                returnToken.value!=null && returnToken.value.equals("4")) {
                returnToken.controlWord = "up";
            }
            return returnToken;
        }

	}


    private boolean isSpecialCharacter(char ch)
    {
        boolean rc = ch=='~' ||      // non-breaking space
                ch=='-' ||      // optional hyphen
                ch=='_';        // non-breaking hyphen
        return rc;
    }

	/*
	 \tab \t
	 \par \n

	*/

	// to avoid an infinite recursive loop. Note n is position of \
	Token resolveEntity(int n)
	{
		String entity = null;
		int start = n;
		int end = start + 1;        // move past \
		char ch = '\0';

        while (true)
        {
            ch = getChar(end);
            if (!Character.isLetter(ch))
                break;
            end++;
        }
		
		if (end == start+1)
			return null;

		entity = rtf.substring(start+1, end);
        if (entity.equals("u"))             // is this a decimal unicode character?
        {
            while (true)
            {
                ch = getChar(end);
                if (!Character.isDigit(ch))
                    break;
                end++;
            }
            if (end != len && getChar(end)==' ')        // can we ignore a blank after a unicode character entity? Let's assume yes.
                end++;      // skip white space that belongs to unicode entity

            String actual = rtf.substring(start, end);
            char[] temp = new char[] {(char)Integer.parseInt(rtf.substring(start+2, end).trim())};
            if (ignoreUnicodeCharacters)
                temp[0] = 0xffff;                       // this will be detected by caller
            return new Token(null,
							new String(temp),         // ignore the character, this will only have an effect if current text is being transcoded.
							actual,
							start,
							end,
							null,
							null);
        }
        else
        {
            if (Character.isWhitespace(ch))
                end++;

            if (platform.equals("mac"))
            {
                if (entity.equals("lquote"))
                    ch = 0xD4;		// 212
                else if (entity.equals("rquote"))
                    ch = 0xD5;		// 213
                else if (entity.equals("bullet"))
                    ch = 0xA5;		// 165
                else if (entity.equals("endash"))
                    ch = 0xD0;		// 208
                else if (entity.equals("emdash"))
                    ch = 0xD1;		// 209
                else if (entity.equals("ldblquote"))
                    ch = 0xD2;      // 210
                else if (entity.equals("rdblquote"))
                    ch = 0xD3;      // 211
                else
                    ch = '\0';
            }
            else if (platform.equals("windows"))
            {
                if (entity.equals("lquote"))
                    ch = 0x91;      // 145
                else if (entity.equals("rquote"))
                    ch = 0x92;      // 146
                else if (entity.equals("bullet"))
                    ch = 0x95;      // 149
                else if (entity.equals("endash"))
                    ch = 0x96;      // 150
                else if (entity.equals("emdash"))
                    ch = 0x97;      // 151
                else if (entity.equals("ldblquote"))
                    ch = 0x93;      // 147
                else if (entity.equals("rdblquote"))
                    ch = 0x94;      // 148
                else
                    ch = '\0';
            }

            if (ch == '\0')
                return null;
            else
            {
                String actual = rtf.substring(start, end);
                return new Token(null,
							new String(new char[]{ch}),
							actual,
							start,
							end,
							null,
							null);
            }
        }
	}

	private char makeHex(char ch1, char ch2) throws Exception
	{
		if (hexDigit[ch1] != -1 && hexDigit[ch2] != -1)
			return (char)(hexDigit[ch1] * 16 + hexDigit[ch2]);
		else
			throw new Exception(String.format("Unexpected hex: %c%c", ch1, ch2));
	}



	public void close() throws Exception
	{
		if (out != null)
			out.close();

        if (log != null)
            log.close();

        if (extractionRaw != null)
            extractionRaw.close();

        if (extractionProcessed != null)
            extractionProcessed.close();
	}




    private void computeCharFrequencies(String fontName, String s)
    {
        Hashtable<String, Integer> h = usedChars.get(fontName);
        if (h == null)
        {
            h = new Hashtable<String, Integer>();
            usedChars.put(fontName, h);
        }

        int numchars = s.length();
        for (int i=0; i<numchars; i++)
        {
            String key = TextFormatter.charToSlashU(s.charAt(i));
            Integer cnt = h.get(key);
            if (cnt == null)
                h.put(key, 1);
            else
            {
                cnt += 1;
                h.put(key, cnt);
            }
        }
    }

    // For purpose of comparing with Indian fellow's transcoding program
    private void doExtractions(String fontName, String text, boolean clearState) throws Exception
    {
        if (text.length() > 0)
        {
            if (fontName.equals("madhushree"))
            {
                if (extractionRaw != null)
                {
                    extractionRaw.write(text);
                    if (clearState)
                        extractionRaw.write(" ");       // so that leading vowel letters do not become attached vowel signs
                }
                if (false && extractionProcessed != null)
                {
                    TransformMap tm = map.get(fontName);
                    String uni = WebServices.transformString(text, tm);  // e.g., convert Vedic 97 to unicode via SLP1
                    extractionProcessed.write(uni);
                    extractionProcessed.write(" ");       // see comment above.
                }
            }
        }
    }

	public ArrayList<String> getUsedFonts() throws Exception
	{
        ArrayList<String> a = new ArrayList<String>();
        skipNonStandard = true;
        out = null;

		if (rtf.charAt(0) != '{')
			throw new Exception(String.format("*****%s not an rtf file\n", file.getAbsolutePath()));
		else
		{
            _getUsedFonts(0, new ParseState(), 0);

            Enumeration<String> en = FonttblParser.fonts.keys();
            while(en.hasMoreElements())
            {
                FontInfo fi = FonttblParser.fonts.get(en.nextElement());
                if (fi.used)
                    a.add(fi.fontName);
            }

			return a;
		}
	}



	// assume it starts with a '{'
	private int _getUsedFonts(int n, ParseState ps, int level) throws Exception
	{
        int start = n;
		char ch = rtf.charAt(n++);
		if (ch != '{')
			throw new Exception("Must begin with {");
		n = skipEOLs(n);
		Token t = nextToken(n);
		Parser p = null;
		if (t.controlWord != null)
        {
            p = rules.lookup(t.controlWord);
            if (p == null && t.nonStandard != null && skipNonStandard)
                p = passThroughParser;

            if (p != null)
            {
                n = p.parse(this, start, ps);		// let custom parser do this group
                return writeEOLs(n);	// move past any whitespace
            }
        }
		ps.pushState(t.controlWord);

        do
		{
            ch = rtf.charAt(n);
			if (ch == '{')
				n = _getUsedFonts(n, ps, level + 1);
            else
            {
                if (t.controlWord != null)
                {
                    p = rules.lookup(t.controlWord);
                    if (p != null)
                        p.handleToken(t, ps);		// handleToken will update state.
                }

                n = t.end;
            }

            n = writeEOLs(n);
            if (n >= len)
                throw new Exception("missing }");

            t = nextToken(n);
		} while (rtf.charAt(n) !='}');
        ps.popState();
		return writeEOLs(n + 1);
	}

    // For best results open fontDisplay.rtf in Word and then resave it.
    static public void createRTFTable(String fontName) throws Exception
    {
       String header = "{\\rtf1\\adeflang1103\\ansi\\ansicpg1252\\uc1{\\fonttbl{\\f1\\fbidi \\froman\\fcharset2\\fprq2 %s;}}\n";
       BufferedWriter bw = IOAssistance.openWriter("fontDisplay.rtf");
       bw.write(String.format(header, fontName));
       String row = "{\\lrtch \\af1 \\i0 %d\\cell    %s\\cell }\\pard\\plain \\ltrpar\\ql \\li0\\ri0\\sa200\\sl276\\slmult1\\widctlpar\\intbl\\wrapdefault\\aspalpha\\aspnum\\faauto\\adjustright\\rin0\\lin0 \\rtlch\\fcs1 \\af\\afs20\\alang1103 \\ltrch\\fcs0 \\f1\\fs22\\lang1033\\langfe1033\\cgrid\\langnp1033\\langfenp1033 {\\rtlch\\fcs1 \\af1 \\ltrch\\fcs0 \\insrsid9728254 \\trowd \\irow0\\irowband0\\ltrrow\\ts15\\trgaph108\\trleft-108\\trbrdrt\\brdrs\\brdrw10\\brdrcf17 \\trbrdrl\\brdrs\\brdrw10\\brdrcf17 \\trbrdrb\\brdrs\\brdrw10\\brdrcf17 \\trbrdrr\\brdrs\\brdrw10\\brdrcf17 \\trbrdrh\\brdrs\\brdrw10\\brdrcf17 \\trbrdrv\\brdrs\\brdrw10\\brdrcf17 \\trftsWidth1\\trftsWidthB3\\trautofit1\\trpaddl108\\trpaddr108\\trpaddfl3\\trpaddft3\\trpaddfb3\\trpaddfr3\\tblrsid9728254\\tbllkhdrrows\\tbllkhdrcols\\tbllknocolband\\tblind0\\tblindtype3 \\clvertalt\\clbrdrt\\brdrs\\brdrw10\\brdrcf17 \\clbrdrl\\brdrs\\brdrw10\\brdrcf17 \\clbrdrb\\brdrs\\brdrw10\\brdrcf17 \\clbrdrr\\brdrs\\brdrw10\\brdrcf17 \\cltxlrtb\\clftsWidth3\\clwWidth4788\\clshdrawnil \\cellx4680\\clvertalt\\clbrdrt\\brdrs\\brdrw10\\brdrcf17 \\clbrdrl\\brdrs\\brdrw10\\brdrcf17 \\clbrdrb\\brdrs\\brdrw10\\brdrcf17 \\clbrdrr\\brdrs\\brdrw10\\brdrcf17 \\cltxlrtb\\clftsWidth3\\clwWidth4788\\clshdrawnil \\cellx9468\\row \\ltrrow}\\pard\\plain \\ltrpar\\ql \\li0\\ri0\\widctlpar\\intbl\\wrapdefault\\aspalpha\\aspnum\\faauto\\adjustright\\rin0\\lin0\\yts15 \\rtlch\\fcs1 \\af1\\afs20\\alang1103 \\ltrch\\fcs0 \\f1\\fs22\\lang1033\\langfe1033\\cgrid\\langnp1033\\langfenp1033\n";

       for (int i=0; i<16; i++)
       {
           bw.write(String.format(row, i, "\\'0" + Integer.toHexString(i)));
       }
/*
       bw.write(String.format(row, 11, "\\'0" + Integer.toHexString(11)));
       bw.write(String.format(row, 12, "\\'0" + Integer.toHexString(12)));
       bw.write(String.format(row, 13, "\\'0" + Integer.toHexString(13)));
       bw.write(String.format(row, 14, "\\'0" + Integer.toHexString(14)));
       bw.write(String.format(row, 15, "\\'0" + Integer.toHexString(15)));
*/
       for (int i=16; i<256; i++)
       {
           bw.write(String.format(row, i, "\\'" + Integer.toHexString(i)));
       }

       bw.write("}");
       bw.close();
    }


    public static void createTextTable() throws Exception
    {
       BufferedWriter bw = IOAssistance.openWriter("allchars.txt", new RawEncoder());
       for (int i=16; i<256; i++)
       {
           bw.write(String.format("%d\t%s\t%c\n", i, TextFormatter.toHexString(i), (char)i));
       }
       bw.close();
    }
}