// Generic SAX parser. Delegates all calls to a SAXParserHandler

package org.sanskritlibrary;

import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import javax.xml.parsers.SAXParserFactory;
import org.xml.sax.helpers.DefaultHandler;

public class SAXParser extends DefaultHandler {
    private SAXParserHandler handler;
    private SAXParserFactory factory;

    public SAXParser (SAXParserHandler handler) {
        this.handler = handler;
        factory = SAXParserFactory.newInstance();
    }

    public void parse (String filename) throws Exception {
        try {
            javax.xml.parsers.SAXParser saxParser = factory.newSAXParser();
            saxParser.parse(filename, this);
        } catch (Exception ex) {
            throw ex;
        }
    }

    @Override
    public void warning(SAXParseException exception) throws SAXException {
        handler.warning(exception);
    }

    @Override
    public void error(SAXParseException exception) throws SAXException {
        handler.error(exception);
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException {
        handler.fatalError(exception);
    }

    @Override
    public void setDocumentLocator(Locator locator) {
        handler.setDocumentLocator(locator);
    }

    @Override
    public void startDocument() throws SAXException {
        handler.startDocument();
    }

    @Override
    public void endDocument() throws SAXException {
        handler.endDocument();
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException {
        handler.startPrefixMapping(prefix, uri);
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException {
        handler.endPrefixMapping(prefix);
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        handler.startElement(uri, localName, qName, atts);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        handler.endElement(uri, localName, qName);
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {
        handler.characters(ch, start, length);
    }


}
