package org.sanskritlibrary;

import java.io.File;
import java.util.ArrayList;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;

public class SandhiHeuristics  extends DefaultHandler
{
    Locator locator;
    ArrayList<SandhiRule> sandhiRules = new ArrayList<SandhiRule>();

    public SandhiHeuristics(String filename) throws Exception
    {
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            saxParser.parse( new File(filename), this);
        }
        catch (Exception t)
        {
            throw new Exception(String.format("Line %d, Column %d: %s", locator.getLineNumber(), locator.getColumnNumber(), t.getMessage()));
        }
    }

    public ArrayList<String> getAlternates(String word)
    {
        ArrayList<String> a = new ArrayList<String>();
        for (SandhiRule sandhiRule : sandhiRules)
        {
            String result = Regex.replace(word, sandhiRule.match, sandhiRule.replace);
            if (!result.equals(word))
                a.add(result);
        }
        
        return a;
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        if (qName.equals("regexp"))
        {
            String match = atts.getValue("match");
            String replace = atts.getValue("replace");
            if (match != null && replace != null)
                sandhiRules.add(new SandhiRule(match, replace));
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
    }

    // For ErrorHandler
    @Override
    public void warning(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void error(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException
    {
        throw exception;
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
    }

    @Override
    public void endDocument() throws SAXException
    {
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    class SandhiRule
    {
        String match;
        String replace;

        SandhiRule(String match, String replace)
        {
            this.match = match;
            this.replace = replace;
        }
    }
}
