package org.sanskritlibrary;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;

public class SegmentSignature extends DefaultHandler
{
    private String fn;
    private boolean debugging = false;
    private boolean debug = false;
    private int offset = 0;
    private Regex intcheck = new Regex("", "^\\d+$");
    private int[] marginSize = new int[]{1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21};
    private String[] delims = null;
    private String tabString = "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t";
    private Locator locator;
    private Hashtable<String, DocumentInfo> h = new Hashtable<String, DocumentInfo>();
    private String currentFile;
    private ArrayList<String> lastSectionValues = new ArrayList<String>();
    private Hashtable<String, FSM[]> fsmh;
    private ResourceInputStream ris;
    private BufferInfo[] bi = new BufferInfo[3];
    private int pageBufferSize = 20000;
    private int verseNumber = 0;
    private Regex wordre =  new Regex("","[a-zA-Z|]");
    private Regex segre = new Regex("", "seg\\[(....)\\]");
    private TransformMap tmRoman;


    public SegmentSignature(String filename, Hashtable<String, FSM[]> fsmh, ResourceInputStream ris, boolean debug) throws Exception
    {
        this.fsmh = fsmh;
        this.ris = ris;
        this.debug = debug;
        tmRoman = new TransformMap("*:slp1->roman", ris, fsmh);
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            saxParser.parse( new File(filename), this);
        }
        catch (Exception t)
        {
            throw t;
        }
    }

    public Hashtable<String, DocumentInfo> getHashtable()
    {
        return h;
    }


    public static void generateCategories(String inputdir) throws Exception
    {
        String[] subdirs =
        {
            "class", "CLASSICAL",
            "jskt",  "EPIC",
            "mbh", "EPIC",
            "ram", "EPIC",
            "vedic", "VEDIC"
        };

        Hashtable<String, String> h = new Hashtable<String,String>();

        Regex re = new Regex("", "seg\\[(....)\\]:");
        int len = subdirs.length;
        for (int i=0; i<len; i+=2)
        {
            String subdir = subdirs[i];
            String category = subdirs[i+1];
            File dir = new File(String.format("%s/%s", inputdir, subdir));
            File[] files = dir.listFiles();
            for (File file : files)
            {
                String fn = file.getName();
                fn = fn.replace(".xml", "");
                h.put(fn, category);
            }
        }

        Enumeration<String> en = h.keys();
        ArrayList<String> a = new ArrayList<String>();
        while (en.hasMoreElements())
        {
            a.add(en.nextElement());
        }

        Collections.sort(a);

        for (String filename : a)
        {
            String category = h.get(filename);
            String flatname = String.format("Catalogs/flattened/%s.txt", filename);
            BufferedReader br = IOAssistance.openReader(flatname);
            String line = null;
            boolean isContinuous = false;
            boolean isSandhi = false;
            while ((line = br.readLine())!= null)
            {
                if (line.length() == 0 || line.charAt(0) == '+')
                    continue;
                re.reset(line);
                if (re.find())
                {
                    String type = re.getGroup(1);
                    if (type.charAt(0) != 'c')          // if it is not commentary
                    {
                        if (type.charAt(1) == 'c')
                            isContinuous = true;
                        else if (type.charAt(1) == 's')
                            isSandhi = true;
                    }
                }
                
                if (isContinuous && isSandhi)
                    break;
            }
            br.close();
            System.out.format("%s: TITUS|%s|", filename, category);
            
            if (isContinuous && isSandhi)
                System.out.println("CONTINUOUS|SANDHI");
            else if (isContinuous)
                System.out.println("CONTINUOUS");
            else if (isSandhi)
                System.out.println("SANDHI");
        }
    }


    public void generateStructures(String inputdir, String outputdir) throws Exception
    {
        StringBuilder vid = new StringBuilder();
        File dir = new File(inputdir);
        File[] files = dir.listFiles();
        for (File file : files)
        {
            verseNumber = 0;
            offset = 0;
            String fn = file.getName();
            System.out.println(fn);
            DocumentInfo di = h.get(fn);
            if (di != null)
            {
                if (di.name.equals("sbm.txt"))
                    di.name = "sbm.txt";

                // create RE signature pattern for this text
                int numSections = di.sections.size();
                int lastSection = numSections;          // lastSection is 1-based
                String[] values = new String[numSections];
                StringBuilder patternBuilder = new StringBuilder();
                int n = 1;
                for (SectionInfo si : di.sections)
                {
                    patternBuilder.append(si.value);
                    patternBuilder.append("\\[([^\\]]+)");
                    if (n < lastSection)
                        patternBuilder.append(".*");
                    else
                        patternBuilder.append(".*?: (.*)");
                    n++;
                }

                String pattern = patternBuilder.toString();
                Regex re = new Regex("", pattern);

                fn = fn.substring(0, fn.length()-4);
                BufferedReader br = IOAssistance.openReader(String.format("%s/%s.txt", inputdir, fn));
                BufferedWriter bwIndex = IOAssistance.openWriter(String.format("%s/%s.index", outputdir, fn));
                bi[0] = new BufferInfo(outputdir, fn, "slp1");
                bi[1] = new BufferInfo(outputdir, fn, "roman");
                bi[2] = new BufferInfo(outputdir, fn, "deva");

                bwIndex.write("<script type='text/javascript'>\n");
                bwIndex.write("var baseOffset = 0;\n");
                bwIndex.write(String.format("var numLevels = %d;\n", numSections));
                bwIndex.write("var levelNames = [");
                String delim = "";
                delims = new String[numSections];
                for (int i=0; i<numSections; i++)
                {
                    bwIndex.write(String.format("%s'%s'", delim, di.sections.get(i).display));
                    delim = ", ";
                    delims[i] = generateTabs(marginSize[i]);
                }
                bwIndex.write("];\n");
                bwIndex.write("var $lists = [];\n");
                bwIndex.write("var bookGeometry =\n");
                bwIndex.write("[\n");
                bwIndex.write("\t{\n");
                bwIndex.write("\t\tv: 0,\n");
                bwIndex.write("\t\ts:\n");
                bwIndex.write("\t\t[\n");
                br.readLine();      // skip Titus URL
                String line;
                String text = "";
                int lineno = 2;

                while ((line = br.readLine()) != null)      // for each line of flattened text
                {
                    if (lineno == 1009)
                        lineno = 1009;
                    boolean newSection = false;
                    re.reset(line);
                    if (re.find())
                    {

                        int gc = re.getGroupCount();
                        for (int i=1; i<=gc-1; i++)             // note: group numbers are 1-based and the gc-th group is the text
                        {
                            int j = i-1;                        // make group number 0-based
                            String s = re.getGroup(i);
                            String v = values[j];               // get last value of this group
                            if (v == null || !v.equals(s))      // if no previous value or the value is new
                            {
                                newSection = true;
                                if (i != gc - 1)                // write ], } for each group that follows the i-th group
                                {
                                    closePreviousElements(bwIndex, v, i, numSections);
                                    write(bwIndex, 0, String.format("%s{\n", delims[i]));
                                    delims[i] = ",\n" + generateTabs(marginSize[i]);
                                    write(bwIndex, marginSize[i] + 1, String.format("v: '%s',\n", s));
                                }
                                if (i < gc - 2)     // if there will definitely be a sections array
                                {       // note: group gc-1 is the last group and hence has no sections
                                        //       group gc-2 is the next to the last group and will not have a sections array if the last group is sequentially numbered.
                                    write(bwIndex, marginSize[i]+1, "s:\n");
                                    write(bwIndex, marginSize[i]+1, "[\n");
                                    delims[i+1] = generateTabs(marginSize[i+1]);
                                }
                                else if (i == gc-1)
                                    lastSectionValues.add(s);       // at last group, save so that sequentiality of values can be tested later

                                if (i != gc-1)                      // if not at last group
                                {
                                    for (int k=i; k<numSections; k++) // reset values that follow
                                    {
                                        values[k] = null;
                                    }
                                }
                                values[j] = s;
                            }
                        }
                        if (newSection)
                        {
                            newSection = false;             // force it to be turned back on again.
                            delim = "";
                            for (String value :values)
                            {
                                vid.append(String.format("%s%s", delim, value));
                                delim=".";
                            }
                        }
                        text = String.format("%s\n", re.getGroup(gc));             // this is what follows the :
                    }
                    else if (line.indexOf(": ") != -1)
                        text = String.format("%s\n", line.split(": ")[1]);
                    else if (line.trim().length()>0)
                        text = "??\n";
                    else
                        text = null;        // empty line
                    lineno++;
                    if (text != null)
                    {
                        String segtype = null;
                        segre.reset(line);
                        if (segre.find())
                            segtype = segre.getGroup(1);
                        writeText(vid.toString(), text, segtype, line.charAt(0)=='+');           //!! Need to pass seg information
                        vid.setLength(0);
                    }
                }

                br.close();

                closePreviousElements(bwIndex, "", 1, numSections);

                bwIndex.write("\t\t]\n");
                bwIndex.write("\t}\n");
                bwIndex.write("];\n");
                bwIndex.write("</script>\n\n");
                bwIndex.close();

                for (int i=0; i<bi.length; i++)
                {
                    bi[i].close();
                }
            }
        }
    }


    private void writeText(String vid, String text, String segtype, boolean isSpan) throws Exception
    {
        int len = bi.length;
        for (int i=0; i<len; i++)
        {
            bi[i].write(vid, text, segtype, isSpan);
        }

        if (vid.length() > 0)
            verseNumber++;
    }


    void closePreviousElements(BufferedWriter bw, String v, int i, int numSections) throws Exception
    {
        if (v != null)      // close off previous elements
        {
            // finish last section
            boolean sequential = true;
            int last = -1;
            if (numSections > 1)
            {
                for (String lv : lastSectionValues)
                {
                    intcheck.reset(lv);
                    if (intcheck.find())          // if value is an integer
                    {
                        int n = Integer.parseInt(lv);
                        if (last == -1)
                            last = n;
                        else if (n == last + 1)     // is it one greater than the previous value?
                            last = n;
                        else
                            sequential = false;     // if not then values are not sequentially and will need an array to represent them.
                    }
                    else
                        sequential = false;

                    if (!sequential)
                        break;
                }
            }
            else
                sequential = false;             // only one section so list possibilities
            
            if (!sequential)        // use an array to hold values
            {
                if (numSections > 1)
                {
                    write(bw, marginSize[numSections-1] + 1, String.format("o:%d, s:\n", offset));
                    write(bw, marginSize[numSections-1] + 1, "[\n");
                }

                String delim = generateTabs(marginSize[numSections]);
                for (String lv : lastSectionValues)
                {
                    write(bw, 0, String.format("%s{", delim));
                    delim = ",\n" + generateTabs(marginSize[numSections]);
                    write(bw, 0, String.format("v:'%s'", lv));
                    write(bw, 0, "}");
                }
                write(bw, 0, "\n");
            }
            else
            {
                if (numSections == 0 || lastSectionValues.size() == 0)
                    numSections = 0;            // for debugging purposes, should never reach here
                else if (numSections > 1)
                    write(bw, marginSize[numSections-1]+1, String.format("o:%d, f:%s, n:%d\n", offset, lastSectionValues.get(0), lastSectionValues.size()));
                else
                    write(bw, marginSize[numSections-1]+1, String.format("{o:%d, f:%s, n:%d}\n", offset, lastSectionValues.get(0), lastSectionValues.size()));
            }
            offset += lastSectionValues.size();
            lastSectionValues.clear();

            for (int k = numSections-1; k >= i; k--)
            {
                if (!sequential || k != numSections-1)
                    write(bw, marginSize[k] + 1, "]\n");
                write(bw, marginSize[k], "}");              // don't write \n in case a comma needs to be written
                if (v.length() == 0)                        // if this is final call
                    write(bw, 0, "\n");                     // write the \n because we know that nothing will follow
            }
        }
    }

    void write(BufferedWriter bw, int indent, String s) throws Exception
    {
        bw.write(generateTabs(indent));
        bw.write(s);
    }

    String generateTabs(int n)
    {
        return tabString.substring(0, n);
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        if (qName.equals("text"))
        {
            currentFile = atts.getValue("name");
            if (debug && atts != null)
            {
                if (atts.getValue("debug").equals("n"))
                    currentFile = null;
            }
            if (currentFile != null)
                h.put(currentFile, new DocumentInfo(currentFile));
        }
        else if (qName.equals("section"))
        {
            if (currentFile != null)
            {
                String display = null;
                DocumentInfo di = h.get(currentFile);
                try
                {
                    display = WebServices.transformString(atts.getValue("display"), tmRoman);
                }
                catch (Exception ex)
                {
                    throw new SAXException(ex.getMessage());
                }
                String value = atts.getValue("value");
                boolean accumulated = atts.getValue("cum") != null;
                di.sections.add(new SectionInfo(display, value, accumulated));
            }
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
    }

    // For ErrorHandler
    @Override
    public void warning(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void error(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException
    {
        throw exception;
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
    }

    @Override
    public void endDocument() throws SAXException
    {
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    private class BufferInfo
    {
        private boolean endparenPending = false;
        private StringBuilder sb = new StringBuilder();
        private String outputdir;
        private String fn;
        private String ext;
        private int sofar;
        private int pagenum;
        private int versecnt;       // keeps track of number of verses in a page
        private String mapdelim;

        private BufferedWriter bwText;
        private BufferedWriter bwMap;
        private TransformMap tm;

        public BufferInfo(String outputdir, String fn, String ext) throws Exception
        {
            this.outputdir = outputdir;
            this.fn = fn;
            this.ext = ext;
            sofar = 0;
            pagenum = 1;
            versecnt = 0;
            mapdelim = "[";
            bwText = IOAssistance.openWriter(String.format("%s/%s_page%d.%s", outputdir, fn, pagenum, ext));
            bwMap = IOAssistance.openWriter(String.format("%s/%s_map.%s", outputdir, fn, ext));
           
            if (!ext.equals("slp1"))
            {
                String map = String.format("*:slp1->%s", ext);
                tm = new TransformMap(map, ris, fsmh);
            }
        }

        public void write(String vid, String text, String segtype, boolean isSpan) throws Exception
        {
            boolean isVerse = vid.length() > 0;

            if (ext.equals("deva"))
                debugging = true;

            if (debugging)
                debugging = true;

            if (!isSpan && endparenPending)
            {
                _write("</p>\n");
                endparenPending = false;
            }


            if (isVerse && sofar > pageBufferSize)
                dumpPage(text);

            boolean isContinuous = segtype != null && segtype.charAt(1) == 'c';
            String tagtype = "p";
            if (isSpan)
                tagtype= "span";

            String id = "";
            if (isVerse)
                versecnt++;

            if (ext.equals("deva") && segtype != null && segtype.charAt(1)=='c')        // squeeze blanks out of continuous devanagari text
                text = deromanize(text, sb);
            String klass = "";
            if (segtype != null)
                klass = String.format(" class='%s'", segtype);
            else
                klass = " class='eng'";

            if (isVerse)
                _write(String.format("<p id='%d' class='verseid'>%s</p>", verseNumber, vid.toString()));

            _write(String.format("<%s%s%s>", tagtype, id, klass));
            if (isContinuous)
                _write("<a class='L'>");
            String[] words = text.split("\\s+");
            String blankDelim = "";
            for (String word : words)
            {
                wordre.reset(word);
                boolean isWord = wordre.find();        // word contains at least one letter

                if (isWord)
                {
                    ArrayList<SLP1Partition> a = FSM.partition(word, segtype != null, sb, FSM.slpCharacter);
                    _write(blankDelim);
                    for (SLP1Partition sp : a)
                    {
                        String part = sp.part;
                        if (segtype != null && tm != null)
                            part = WebServices.transformString(part, tm);
                        if (sp.isSLP1)
                        {
                            if (isContinuous)
                                _write(String.format("%s ", part));
                            else
                                _write(String.format("<a>%s</a>", part));
                        }
                        else
                            _write(part);
                    }
                }
                else
                {
                    if (segtype != null && tm != null)      // e.g., convert digits to devanagari
                        word = WebServices.transformString(word, tm);
                    _write(String.format("%s%s", blankDelim, word));
                }
                blankDelim = " ";
            }
            if (isContinuous)
                _write("</a>");
            // close off element, waiting if it is a p
            if (isSpan)
                _write(String.format("</%s> ", tagtype));
            else
                endparenPending = true;     // delay writing </p> until the next <p>
        }


        private String deromanize(String text, StringBuilder sb)
        {
            sb.setLength(0);
            int len = text.length();
            char nextch = text.charAt(0);
            for (int i=1; i<len; i++)
            {
                if (nextch == '\0')             // char i is a space that needs to be skipped.
                {
                    for (int j=i; j<len; j++)   // find next non-blank
                    {
                        nextch = text.charAt(j);
                        if (nextch != ' ')
                            break;
                    }
                }
                else
                {
                    char ch = nextch;
                    sb.append(ch);
                    nextch = text.charAt(i);
                    if (nextch == ' ' && FSM.slpVowel.get(ch) == null)      // if next char is blank and this char is not a vowel
                        nextch = '\0';                              // skip the next space
                }
            }

            if (nextch != ' ')
                sb.append(nextch);

            return sb.toString();
        }

        private void _write(String s) throws Exception
        {
            bwText.write(s);
            sofar += s.length();
        }

        private void _write(char ch) throws Exception
        {
            bwText.write(ch);
            sofar++;
        }

        private void dumpPage(String text) throws Exception
        {
            bwMap.write(String.format("%s%d", mapdelim, versecnt));
            mapdelim = ",";
            pagenum++;
            bwText.close();
            if (text != null)
            {
                String path = String.format("%s/%s_page%d.%s", outputdir, fn, pagenum, ext);
                bwText = IOAssistance.openWriter(path);
                versecnt = 0;
                sofar = 0;
            }
        }

        public void close() throws Exception
        {
            dumpPage(null);
            bwMap.write("]");
            bwMap.close();
        }
    }
}
