package org.sanskritlibrary;

import java.io.File;
import java.util.Hashtable;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;

public class SubjectClassifications extends DefaultHandler
{
    static TransformMap tmSLP1toRoman = null;
    Hashtable<String, FSM[]>fsmh;
    ResourceInputStream ris;
    static int elementNumber = 0;
    String currentID = null;
    boolean inCatDesc = false;
    boolean inForeign = false;
    String catlang = null;
    String foreignlang = null;
    StringBuilder catDescBuffer = new StringBuilder();
    StringBuilder foreignBuffer = new StringBuilder();
    Locator locator;
    File filename;
    Hashtable<String, String> h = new Hashtable<String, String>();

    public SubjectClassifications(File filename, Hashtable<String, FSM[]>fsmh, ResourceInputStream ris) throws Exception
    {
        this.filename = filename;
        this.fsmh = fsmh;
        this.ris = ris;
        tmSLP1toRoman = new TransformMap("*:slp1->roman", ris, fsmh);
    }

    public Hashtable<String, String> getClassifications()
    {
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            saxParser.parse( filename, this);
        }
        catch (Throwable t)
        {
            t.printStackTrace ();
        }

        return h;
    }

    // For ErrorHandler
    @Override
    public void warning(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void error(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
    }

    @Override
    public void endDocument() throws SAXException
    {
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        elementNumber++;
        
        if (qName.equals("category"))
        {
            currentID = atts.getValue("xml:id");
        }
        else if (qName.equals("catDesc"))
        {
            catlang = atts.getValue("xml:lang");
            if (catlang != null)
                inCatDesc=true;
            inCatDesc = true;
        }
        else if (inCatDesc && qName.equals("foreign"))
        {
            foreignlang = atts.getValue("xml:lang");
            inForeign = true;
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (qName.equals("catDesc"))
        {
            if (currentID != null)
            {
                String transcodedText = transcodeText(catDescBuffer.toString(), catlang);   // assume no foreign sub elements
                h.put(currentID, transcodedText);
            }
            inCatDesc = false;
            catlang = null;
            catDescBuffer.setLength(0);
        }
        else if (qName.equals("foreign"))
        {
            if (currentID != null && inCatDesc)
            {
                String transcodedText = transcodeText(foreignBuffer.toString(), foreignlang);
                catDescBuffer.append(transcodedText);
            }
            inForeign = false;
            foreignlang = null;
            foreignBuffer.setLength(0);
        }
    }

    String transcodeText(String s, String lang) throws SAXException
    {
        try
        {
            if (lang != null && lang.equals("sa-Latn-x-SLP1"))
            {
                s = WebServices.transformString(s, tmSLP1toRoman);
                String pattern = "\\[(.)\\]";
                s = s.replaceAll(pattern, "$1");
            }
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }

        return s;
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        if (inCatDesc)
        {
            if (currentID == null)
                return;
            String content = new String(ch, start, length);
            if (inForeign)
                foreignBuffer.append(content);
            else if (inCatDesc)
            {
                catDescBuffer.append(content);
            }
        }
    }

}
