package org.sanskritlibrary;
import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Enumeration;

public class SymbolTable
{
	private FSMEntry[] fsmEntry;
    private Hashtable<String, FSMEntry> fsmHash = null;
    private Hashtable<String, Hashtable<String, Boolean>> classCollections = new Hashtable<String, Hashtable<String, Boolean>>();
    private Hashtable<String, Integer> maxKeyLengths = new Hashtable<String, Integer>();

	public SymbolTable()
	{
		
	}
	
	public SymbolTable(FSMEntry[] fsmEntry) throws Exception
	{
		this.fsmEntry = fsmEntry;
	}


    public void validateClassNames(Buffers buffers) throws Exception
    {
        for (FSMEntry fe : fsmEntry)
        {
            if (fe.klass == null)
            {
                if (fe.key == null)
                    throw new Exception(String.format("%s has neither klass or key defined", fe.id));
            }
            else if (buffers.getMember(fe.klass) == null)
                throw new Exception(String.format("%s not a known member", fe.klass));
        }
    }


    public Hashtable<String, Boolean> getClassCollection(String className) throws Exception
    {
        // classname may contain class names, keys and literals
        Hashtable<String, Boolean> h = classCollections.get(className);
        if (h == null)
        {
            h = new Hashtable<String, Boolean>();
            Regex re = new Regex(className, "(\\$[a-zA-Z_0-9]+)|(%[a-zA-Z_0-9.]+)|(\\\\u\\d+)");
            while (re.find())
            {
                String s = re.getGroup(1);
                if (s.startsWith("$"))
                    getClassMembers(s.substring(1), h);
                else if (s.startsWith("%"))
                    getEntryMembers(s.substring(1), h);
                else
                    h.put(TextFormatter.slashUToChar(s), true);
            }

            classCollections.put(className, h);

            int maxKeyLength = 0;
            Enumeration<String> e = h.keys();
            while(e.hasMoreElements())
            {
                int len = e.nextElement().length();
                if (len > maxKeyLength)
                    maxKeyLength = len;
            }
            maxKeyLengths.put(className, maxKeyLength);
        }

        return h;
    }


    private void getClassMembers(String klass, Hashtable<String, Boolean> h) throws Exception
    {
        ArrayList<String> a = new ArrayList<String>();

        boolean found = false;
        for (FSMEntry fe : fsmEntry)
        {
            if (fe.klass.equals(klass))
            {
                found = true;
                addEdges(fe, h);
            }
        }

        if (!found)
            throw new Exception();
    }


    private void getEntryMembers(String key, Hashtable<String, Boolean> h) throws Exception
    {
        ArrayList<String> a = new ArrayList<String>();
        FSMEntry fe = getFSMEntry(key);
        if (fe == null)
            throw new Exception(String.format("key %s referenced in match item not found", key));

        addEdges(fe, h);
    }


    private void addEdges(FSMEntry fe, Hashtable<String, Boolean> h)
    {
        if (fe.edges != null)
        {
            for (String edge : fe.edges)
            {
                h.put(edge, Boolean.TRUE);
            }
        }
        else
            h.put(fe.edge, true);

    }


    FSMEntry getFSMEntry(String name)
    {
        if (fsmHash == null)
        {
            fsmHash = new Hashtable<String, FSMEntry>();
            for (FSMEntry fe : fsmEntry)
            {
                if (fe.key != null)
                    fsmHash.put(fe.key, fe);
            }
        }

        return fsmHash.get(name);
    }


    public int getMaxKeyLength(String className)
    {
        return maxKeyLengths.get(className);
    }
}
