package org.sanskritlibrary;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;

public class TEICatalog extends DefaultHandler
{
    static String headerTemplate =
    "<%@page contentType='text/html' pageEncoding='UTF-8'%>\n" +
    "<%@include file='../WEB-INF/jspf/Head.jspf'%>\n" +
    "<%@include file='../WEB-INF/jspf/BreadCrumbs.jspf'%>\n" +
    "<%@include file='../WEB-INF/jspf/Menu.jspf'%>\n" +
    "<link rel='stylesheet' type='text/css' href='css/bib.css'/>\n" +
    "<script type='text/javascript'>\n" +
	"function doSubmit(n)\n" +
	"{\n" +
	"\t$(\"#format\").val(n);\n" +
	"\t$(\"#buttons\").submit();\n" +
	"}\n" +
    "</script>\n" +
    "<div class='contentEnglish'>\n" +
    "<div class='text'>\n";

    
    static String[] headerButtonsTemplate = new String[]
    {
    "<input class='TXTmenuButton' style='display:%s' value='Text' onclick='doSubmit(%d)' type='button'/>\n",
    "<input class='TXTmenuButton' style='display:%s' value='Sandhi-analyzed Text' onclick='doSubmit(%d)' type='button'/>\n",
    "<input class='TXTmenuButton' style='display:%s' value='Kramapāṭha Reader' onclick='doSubmit(%d)' type='button'/>\n",
    };
    

    static String dataTemplate =
    "<div class='BIBentry'>\n" +
    "\t<div class='BIBname'>%s</div>\n" +
    "\t<div class='BIBvalue'>%s</div>\n" +
    "</div>\n";

    static String trailerTemplate = "<%@include file='../WEB-INF/jspf/BodyEnd.jspf'%>";
    static String linkTemplate = "<li class='spaced'><a href='LoadPage?name=Catalogs/%s&context=Texts&title=%s'>%s</a></li>\n";
    static int elementNumber = 0;
    static Integer zero = new Integer(0);
    Locator locator;
    String filename;
    String fileid;
    StringBuilder sb = new StringBuilder();
    Hashtable<String, String> h;
    ArrayList<String> currentTypePath = new ArrayList<String>();
    ArrayList<String> elements = new ArrayList<String>();
    ArrayList<ElementInfo> elementInfo = new ArrayList<ElementInfo>();
    ArrayList<String> div1 = new ArrayList<String>();
    ArrayList<String> lastVerse = new ArrayList<String>();
    ArrayList<Integer> elementIndexes = new ArrayList<Integer>();
    String currentTextType;
    Hashtable<String, FSM[]>fsmh;
    ResourceInputStream ris;
    TransformMap tmSLP1toRoman ;
    Hashtable<String, ArrayList<ElementInfo>> buffer = new Hashtable<String, ArrayList<ElementInfo>>();
    Hashtable<String, String> nameExtensions = new Hashtable<String, String>();
    Hashtable<String, Integer> groupExtensions = new Hashtable<String, Integer>();
    static Hashtable<String, ArrayList<String>> hSegType = new Hashtable<String, ArrayList<String>>();
    StringBuilder foreignText = new StringBuilder();
    String copyElement;
    String lv = null;
    StringBuilder debugBuffer = new StringBuilder();
    StringBuilder keybfr = new StringBuilder();
    StringBuilder accentbfr = new StringBuilder();
    boolean doTranscoding = false;
    String title = null;
    int parseType = 0;
    boolean startFlattening = false;
    String titusLink = null;
    private int segCount = 0;
    private String segtype = null;
    private boolean processWords = false;
    private StringBuilder wordBuffer = new StringBuilder();
    StringBuilder flattenContent = new StringBuilder();
    Hashtable<String, ArrayList<FreqInfo>> analysis = new Hashtable<String, ArrayList<FreqInfo>>();
    static Hashtable<Character, Boolean> mystery = new Hashtable<Character, Boolean>();
    private Regex wordre =  new Regex("","[a-zA-Z|]");
    private Hashtable<String, WordInfo> wordh;
    public static Hashtable<String, Integer> missingAna = new Hashtable<String, Integer>();
    public static Hashtable<String, ArrayList<String>> types = new Hashtable<String, ArrayList<String>>();
    public static Hashtable<String, ArrayList<String>> paths = new Hashtable<String, ArrayList<String>>();


    public TEICatalog(String filename, Hashtable<String, String> h, Hashtable<String, FSM[]> fsmh, ResourceInputStream ris)
            throws Exception
    {
        //System.out.println(filename);
        this.filename = filename;
        this.h = h;
        this.fsmh = fsmh;
        this.ris = ris;
        tmSLP1toRoman = new TransformMap("*:slp1->roman", ris, fsmh);
        buffer.put("*.foreign", new ArrayList<ElementInfo>());
        buffer.put("*.ptr", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.titleStmt.title[main]", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.titleStmt.title[sub]", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.titleStmt.title[desc]", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.titleStmt.author", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.titleStmt.editor", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.titleStmt.series", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.titleStmt.series.biblScope", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.titleStmt.respStmt.name", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.titleStmt.respStmt.resp", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.editionStmt.edition", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.publicationStmt.publisher", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.publicationStmt.pubPlace.address.addrLine", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.publicationStmt.date", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.publicationStmt.availability.p", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.sourceDesc", new ArrayList<ElementInfo>());          // to count # of sourceDescs
        buffer.put("TEI.teiHeader.fileDesc.sourceDesc.bibl.title[main]", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.sourceDesc.bibl.title[desc]", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.sourceDesc.bibl.title[sub]", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.sourceDesc.bibl.editor", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.sourceDesc.bibl.pubPlace", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.sourceDesc.bibl.publisher", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.sourceDesc.bibl.date", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.sourceDesc.bibl.ptr", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.sourceDesc.bibl.series", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.fileDesc.sourceDesc.bibl.series.biblScope", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.profileDesc.textClass.keywords.list.item", new ArrayList<ElementInfo>());
        buffer.put("TEI.teiHeader.profileDesc.textClass.catRef", new ArrayList<ElementInfo>());

        nameExtensions.put("TEI.teiHeader.fileDesc.sourceDesc.bibl.title", "type");
        nameExtensions.put("TEI.teiHeader.fileDesc.titleStmt.title", "type");

        groupExtensions.put("TEI.teiHeader.fileDesc.sourceDesc", new Integer(0));

        Regex re = new Regex(filename, "\\\\([^\\\\]+)\\.xml$");
        if (re.find())
            this.fileid = re.getGroup(1);

    }


    public static void createStructure(String directory, String outfile) throws Exception
    {
        BufferedWriter bw = IOAssistance.openWriter(outfile);
        bw.write("<texts>\n");
        File dir = new File(directory);
        File[] files = dir.listFiles();
        ArrayList<String> a = new ArrayList<String>();
        Regex re = new Regex("", "([^\\[]+)\\[");
        for (File fn : files)
        {
            if (!fn.getName().endsWith(".txt"))
                continue;
            a.clear();
            bw.write(String.format("\t<text name='%s'>\n", fn.getName()));
            BufferedReader br = IOAssistance.openReader(fn.getAbsolutePath());
            String line = null;
            while ((line = br.readLine()) != null)
            {
                if (line.length() == 0 || line.charAt(0) == '*')
                    continue;

                if (line.indexOf(":") == -1)
                    continue;

                String path = line.split(":")[0];
                String[] sections = path.split("\\.");
                for (String section : sections)
                {
                    re.reset(section);
                    if (re.find())
                    {
                        String s = re.getGroup(1);
                        if (!s.equals("seg") && !a.contains(s))
                            a.add(s);
                    }
                }
            }
            for (String s : a)
            {
                bw.write(String.format("\t\t<section display='%s' value='%s'/>\n", s, s));
            }
            bw.write(String.format("\t</text>\n", fn.getName()));
        }
        bw.write("</texts>");
        bw.close();
    }

    public String extractInfo() throws Throwable
    {
        parseType = 1;
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            saxParser.parse( new File(filename), this);
        }
        catch (Throwable t)
        {
            throw t;
        }

        return sb.toString();
    }

    public String dataSummary() throws Throwable
    {
        h = new Hashtable<String, String>();
        parseType = 2;
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            saxParser.parse( new File(filename), this);
        }
        catch (Throwable t)
        {
            throw t;
        }

        Enumeration<String> e = analysis.keys();
        ArrayList<String> helper = new ArrayList<String>();
        while (e.hasMoreElements())
        {
            helper.add(e.nextElement());
        }
        Collections.sort(helper);

        for (String key : helper)
        {
            sb.append("\t");
            sb.append(key);
            sb.append(":\n");
            ArrayList<FreqInfo> anas = analysis.get(key);
            for (FreqInfo ana : anas)
            {
                if (ana.s.equals("[?]"))
                    sb.append(String.format("\t\t%s (%d) first location: %d\n", ana.s, ana.count, ana.location));
                else
                    sb.append(String.format("\t\t%s (%d)\n", ana.s, ana.count));
            }
        }
        return sb.toString();
    }

    public String flatten() throws Throwable
    {
        parseType = 3;
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            saxParser.parse( new File(filename), this);
        }
        catch (Throwable t)
        {
            throw t;
        }

        sb.append("\n\n");
        return sb.toString();
    }


    public String createWordList(Hashtable<String, WordInfo> h, String segtype) throws Throwable
    {
        this.segtype = segtype;
        this.wordh = h;
        parseType = 4;
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            saxParser.parse( new File(filename), this);
        }
        catch (Throwable t)
        {
            throw t;
        }

        sb.append("\n\n");
        return sb.toString();
    }



    public String getLink()
    {
        String jspname = filename;
        Regex re = new Regex(jspname, "\\\\([^\\\\]+)\\.xml$");
        if (re.find())
            jspname = re.getGroup(1);
        return String.format(linkTemplate, jspname, title, title);
    }

    // This will be modified by human
    public String getTextInfo(boolean auto)
    {
        StringBuilder sbb = new StringBuilder();
        Regex re = new Regex(filename, "\\\\([^\\\\]+)\\.xml$");
        if (re.find())
            filename = re.getGroup(1);
        sbb.append("\t<texts>\n");
        sbb.append(String.format("\t\t<name>%s</name>\n", title));
        sbb.append("\t\t<structure>\n");
        if (!auto)
            generateBookElement(sbb);
        else
            generateBookElementEx(sbb);
        sbb.append("\t\t</structure>\n");
        sbb.append(String.format("\t\t<fileName>texts/%s</fileName>\n", filename));
		sbb.append("\t\t<versePattern>\n");
		sbb.append("\t\t\t<regexp>^\\(?(\\p{N}+)[.\\u0964](\\p{N}+)[.\\u0964](\\p{N}+)[^\\s]+\\s+([^\t]*\t[^\t]*)</regexp>\n");
		sbb.append("\t\t\t<book></book>\n");
        sbb.append("\t\t\t<chapter>2</chapter>\n");
		sbb.append("\t\t\t<verse>3</verse>\n");
		sbb.append("\t\t\t<text>4</text>\n");
		sbb.append("\t\t</versePattern>\n");
		sbb.append("\t\t<originalEncoding>\n");
		sbb.append("\t\t\t<encoding>slp1</encoding>\n");
		sbb.append("\t\t</originalEncoding>\n");
        sbb.append("\t<\\texts>\n\n");
        return sbb.toString();
    }


    public static String getMissingAna()
    {
        StringBuilder sbb = new StringBuilder();
        ArrayList<String> a = new ArrayList<String>();
        Enumeration<String> en = missingAna.keys();
        while (en.hasMoreElements())
        {
            a.add(en.nextElement());
        }
        Collections.sort(a);
        for (String s : a)
        {
            sbb.append(String.format("%s %d\n", s, missingAna.get(s).intValue()));
        }

        if (sbb.length() == 0)
            sbb.append("No missing ana\n");

        return sbb.toString();
    }


    public static String getTypes()
    {
        StringBuilder sbb = new StringBuilder();
        Enumeration<String> en = types.keys();
        ArrayList<String> a = new ArrayList<String>();

        while (en.hasMoreElements())
        {
            a.add(en.nextElement());
        }
        Collections.sort(a);

        for (String type : a)
        {
            ArrayList<String> aa = types.get(type);
            Collections.sort(aa);
            String delim = "";
            sbb.append(String.format("%s:\n", type));
            for (String s: aa)
            {
                sbb.append(String.format("%s%s", delim, s));
                delim = ", ";
            }
            sbb.append("\n\n");
        }

        en = paths.keys();
        a.clear();
        while (en.hasMoreElements())
        {
            a.add(en.nextElement());
        }

        Collections.sort(a);

        for (String path : a)
        {
            ArrayList<String> aa = paths.get(path);
            Collections.sort(aa);
            sbb.append(String.format("%s:\n", path));
            for (String s: aa)
            {
                sbb.append(String.format("\t%s\n", s));
            }
            sbb.append("\n");
        }

        return sbb.toString();
    }


    public static String getSegTypes()
    {
        StringBuilder sbb = new StringBuilder();
        sbb.append("\nSegment types\n");
        Enumeration<String> en = hSegType.keys();
        ArrayList<String> a = new ArrayList<String>();
        while (en.hasMoreElements())
        {
            a.add(en.nextElement());
        }

        Collections.sort(a);

        for (String key : a)
        {
            ArrayList<String> aa = hSegType.get(key);
            Collections.sort(aa);
            sbb.append(String.format("%s:\n", key));
            int n = 0;
            String delim = "\t";
            for (String fn : aa)
            {
                sbb.append(String.format("%s%s", delim, fn));
                delim = ", ";
                n++;
                if (n == 10)
                {
                    delim = "\t";
                    sbb.append("\n");
                    n = 0;
                }
            }
            if (n != 0)
                sbb.append("\n");
        }
        
        return sbb.toString();
    }

    private void generateBookElement(StringBuilder sbb)
    {
        sbb.append("\t\t\t<book number=''>\n");
        for (int i=0; i<4; i++)
        {
            sbb.append("\t\t\t\t<chapter number='' numVerses=''/>\n");
        }
        sbb.append("\t\t\t</book>\n");
    }

    private void generateBookElementEx(StringBuilder sbb)
    {
        int n = 0;
        boolean error = false;
        int len = div1.size();
        for (int i=0; i<len; i++)
        {
            String s = div1.get(i);
            if (n < lastVerse.size())
            {
                String vn = lastVerse.get(n++);
                while(vn.equals("?") && n < len)
                {
                    vn = lastVerse.get(n++);
                }
                i = n - 1;      // in case some question marks were skipped
                sbb.append(String.format("\t\t\t\t<chapter number='%s' numVerses='%s'/>\n", s, vn));
            }
            else
            {
                error = true;
                sbb.append(String.format("\t\t\t\t<chapter number='%s' numVerses='?'/>\n", s));
            }
        }

        if (error)
            System.out.format("%s has mismatched verse counts\n", filename);
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        elementNumber++;

        if (elementNumber == 131)
            elementNumber = 131;

        pushName(qName, atts);

        switch (parseType)
        {
            case 1:
                startElementTEIHeader(uri, localName, qName, atts);
                break;
            case 2:
                startElementData(uri, localName, qName, atts);
                break;

            case 3:
                startElementFlatten(uri, localName, qName, atts);
                break;

            case 4:
                startElementWordList(uri, localName, qName, atts);
                break;
        }
    }

    public void startElementTEIHeader(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        if (copyElement != null)
        {
            copyElementTag(qName, atts, "");
        }
        else
        {
            ArrayList<ElementInfo> eis = null;

            eis = buffer.get(buildIndexedKey());
            if (eis != null)
            {
                ElementInfo ei = new ElementInfo(elementNumber);
                if (atts.getLength() > 0)
                    ei.copyAttributes(atts);
                eis.add(ei);
            }
        }
    }

    public void startElementData(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        if (checkPathPrefix("TEI", "text", "body"))
        {
            if (qName.startsWith("div"))
            {
                String type = atts.getValue("type");
                if (type == null)
                    type = "?";
                currentTypePath.add(type);
            }
            else if (qName.equals("seg"))
            {
                if (elements.get(elements.size()-2).equals("p"))
                {
                    String key = concat(currentTypePath);
                    ArrayList<FreqInfo> anas = analysis.get(key);
                    if (anas == null)
                    {
                        anas = new ArrayList<FreqInfo>();
                        analysis.put(key, anas);
                    }

                    String ana = atts.getValue("ana");
                    if (ana != null)
                    {
                        String[] fs = ana.split(" ");
                        StringBuilder sbb = new StringBuilder();
                        sbb.append("[");
                        String delim = "";
                        for (String s : fs)
                        {
                            sbb.append(delim);
                            sbb.append(s.trim());
                            delim = ", ";
                        }
                        sbb.append("]");
                        FreqInfo fi = getFreqInfo(anas, sbb.toString());
                        if (fi == null)
                            anas.add(new FreqInfo(sbb.toString(), 1, locator.getLineNumber()));
                        else
                            fi.count++;
                    }
                    else
                    {
                        FreqInfo fi = getFreqInfo(anas, "[?]");
                        if (fi ==  null)
                            anas.add(new FreqInfo("[?]", 1, locator.getLineNumber()));
                        else
                            fi.count++;
                    }
                }
                else
                    throw new SAXException("Unexpected seg element");
            }
        }
    }

    public void startElementFlatten(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        if (qName.equals("body"))
            startFlattening = true;
        else if (getPath().equals("TEI.teiHeader.fileDesc.sourceDesc.bibl.ptr"))
        {
            if (titusLink == null)
            {
                titusLink = atts.getValue("target");
                if (titusLink != null)
                    sb.append(String.format("%s\n", titusLink));
                else
                    titusLink = null;
            }
            else
                titusLink = null;
        }
        else if (startFlattening)
        {
            if (qName.equals("p"))
                segCount = 0;
            else if (qName.equals("seg"))
                segCount++;
            ElementInfo ei = new ElementInfo(qName);
            if (atts.getLength() > 0)
                ei.copyAttributes(atts);
            elementInfo.add(ei);
        }

    }

    public void startElementWordList(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        if (qName.equals("seg"))
        {
            String ana = atts.getValue("ana");
            if (ana != null && ana.indexOf(segtype) != -1 && ana.indexOf("#recons") == -1)
                processWords = true;
            else
                processWords = false;
        }
        else
            processWords = false;
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        switch (parseType)
        {
            case 1:
                endElementTEIHeader(uri, localName, qName);
                break;
            case 2:
                endElementData(uri, localName, qName);
                break;
            case 3:
                endElementFlatten(uri, localName, qName);
                break;
            case 4:
                endElementWordList(uri, localName, qName);
                break;
        }

        popName();
    }

    public void endElementTEIHeader(String uri, String localName, String qName) throws SAXException
    {
        if (copyElement != null)        // for proof reading purposes
        {
            debugBuffer.append(String.format("</%s>", qName));

            if (qName.equals(copyElement))
            {
                System.out.println(debugBuffer.toString());
                debugBuffer.setLength(0);
                copyElement = null;
            }
        }

        ElementInfo ei = getCurrentElement();

        if (ei != null)
        {
            if (qName.equals("foreign"))    // try to transcode and append to current buffer.
            {
                transcodeText(ei);
                String parentKey = buildParentKey();
                ArrayList<ElementInfo> eis = buffer.get(buildIndexedKey());
                ElementInfo parent = getLastElement(parentKey);
                parent.buffer.append(eis.get(0).buffer);
                eis.remove(0);
                return;
            }

            if (ei.getValue("xml:lang") != null)
                transcodeText(ei);

            if (qName.equals("ptr"))
            {
                String parentKey = buildParentKey();
                if (parentKey.endsWith(".p"))           // if <ptr> occurs in text then
                {
                    ArrayList<ElementInfo> eis = buffer.get(buildIndexedKey());
                    ElementInfo parent = getLastElement(parentKey);
                    if (eis.size() > 0)
                    {
                        String url = eis.get(0).atts.get("target");
                        if (url != null)
                            parent.buffer.append(String.format("<a target='copyright' href='%s'>%s</a>", url, url));
                        eis.remove(0);
                    }
                }
            }
        }
        else if (isTeiHeader(qName))
            dumpCatalogEntry();
    }

    public void endElementData(String uri, String localName, String qName) throws SAXException
    {
        if (checkPathPrefix("TEI", "text", "body"))
        {
            if (qName.startsWith("div"))
                currentTypePath.remove(currentTypePath.size()-1);
        }
    }

    public void endElementFlatten(String uri, String localName, String qName) throws SAXException
    {
        if (qName.equals("body"))
            startFlattening = false;

        if (startFlattening)
        {
            String content = flattenContent.toString().trim();
            if (content.length() > 0)
            {
                boolean isSLP1 = true;

                if (qName.equals("seg")) // ignore pb, biblScope, bibl and hi
                {
                    StringBuilder path = new StringBuilder();
                    String pathDelim = "";
                    StringBuilder sbb = new StringBuilder();
                    String ignore = "";
                    String delim = "";
                    boolean haveDiv1 = true;

                    if (segCount > 1)
                        sbb.append("+");
                    // Build the signature of the seg
                    for (ElementInfo ei : elementInfo)
                    {
                        if (ei.qName.startsWith("div"))             // signature contribution includes type and n
                        {
                            if (!ei.qName.startsWith("div1"))
                                haveDiv1 = false;

                            String type = ei.atts.get("type");
                            if (type == null)
                                type = ei.qName + "?";

                            String n = ei.atts.get("n");
                            if (n == null)
                                n = "";
                            else
                                n = String.format("[%s]", n);

                            sbb.append(String.format("%s%s%s", delim, type, n));        // append piece of the signature

                            // for getTextInfo
                            ArrayList<String> a = types.get(ei.qName);
                            if (a == null)
                            {
                                a = new ArrayList<String>();
                                types.put(ei.qName, a);
                            }

                            if (!a.contains(type))
                                a.add(type);

                            path.append(String.format("%s%s", pathDelim, type));
                            pathDelim = ".";
                        }
                        else if (ei.qName.equals("seg"))      // signature contribution include ana
                        {
                            String ana = "";
                            if (filename.endsWith("ab.xml"))
                            {
                                ana = "";
                            }

                            if (ei.atts != null)
                            {
                                ana = ei.atts.get("ana");
                                if (ana == null)
                                {
                                    ArrayList<String> a = hSegType.get("eng");
                                    if (a == null)
                                    {
                                        a = new ArrayList<String>();
                                        hSegType.put("eng", a);
                                    }
                                    if (!a.contains(fileid))
                                        a.add(fileid);
                                    ana = "";
                                }
                                else
                                {
                                    ana = normalizeAna(ana);
                                    ana = String.format("[%s]", ana);
                                    ArrayList<String> a = hSegType.get(ana);
                                    if (a == null)
                                    {
                                        a = new ArrayList<String>();
                                        hSegType.put(ana, a);
                                    }
                                    if (!a.contains(fileid))
                                        a.add(fileid);
                                }
                                String lang = ei.atts.get("xml:lang");
                                isSLP1 = lang != null && lang.endsWith("-SLP1");
                            }

                            if (ana.length() == 0)              // is this always an error?
                            {
                                //ignore = "***";

                                // for getTextInfo
                                int n = 0;
                                if (missingAna.get(filename) != null)
                                    n = missingAna.get(filename);
                                n++;
                                missingAna.put(filename, n);
                            }

                            sbb.append(String.format("%s%s%s", delim, ei.qName, ana));
                        }
                        else        // something other than div or seg      // signature contribution includes name only
                            sbb.append(String.format("%s%s", delim, ei.qName));
                        delim = ".";
                    }

                    if (haveDiv1)        // i.e., div1 only, not div2, etc. for creation of texts.txt
                    {
                        div1.add(content);
                        lastVerse.add("?");
                    }

                    if (ignore.length() == 0)
                    {
                        lv = sbb.toString();
                        addLastVerse();
                        lv = null;
                    }

                    sb.append(String.format("%s", sbb.toString()));
                    sb.append(": ");
                    //if (!isSLP1)
                    //    sb.append("[");
                    sb.append(flattenContent.toString().trim());
                    //if (!isSLP1)
                    //    sb.append("]");
                    sb.append("\n");

                    // for getTextInfo
                    String spath = path.toString();
                    ArrayList<String> a = paths.get(spath);
                    if (a == null)
                    {
                        a = new ArrayList<String>();
                        paths.put(spath, a);
                    }
                    if (!a.contains(filename))
                        a.add(filename);
                }
            }
            int last = elementInfo.size() - 1;
            elementInfo.remove(last);
        }

        flattenContent.setLength(0);
    }

    public void endElementWordList(String uri, String localName, String qName) throws SAXException
    {
        if (qName.equals("seg") && processWords)
            doWords();
        processWords = false;
    }


    private void doWords()
    {
        String[] words = wordBuffer.toString().split("\\s+");
        wordBuffer.setLength(0);
        for (String word : words)
        {
            wordre.reset(word);
            boolean isWord = wordre.find();        // word contains at least one letter

            if (isWord)
            {
                ArrayList<SLP1Partition> a = FSM.partition(word, segtype != null, sb, FSM.slpCharacter);
                for (SLP1Partition sp : a)
                {
                    if (sp.isSLP1)
                    {
                        String part = sp.part;
                        WordInfo wi = wordh.get(part);
                        part = removeAccents(part);
                        boolean sandhiRemoved = segtype.equals("#trc");
                        if (wi == null)
                        {
                            wi = new WordInfo();
                            wordh.put(part, wi);
                            wi.sandhiRemoved = sandhiRemoved;
                        }
                        if (wi.sandhiRemoved == sandhiRemoved)  // don't bump counts of words in continuous text that have no sandhi
                        {
                            if (wi.workingfrequency == 0)
                                wi.numTexts++;
                            wi.workingfrequency++;
                            wi.totalfrequency++;
                        }
                    }
                }
            }
        }
    }

    private String removeAccents(String word)
    {
        accentbfr.setLength(0);
        int len = word.length();
        for (int i=0; i<len; i++)
        {
            char ch = word.charAt(i);
            if (ch == '|' || ch == '\'' || Character.isLetter(ch))
                accentbfr.append(ch);
            else if (mystery.get(ch) == null)
                mystery.put(ch, true);
        }

        return accentbfr.toString();
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        switch (parseType)
        {
            case 1:
                charactersTEIHeader(ch, start, length);
                break;
            case 2:
                charactersData(ch, start, length);
                break;
            case 3:
                charactersFlatten(ch, start, length);
                break;
            case 4:
                charactersWordList(ch, start, length);
                break;
        }

    }

    public void charactersTEIHeader(char[] ch, int start, int length) throws SAXException
    {
        String content = new String(ch, start, length);
        content = Regex.replace(content, "[\\n\\t]", " ");
        content = Regex.replace(content, " {2,}", " ");

        if (copyElement != null)            // for proof reading purposes
            debugBuffer.append(content);
        else
        {
            ElementInfo ei = getCurrentElement();
            if (ei != null)         // if this is an element we are interested in
                getCurrentElement().buffer.append(content);
        }
    }

    public void charactersData(char[] ch, int start, int length) throws SAXException
    {

    }

    public void charactersFlatten(char[] ch, int start, int length) throws SAXException
    {
        if (startFlattening)
        {
            String s = new String(ch, start, length);
            flattenContent.append(s.replaceAll("[\n\t]", " "));
        }
    }

    public void charactersWordList(char[] ch, int start, int length) throws SAXException
    {
        if (processWords)
            wordBuffer.append(new String(ch, start, length));
    }

    /*
     [mpc][csr_][a_][abcx_]

     mc__
     mc_[abcx]
     mca_
     mca[abcx]

     ms__
     ms_[abcx]
     msa_
     msa[abcx]

     cc__
     cs__

     */
    private String normalizeAna(String ana)
    {
        StringBuilder sbb = new StringBuilder();

        if (ana.indexOf("#metr") != -1)
            sbb.append("m");
        else if (ana.indexOf("#prose") != -1)
            sbb.append("p");
        else if (ana.indexOf("com") != -1)
            sbb.append("c");
        else
            sbb.append("?");

        if (ana.indexOf("#trl") != -1)
            sbb.append("c");
        else if (ana.indexOf("#trc") != -1)
            sbb.append("s");
        else if (ana.indexOf("#recon") != -1)
            sbb.append("r");
        else
            sbb.append("_");

        if (ana.indexOf("#acc") != -1)
            sbb.append("a");
        else
            sbb.append("_");

        int n = ana.indexOf("#var-");
        if (n != -1)
        {
            char ch = ana.charAt(n+5);
            sbb.append(ch);
        }
        else
            sbb.append("_");

        return sbb.toString();
    }


    private String concat(ArrayList<String> a)
    {
        StringBuilder sbb = new StringBuilder();
        String delim = "";
        for (String s : a)
        {
            sbb.append(delim);
            sbb.append(s);
            delim = ".";
        }

        return sbb.toString();
    }

    private void pushName(String qName, Attributes atts)
    {
        elements.add(nameFixup(qName, atts));       // add attribute extension
        String key = buildKey();                    // without any indexes, so we can look things up
        Integer cnt = groupExtensions.get(key);
        if (cnt != null)
        {
            cnt = new Integer(cnt.intValue() + 1);
            groupExtensions.put(key, cnt);
            elementIndexes.add(cnt);
            buffer.put(buildIndexedKey(), new ArrayList<ElementInfo>());
        }
        else
        {
            elementIndexes.add(zero);
            String indexedKey = buildIndexedKey();
            if (buffer.get(indexedKey)==null && buffer.get(key) != null)   // e.g, indexed keys or *. keys
                buffer.put(indexedKey, new ArrayList<ElementInfo>());
        }
    }

    private void popName()
    {
        int last = elements.size() - 1;
        elements.remove(last);
        elementIndexes.remove(last);
    }

    private String nameFixup(String qName, Attributes atts)
    {
        String key = buildKey();            // biuld without indexes so we can do lookup in nameExtensions;
        key = String.format("%s.%s", key, qName);
        String attName = nameExtensions.get(key);
        if (attName != null)
        {
            String value = atts.getValue(attName);
            if (value != null)
                return String.format("%s[%s]", qName, value);
            else
                return qName;
        }
        else
            return qName;
    }

    private void transcodeText(ElementInfo ei) throws SAXException
    {
        if (ei.atts != null)
        {
            if (ei.getValue("xml:lang").equals("sa-Latn-x-SLP1"))
            {
                try
                {
                    String transcodedText = WebServices.transformString(ei.buffer.toString(), tmSLP1toRoman);
                    ei.buffer.setLength(0);
                    String pattern = "\\[(.)\\]";
                    transcodedText = transcodedText.replaceAll(pattern, "$1");
                    ei.buffer.append(transcodedText);
                } catch (Exception ex)
                {
                    throw new SAXException(ex.getMessage());
                }
            }
            else
                ei.buffer.append(foreignText);
        }
        else
            ei.buffer.append(foreignText);
    }

    private void dumpCatalogEntry()
    {
        sb.append(headerTemplate);
        sb.append("<div class='TXTbuttons'>\n");
        sb.append("<form id='buttons' action='TextViewerEx' method='get'>\n");
        for (int i=0; i<headerButtonsTemplate.length; i++)
        {
            String template = headerButtonsTemplate[i];
            String displayType = hasFormat(i) ? "inline" : "none";
            sb.append(String.format(template, displayType, i));
        }
        sb.append(String.format("<input id='format' name='format' type='hidden' value=''/>\n"));
        sb.append(String.format("<input id='name' name='name' type='hidden' value='%s'/>\n", fileid));
        sb.append(String.format("<input id='title' name='title' type='hidden' value='%s'/>\n",
            buffer.get("TEI.teiHeader.fileDesc.titleStmt.title[main]").get(0).buffer.toString()));
        sb.append("<input id='texttype' name='texttype' type='hidden' value=${texttype}/>\n");
        sb.append("</form>\n</div>\n");

        dumpTitle();
        dumpPublication();
        dumpProfile();
        dumpSources();
        sb.append(trailerTemplate);
    }

    private boolean hasFormat(int n)
    {
        if (n == 0)
            return true;
        else
            return false;
    }

    private void dumpTitle()
    {
        processTitle("TEI.teiHeader.fileDesc.titleStmt.title");
        title = getBuffer("TEI.teiHeader.fileDesc.titleStmt.title[main]");
        String subtitle = getBuffer("TEI.teiHeader.fileDesc.titleStmt.title[sub]");
        if (subtitle.length() > 0)
            title = String.format("%s: %s", title, subtitle);

        dumpElement("TEI.teiHeader.fileDesc.titleStmt.author", "Author: ");
        dumpElement("TEI.teiHeader.fileDesc.titleStmt.editor", "Editor: ");

        ArrayList<ElementInfo> respNames = buffer.get("TEI.teiHeader.fileDesc.titleStmt.respStmt.name");
        ArrayList<ElementInfo> respResps = buffer.get("TEI.teiHeader.fileDesc.titleStmt.respStmt.resp");
        if (respNames.size() == respResps.size())       // !!TODO handle this the way that Sources are done.
        {
            int len = respNames.size();
            for (int i=0; i<len; i++)
            {
                ElementInfo eiName = respNames.get(i);
                ElementInfo eiResp = respResps.get(i);
                sb.append(String.format(dataTemplate, "Contributor: ", eiName.buffer.toString()));
                sb.append(String.format(dataTemplate, "Contributions: ", eiResp.buffer.toString()));
            }
        }
        else 
            System.out.println("names != resps");

        dumpElement("TEI.teiHeader.fileDesc.editionStmt.edition", "Version: ");
        String series = getSeries(-1);
        if (series.length() != 0)
            sb.append(String.format(dataTemplate, "Series: ", series));

    }

    private void dumpElement(String key, String intro)
    {
        ArrayList<ElementInfo>eis = buffer.get(key);
        String delim = "";
        StringBuilder sbb = new StringBuilder();
        for (ElementInfo ei : eis)
        {
            if (!isBufferEmpty(ei.buffer))
            {
                sbb.append(String.format("%s%s", delim, ei.buffer));
                delim = "; ";
            }
        }
        if (sbb.length() > 0)
            sb.append(String.format(dataTemplate, intro, sbb.toString()));
    }

    private void dumpPublication()
    {
       String publisher = getBuffer("TEI.teiHeader.fileDesc.publicationStmt.publisher");
       String address = getBuffer("TEI.teiHeader.fileDesc.publicationStmt.pubPlace.address.addrLine");
       String date = getBuffer("TEI.teiHeader.fileDesc.publicationStmt.date");

       Regex re = new Regex(date, "(\\d+)$");
       if (re.find())
           date = ", " + re.getGroup(1);

       String publication = String.format("%s: %s%s", address, publisher, date);
       sb.append(String.format(dataTemplate, "Publication: ", publication));

       ArrayList<ElementInfo> eis = buffer.get("TEI.teiHeader.fileDesc.publicationStmt.availability.p");
       for (ElementInfo ei : eis)
       {
           if (ei.buffer.lastIndexOf("COPYRIGHT") == -1)
               sb.append(String.format(dataTemplate, "Copyright: ", ei.buffer.toString()));
       }
    }

    private void dumpSources()
    {
        int i = 0;

        while (true)
        {
            int n = i+1;
            ArrayList<ElementInfo> eis = buffer.get(String.format("TEI.teiHeader.fileDesc.sourceDesc[%d]", n));
            if (eis == null)
                break;
            String sourceDesc = String.format("TEI.teiHeader.fileDesc.sourceDesc[%d].bibl", n);
            String sourceTitle = getBuffer(sourceDesc + ".title[main]", 0);
            String editor = getBuffer(sourceDesc + ".editor", 0);
            String pubPlace = getBuffer(sourceDesc + ".pubPlace", 0);
            String publisher = getBuffer(sourceDesc + ".publisher", 0);
            String date = getBuffer(sourceDesc + ".date", 0);
            eis = buffer.get(sourceDesc + ".ptr");
            String ptr = null;
            if (eis != null && eis.size()>0)
            {
                ElementInfo ei = eis.get(0);
                ptr = ei.atts.get("target");
            }

            String series = getSeries(n);
            String publication = "";

            boolean empty = true;

            if (sourceTitle.length() > 0 || editor.length() > 0)
                empty = false;
            if (pubPlace.length() > 0 || publisher.length() > 0 || date.length() > 0)
            {
                if (pubPlace.length() > 0)
                    pubPlace = pubPlace + ": ";
                if (publisher.length() > 0 && date.length() > 0)
                    date = ", " + date;
                publication = pubPlace + publisher + date;
                empty = false;
            }

            if (!empty)
            {
                sb.append("<div class='BIBsec'>Source</div>\n");
                processTitle(sourceDesc + ".title");
                sb.append(String.format(dataTemplate, "Editor: ", editor));
                sb.append(String.format(dataTemplate, "Publication: ", publication));
                if (series.length() > 0)
                    sb.append(String.format(dataTemplate, "Series: ", series));
                if (ptr!=null)
                {
                    String link = String.format("<a target='source' href='%s'>%s</a>", ptr, ptr);
                    sb.append(String.format(dataTemplate, "Web Page:", link));
                }
            }

            i++;
        }
    }


    private void processTitle(String key)
    {
        String theTitle = "";
        ElementInfo main = null;
        ElementInfo desc = null;
        ElementInfo sub = null;

        ArrayList<ElementInfo> titles = buffer.get(key + "[main]");
        if (titles == null || titles.size() == 0)
        {
            sb.append(String.format(dataTemplate, "Title: ", "No main title"));
            return;
        }
        else
            main = titles.get(0);

        ArrayList<ElementInfo> desctitle = buffer.get(key + "[desc]");
        if (desctitle != null && desctitle.size() > 0)
            desc = desctitle.get(0);

        ArrayList<ElementInfo> subtitle = buffer.get(key + "[sub]");
        if (subtitle != null && subtitle.size() > 0)
            sub = subtitle.get(0);

        if (main != null)
            theTitle = main.buffer.toString();
        else
            return;             // if no main title then desc and sub don't make sense.

        if (desc != null)
        {
            if (sub==null || desc.number < sub.number)
            {
                theTitle = String.format("%s: %s", theTitle, desc.buffer);
                if (desctitle.size() > 1)               // prepare for subtitle, if any
                    desc = desctitle.get(1);
                else
                    desc = null;
            }
        }

        sb.append(String.format(dataTemplate, "Title: ", theTitle));

        if (sub != null)
        {
            String st = sub.buffer.toString();
            if (desc != null)
                st = String.format("%s: %s", st, desc.buffer.toString());

            sb.append(String.format(dataTemplate, "Subtitle: ", st));
        }
    }

    private String getSeries(int n)
    {
        String series = null;
        String biblscope = null;
        if (n != -1)
        {
            series = getBuffer(String.format("TEI.teiHeader.fileDesc.sourceDesc[%d].bibl.series", n), 0);
            biblscope = getBuffer(String.format("TEI.teiHeader.fileDesc.sourceDesc[%d].bibl.series.biblScope", n), 0);
        }
        else
        {
            series = getBuffer("TEI.teiHeader.fileDesc.titleStmt.series", 0).trim();
            biblscope = getBuffer("TEI.teiHeader.fileDesc.titleStmt.series.biblScope", 0).trim();
        }

        if (series.length() > 0 || biblscope.length() > 0)
            return String.format("%s %s", series, biblscope);
        else
            return "";
    }


    private void dumpProfile()
    {
        ArrayList<ElementInfo> eis = buffer.get("TEI.teiHeader.profileDesc.textClass.keywords.list.item");
        StringBuilder sbb = new StringBuilder();
        String delim = "";
        for (ElementInfo ei : eis)
        {
            sbb.append(String.format("%s%s", delim, ei.buffer.toString()));
            delim = ", ";
        }

        ElementInfo ei = buffer.get("TEI.teiHeader.profileDesc.textClass.catRef").get(0);
        String catref = ei.getValue("target");
        if (catref != null)
        {
            Regex re = new Regex(catref, "#(.*)$");
            if (re.find())
            {
                String sl = h.get(re.getGroup(1));
                if (sl != null)
                    catref = sl;
            }
        }

        if (!isBufferEmpty(sbb))
            sb.append(String.format(dataTemplate, "SubjectLC: ", sbb.toString()));
        if (catref != null)
            sb.append(String.format(dataTemplate, "SubjectSL: ", catref));
    }


    private String getBuffer(String key)
    {
        ElementInfo ei = getFirstElement(key);
        if (ei != null)
            return ei.buffer.toString();
        else
            return "";
    }
    
    private String getBuffer(String key, int idx)
    {
        ElementInfo ei = getElement(key, idx);
        if (ei != null)
            return ei.buffer.toString();
        else
            return "";        
    }


    private ElementInfo getElement(String key, int idx)
    {
        ArrayList<ElementInfo> eis = buffer.get(key);
        if (eis != null)
        {
            if (eis.size() > idx)
                return eis.get(idx);
            else
                return null;
        }
        else
            return null;

    }

    private ElementInfo getFirstElement(String key)
    {
        ArrayList<ElementInfo> eis = buffer.get(key);
        if (eis != null)
        {
            if (eis.size() > 0)
                return eis.get(0);
            else
                return null;
        }
        else
            return null;
    }

    private ElementInfo getLastElement(String key)
    {
        ArrayList<ElementInfo> eis = buffer.get(key);
        if (eis != null)
        {
            int idx = eis.size() - 1;
            if (idx >= 0)
                return eis.get(idx);
            else
                return null;
        }
        else
            return null;
    }

    private ElementInfo getCurrentElement()
    {
        String key = buildIndexedKey();
        return getLastElement(key);
    }


    private void copyElementTag(String qName, Attributes atts, String prefix)
    {
        debugBuffer.append(String.format("%s<%s ", prefix, qName));
        debugBuffer.append(copyAtts(atts));
        debugBuffer.append(">");
    }


    private String copyAtts(Attributes atts)
    {
        StringBuilder sbb = new StringBuilder();

        int numatts = atts.getLength();
        for (int i=0; i<numatts; i++)
        {
            sbb.append(atts.getQName(i));
            sbb.append("='");
            sbb.append(atts.getValue(i));
            sbb.append("' ");
        }

        return sbb.toString();
    }


    private boolean isTeiHeader(String qName)
    {
        boolean rc = qName.equals("teiHeader") && checkPath("TEI");
        return rc;
    }

    private boolean isBufferEmpty(StringBuilder buffer)
    {
        for (int i=0; i<buffer.length(); i++)
        {
            if (buffer.charAt(i) > ' ')
                return false;
        }

        return true;
    }


    private String getPath()
    {
        StringBuilder sbb = new StringBuilder();
        int len = elements.size();
        String delim = "";
        for (int i=0; i<len; i++)
        {
            sbb.append(delim);
            sbb.append(elements.get(i));
            delim = ".";
        }

        return sbb.toString();
    }

    private boolean checkPath(String... path)
    {
        boolean okay = false;
        if (elements.size() == path.length+1)
        {
            okay = true;
            for (int i=0; i<path.length && okay; i++)
            {
                okay = elements.get(i).equals(path[i]);
            }
        }

        return okay;
    }

    private boolean checkPathPrefix(String... path)
    {
        boolean okay = elements.size() >= path.length;
        for (int i=0; i<path.length && okay; i++)
        {
            okay = elements.get(i).equals(path[i]);
        }

        return okay;
    }

    private String buildKey()           // does not use indexes
    {
        keybfr.setLength(0);

        String delim = "";
        for (String element : elements)
        {
            keybfr.append(delim);
            keybfr.append(element);
            delim = ".";
        }
        String key = keybfr.toString();
        if (key.endsWith(".foreign"))
            key = "*.foreign";
        else if (key.endsWith(".ptr"))
            key = "*.ptr";
        return key;
    }

    private String buildIndexedKey()
    {
        keybfr.setLength(0);
        int len = elements.size();
        String delim = "";
        for (int i=0; i<len; i++)
        {
            keybfr.append(delim);
            keybfr.append(elements.get(i));
            int index = elementIndexes.get(i).intValue();
            if (index != 0)
                keybfr.append(String.format("[%d]", index));
            delim = ".";
        }
        return keybfr.toString();
    }

    private String buildParentKey()
    {
        String key = buildIndexedKey();
        return key.substring(0, key.lastIndexOf("."));
    }

    FreqInfo getFreqInfo(ArrayList<FreqInfo> a, String val)
    {
        for (FreqInfo fi : a)
        {
            if (fi.s.equals(val))
                return fi;
        }

        return null;
    }

    private void addLastVerse()
    {
        if (lv == null)
            return;

        String x = "?";

        Regex re = new Regex(lv, "(\\[[^\\]]+\\])");
        while (re.find())
        {
            String s = re.getGroup(1);
            Regex re2 = new Regex(s, ".(\\d+)");
            if (re2.find())
                x = re2.getGroup(1);
        }
        int n = lastVerse.size() - 1;
        if (n >= 0)
            lastVerse.set(n, x);
    }


    public void msg(String text) throws SAXException
    {
        throw new SAXException(String.format("%d.%d %s", locator.getLineNumber(), locator.getColumnNumber(), text));
    }

    // For ErrorHandler
    @Override
    public void warning(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void error(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException
    {
        throw exception;
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
    }

    @Override
    public void endDocument() throws SAXException
    {
        if (parseType == 3)
            addLastVerse();
        else if (parseType == 4)
        {
            Enumeration<String> en = wordh.keys();
            while (en.hasMoreElements())
            {
                WordInfo wi = wordh.get(en.nextElement());
                if (wi.numTexts > 1)
                    parseType = 4;
                if (wi.workingfrequency > wi.maxfrequency)
                {
                    wi.maxfrequency = wi.workingfrequency;
                    wi.filename = fileid;
                }
                wi.workingfrequency = 0;
            }
        }
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    private class ElementInfo
    {
        String qName;
        Hashtable<String, String> atts;
        StringBuilder buffer = new StringBuilder();
        int number;

        public ElementInfo(int number)
        {
            this.number = number;
        }

        public ElementInfo(String qName)
        {
            this.qName = qName;
        }

        void copyAttributes(Attributes a)
        {
            atts = new Hashtable<String, String>();
            int len = a.getLength();
            for (int i=0; i<len; i++)
            {
                atts.put(a.getQName(i), a.getValue(i));
            }
        }

        String getValue(String name)
        {
            if (atts != null)
                return atts.get(name);
            else
                return null;
        }
    }

    private class FreqInfo
    {
        public String s;
        public int count;
        public int location;

        public FreqInfo(String s, int count, int location)
        {
            this.s = s;
            this.count = count;
            this.location = location;
        }
    }


}
