package org.sanskritlibrary;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.List;
import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;

public class TranscodeFile  implements Comparator {
    private enum FileType { RTF, HTML, TEXT};
    //private char[] delimsOption = new char[] {'[', ']'};
    //private FSM.OptionsUnknown unknownOption;
    //private FSM.OptionsFlagged flaggedOption;
    //private boolean removeTags = false;

    public List<TextSegment> findSegments(Hashtable<String, Object> h, 
                                               List<TranscodingRule> transcodingRules,
                                               String text, String regexName) throws Exception {

        if (regexName.length() == 0) {
            return findSegmentsUsingTags(h, transcodingRules, text);
        } else {
            return findSegmentsUsingRegex(h, transcodingRules, text, regexName);
        }
	}


    public List<TextSegment> findSegmentsUsingTags(Hashtable<String, Object> h,
                                               List<TranscodingRule> transcodingRules,
                                               String text) {
        List<TextSegment> segments = new ArrayList<TextSegment>();

        int len = transcodingRules.size();
        for (int i = 0; i<len; i++)
        {
            TranscodingRule tr = transcodingRules.get(i);
            String startTag = tr.startTag;
            if (startTag.equals("*")) {
                h.put("defaultTranscodingRule", tr);
                continue;
            }
            String endTag = tr.endTag;
            TransformMap tm = tr.tm;
            int startTagLength = startTag.length();
            int endTagLength = endTag.length();
            if (startTagLength == 0 || endTagLength == 0) {    // user submitted empty tag, continue to next one to avoid infinite loop.
                continue;
            }
            int offset = 0;
            do
            {
                offset = text.indexOf(startTag, offset);
                if (offset != -1)
                {
                    int startOffset = offset;
                    offset = text.indexOf(endTag, startOffset + startTagLength);
                    if (offset != -1)
                    {
                        TextSegment ts = new TextSegment(startOffset, offset, tm, startTagLength, endTagLength, tr.newStartTag, tr.newEndTag);
                        segments.add(ts);
                    }
                    else
                    {
                        // ignore the tag, let user discover that it is still there.
                    }
                }
            } while (offset != -1);
        }

        Collections.sort(segments, this);

        return segments;
	}


    public List<TextSegment> findSegmentsUsingRegex(Hashtable<String, Object> h,
                                               List<TranscodingRule> trs,
                                               String text, String regexName) throws Exception {

        List<TextSegment> segments = new ArrayList<TextSegment>();
        String pattern = null;
        for (TranscodingRule tr : trs) {
            if (tr.regexName.equals(regexName)) {
                pattern = tr.pattern;
                break;
            }
        }
        if (pattern == null) {
            return segments;
        }

        PerlRegex re = new PerlRegex(text, pattern);
        while (re.find()) {
            int numGroups = re.getGroupCount();
            for (int i=1; i<=numGroups; i++) {
                TranscodingRule tr = findTranscodingRule(trs, regexName, i);
                if (tr.tm != null) {
                        int startOffset = re.getOffset(i);
                        int endOffset = startOffset + re.getGroup(i).length();
                        TextSegment ts = new TextSegment(startOffset, endOffset, tr.tm, 0, 0, tr.newStartTag, tr.newEndTag);
                        segments.add(ts);
                    }
                }
         }

        return segments;
    }


    public String doEncodings(Hashtable<String, Object> h, String text, List<TextSegment> segments,
            String uploadedFilename, List<TranscodingRule> tr) {
        FileType ft = FileType.TEXT;
        boolean uc = false;
        if (uploadedFilename.toLowerCase().endsWith(".rtf")) {
            ft = FileType.RTF;
            uc = text.contains("\\uc1");
        } else if (uploadedFilename.toLowerCase().endsWith(".html")) {
            ft = FileType.HTML;
        }
        TranscodingRule dtr = (TranscodingRule) h.get("defaultTranscodingRule");
        StringBuilder in = new StringBuilder(text);
        StringBuilder out = new StringBuilder();
        int offset = 0;
        String s = null;
        String intext = null;
        try {
            for (TextSegment ts : segments) {
                if (offset < ts.start) // if untagged text is before the tagged segment
                {
                    intext = in.substring(offset, ts.start);
                    if (dtr != null) {          // if transcode text with no tags
                        if (ft == FileType.HTML) {
                            intext = preprocessString(intext);
                        }
                        if (dtr.tm == null) {       // have startTag='*', sourceEncoding='asis'
                            s = intext;             // pass through as is
                        } else {
                            s = WebServices.transformString(intext, dtr.tm);
                            s = fixup(s, ft, uc);
                        }
                    } else {
                        s = intext;
                    }

                    out.append(s);
                }

                String startTag = "";
                String endTag = "";
                if (ts.tm != null) {
                    if (ts.tm.getRemoveTags() == false) {
                        if (ts.newStartTag != null) {
                            startTag = ts.newStartTag;
                            endTag = ts.newEndTag;
                        } else {
                            startTag = in.substring(ts.start, ts.start + ts.startTagLength);
                            endTag = in.substring(ts.end, ts.end + ts.endTagLength);
                        }
                    }

                    intext = in.substring(ts.start + ts.startTagLength, ts.end);
                    if (ft == FileType.HTML) {
                        intext = preprocessString(intext);
                    }
                    s = WebServices.transformString(intext, ts.tm);
                } else {
                    s = in.substring(ts.start + ts.startTagLength, ts.end); // must have startEncoding = 'asis'
                }

                out.append(startTag);
                out.append(fixup(s, ft, uc));
                out.append(endTag);
                offset = ts.end + ts.endTagLength;
            }

            if (offset < text.length()) // if untagged text remains
            {
                intext = in.substring(offset, text.length());
                if (dtr != null) {  // if transcode text with no tags
                    if (ft == FileType.HTML) {
                        intext = preprocessString(intext);
                    }
                    s = fixup(WebServices.transformString(intext, dtr.tm), ft, uc);
                } else {
                    s = intext;
                }

                out.append(s);
            }
        } catch (Exception ex) {
            System.out.println(ex.getMessage());
        }

        return out.toString();
    }



    private String preprocessString(String s) throws Exception {
        StringBuilder sb = new StringBuilder();
        int len = s.length();
        for (int i = 0; i < len; i++) {
            boolean append = true;
            if (s.charAt(i) == '&' && i < len - 1 && s.charAt(i + 1) == '#') {      // possible entity?
                int j = i + 2;   // move past &#
                while (j < len && s.charAt(j) != ';') {         // find end of entity
                    j++;
                }
                if (s.charAt(j) == ';') {       // might have an entity
                    char ch = 0;
                    try {
                        if (Character.toLowerCase(s.charAt(i + 2)) == 'x') {
                            ch = (char) Integer.parseInt(s.substring(i + 3, j), 16);    // convert from hexadecimal
                        } else {
                            ch = (char) Integer.parseInt(s.substring(i + 2, j), 10);    // convert from decimal
                        }
                        i = j;          // i is now the index of the ';'
                        sb.append(ch);  // ch is a 16 bit character
                        append = false;
                    } catch (Exception ex) {
                    }
                }
            }
            
            if (append) {                   // not an entity so just append character.
                sb.append(s.charAt(i));
            }
        }
        return sb.toString();
    }


    private TranscodingRule findTranscodingRule(List<TranscodingRule> trs, String regexName, int i) {
        for (TranscodingRule tr : trs) {
            if (tr.regexName != null && tr.regexName.equals(regexName) && tr.groupNumber == i) {
                return tr;
            }
        }
        return null;
    }

    private String fixup(String s, FileType ft, boolean uc) {
        if (ft == FileType.RTF) {
            StringBuilder sb = new StringBuilder();
            int len = s.length();
            for (int i=0; i<len; i++) {
                char ch = s.charAt(i);
                if (ch > 0x7f) {            // not lower ASCII
                    sb.append(TextFormatter.charToSlashUDecimal(ch));
                    if (uc)
                        sb.append("\\'3f");
                }
                else {
                    sb.append(ch);
                }
            }
            s = sb.toString();
        }
        else if (ft == FileType.HTML) {
            s = TextFormatter.charToEntity(s);
        }

        return s;
    }



    public int compare(Object o1, Object o2)
    {
        return ((TextSegment)o1).start - ((TextSegment)o2).start;
    }

}
