package org.sanskritlibrary;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Attr;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathFactory;
import javax.xml.xpath.XPathConstants;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.StringReader;
import java.util.ArrayList;
import org.xml.sax.InputSource;

public class XML
{
	static private XPath xpath;
	private Document d;
    private ArrayList<Node> nodes = new ArrayList<Node>();
    private Node node;

	public XML(String inputFileName) throws Exception
	{
		InputStream is = new FileInputStream(inputFileName);
		init(is);
		is.close();
	}

    public Node createNode(String name)
    {
        return d.createElement(name);
    }

    public Node copy(Node oldEntry)
    {
        Node newEntry = null;
        int nodeType = oldEntry.getNodeType();
        if (nodeType == d.ELEMENT_NODE)
        {
            newEntry = d.createElement(oldEntry.getNodeName());
            NamedNodeMap attributes = oldEntry.getAttributes();
            int len = attributes.getLength();
            // Copy attributes
            for (int i = 0; i < len; i++)
            {
                Node oldattr = attributes.item(i);
                Attr newattr = d.createAttribute(oldattr.getNodeName());
                newattr.setNodeValue(oldattr.getNodeValue());
                newEntry.getAttributes().setNamedItem(newattr);
            }

            // copy child elements
            NodeList oldChildren = oldEntry.getChildNodes();
            len = oldChildren.getLength();
            for (int i = 0; i < len; i++)
            {
                Node oldchild = oldChildren.item(i);
                newEntry.appendChild(copy(oldchild));
            }
        }
        else if (nodeType == d.TEXT_NODE)
        {
            newEntry = d.createTextNode(oldEntry.getNodeValue());
        }

        return newEntry;
    }


	// It is the user's responsibility to close the input stream.
	public XML(InputStream inputStream) throws Exception
	{
		init(inputStream);
	}

	public XML(StringReader inputStream) throws Exception
	{
		init(inputStream);
	}

	private void init(InputStream is) throws Exception
	{
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		DocumentBuilder db = dbf.newDocumentBuilder();
		d = db.parse(is);
	}

    private void init(StringReader sr) throws Exception
    {
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		DocumentBuilder db = dbf.newDocumentBuilder();
        InputSource is = new InputSource(sr);
		d = db.parse(is);
    }

	public Node getNode(String query, String... args) throws Exception
	{
		if (xpath == null)
			xpath = XPathFactory.newInstance().newXPath();

		String xq = String.format(query, (Object[])args);
		return (Node) xpath.evaluate(xq, d, XPathConstants.NODE);
	}


    public void pushNode(String query, String... args) throws Exception
    {
        node = getNode(query, args);
        nodes.add(node);
    }

    public void pushRelativeNode(String query, String... args) throws Exception
    {
        node = getRelativeNode(query, args);
        nodes.add(node);
    }


    public void pushNode(Node node)
    {
        this.node = node;
        nodes.add(node);
    }
    
    public Node popNode() throws Exception
    {
        int len = nodes.size();
        if (len == 0)
            throw new Exception("Stack is empty");

        Node top = node;
        nodes.remove(len-1);

        len = nodes.size();
        if (len > 0)
            node = nodes.get(len-1);
        else
            node = null;

        return top;
    }

	public NodeList getNodeList(String query, String... args) throws Exception
	{
		if (xpath == null)
			xpath = XPathFactory.newInstance().newXPath();

		String xq = String.format(query, (Object)args);
		return (NodeList) xpath.evaluate(xq, d, XPathConstants.NODESET);
	}

	static public Node getRelativeNode(Node node, String query, String... args) throws Exception
	{
		String xq = String.format(query, (Object)args);
		return (Node)xpath.evaluate(xq, node, XPathConstants.NODE);
	}

    public Node getRelativeNode(String query, String... args) throws Exception
    {
        return getRelativeNode(node, query, args);
    }

	static public NodeList getRelativeNodeList(Node node, String query, String... args) throws Exception
	{
		String xq = String.format(query, (Object)args);
		return (NodeList)xpath.evaluate(xq, node, XPathConstants.NODESET);
	}

    public NodeList getRelativeNodeList(String query, String... args) throws Exception
    {
        return getRelativeNodeList(node, query, args);
    }

	static public String getValue(Node node, String childName) throws Exception
	{
		Node n = XML.getRelativeNode(node, childName);
		if (n != null)
			return n.getTextContent();
        else
            throw new Exception(String.format("Child node %s does not exist.", childName));
	}

    public String getValue(String childName) throws Exception
    {
        return getValue(node, childName);
    }

    static public String getOptionalValue(Node node, String childName, String deflt) throws Exception
	{
		String result = deflt;
		Node n = XML.getRelativeNode(node, childName);
		if (n != null)
			result = n.getTextContent();

		return result;
	}

    public String getOptionalValue(String childName, String deflt) throws Exception
    {
        return getOptionalValue(node, childName, deflt);
    }

	static public String getOptionalNumber(Node node, String childName, String deflt) throws Exception
	{
		String result = deflt;
		Node n = XML.getRelativeNode(node, childName);
		if (n != null)
		{
			String s = n.getTextContent().trim();
			if (s.length() > 0)
				result = n.getTextContent();
		}

		return result;
	}

    public String getOptionalNumber(String childName, String deflt) throws Exception
    {
        return getOptionalNumber(node, childName, deflt);
    }

	static public String getAttribute(Node node, String attrName) throws Exception
	{
		Node n = XML.getRelativeNode(node, "@"+attrName);
		if (n != null)
			return n.getNodeValue();
		else
			return null;
	}

    public String getAttribute(String attrName) throws Exception
    {
        return getAttribute(node, attrName);
    }

	static public String getOptionalAttribute(Node node, String attrName, String dflt) throws Exception
	{
		Node n = XML.getRelativeNode(node, "@"+attrName);
		if (n != null)
			return n.getNodeValue();
		else
			return dflt;
	}

    public String getOptionalAttribute(String attrName, String dflt) throws Exception
    {
        return getOptionalAttribute(node, attrName, dflt);
    }

    public String getTextContent()
    {
        return node.getTextContent();
    }

    public String getCurrentNodeName()
    {
        if (node != null)
            return node.getNodeName();
        else
            return null;
    }
}
