package org.sanskritlibrary;

import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.ext.DefaultHandler2;
import java.util.Hashtable;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import org.xml.sax.InputSource;

public class XMLUtils extends DefaultHandler2 {
    public enum Types {PATHS, FORMAT, PARSE};
    private Locator locator;
    private Hashtable<String, Integer> h = new Hashtable<String, Integer>();
    private SAXParser saxParser;
    private BufferedWriter outxml;
    private int margin=-1;
    private String blanks = "                                                    ";
    private StringBuilder sbxml = new StringBuilder();
    boolean endCommentWithLineFeed = true;      // for comments before root element
    private Hashtable<String, Integer> paths;
    private Hashtable<String, Integer> lineNumber;
    private Hashtable<String, Boolean> textMissing;
    private String inxml;
    private StringBuilder buffer;
    Types type;
    String attName;
    ArrayList<ContextInfo> context = new ArrayList<ContextInfo>();

    public XMLUtils (Types type, String xml) throws Throwable {
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            this.type = type;
            if (type == Types.PARSE) {
                saxParser = factory.newSAXParser();
                InputSource is = new InputSource(new StringReader(xml));
                saxParser.parse(is, this);
            }
        }
        catch (Throwable t)
        {
            throw t;
        }

    }

    public XMLUtils (Types type, String inxml, String outxml) throws Throwable {
        this.type = type;
        this.inxml = inxml;
        if (outxml != null) {
            this.outxml = IOAssistance.openWriter(outxml);
        }
        if (type != Types.PATHS) {
            sbxml.append("<?xml version='1.0' encoding='UTF-8'?>\n");
        }

        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            saxParser = factory.newSAXParser();
            if (type == Types.FORMAT) {
                saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
            } else if (type == Types.PATHS) {
            }
        }
        catch (Throwable t)
        {
            closeFiles();
            if (type != Types.PARSE) {
                t.printStackTrace ();
            }
            throw t;
        }
    }

    public void setAttributeName (String attName) {
        this.attName = attName;
    }

    public String getResult () {
        switch (type) {
            case FORMAT:
                return sbxml.toString();
            case PATHS:
                return serializePaths();
        }

        return null;
    }

    public void parse() throws Throwable
    {
        InputStream is = null;
        try
        {
            is = new FileInputStream(inxml);
            saxParser.parse(is, this);
        }
        catch (Throwable t)
        {
            try {is.close();}catch (Exception ex){}
            throw t;
        }
    }

    public Hashtable<String, Integer> getResults () {
        return h;
    }


        // For ErrorHandler
    @Override
    public void warning(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void error(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    private String serializePaths () {
        StringBuilder sb = new StringBuilder();

        ArrayList<String> a = new ArrayList<String>();
        Enumeration<String> en = paths.keys();
        while (en.hasMoreElements()) {
            a.add(en.nextElement());
        }
        Collections.sort(a);
        ArrayList<PathInfo> pis = new ArrayList<PathInfo>();
        for (String s : a) {
            String append = "";
            if (textMissing.get(s) != null) {
                append = "#";
            }
            String ps = String.format("<path\n\tpattern=\"%s%s\"\n\ttemplate=''\n\tline='%d'\n\tcount='%d'\n</path>\n\n",
                    s, append, lineNumber.get(s).intValue(), paths.get(s).intValue());
            pis.add(new PathInfo(ps, lineNumber.get(s).intValue()));
        }
        
        Collections.sort(pis, new PathInfo());
        
        for (PathInfo pi : pis) {
            sb.append(pi.path);
        }

        return sb.toString();
    }


    private String getMargin() {
        if (margin > 0) {
            return blanks.substring(0, margin);
        } else {
            return "";
        }
    }



    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
        switch (type) {
            case FORMAT: break;
            case PATHS: 
                buffer = new StringBuilder();
                paths = new Hashtable<String, Integer>();
                lineNumber = new Hashtable<String, Integer>();
                textMissing = new Hashtable<String, Boolean>();
                break;
        }
    }

    private void closeFiles () {
        try {
            if (outxml != null) {
                outxml.close();
            }
        }
        catch (Exception ex) {

        }
    }

    private void pushContext (String qName, Attributes atts) {
        context.add(new ContextInfo(qName, atts));
    }

    private ContextInfo popContext () {
        ContextInfo ci = context.remove(context.size() - 1);
        return ci;
    }

    private ContextInfo peekContext () {
        return context.get(context.size() - 1);
    }

    private void savePath (String append) {
        StringBuilder sb = createPath();
        sb.append(append);
        String key = sb.toString();
        if (lineNumber.get(key) == null) {
            lineNumber.put(key, locator.getLineNumber());
        }
        Integer count = paths.get(key);
        if (count == null) {
            paths.put(key, new Integer(1));
        } else {
            paths.put(key, new Integer(count.intValue() + 1));
        }
    }

    private StringBuilder createPath () {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        ArrayList<String> a = new ArrayList<String> ();         // for sorting attribute names.
        for (ContextInfo ci : context) {
            sb.append(delim);
            delim = "/";
            sb.append(ci.qName);
            if (ci._atts.size() > 0) {
                Hashtable atts = ci._atts;
                Enumeration<String> en = atts.keys();
                a.clear();
                while (en.hasMoreElements()) {
                    a.add(en.nextElement());
                }
                Collections.sort(a);

                int attCount = 0;
                int len = a.size();
                for (int i=0; i<len; i++) {
                    if (attCount > 0) {
                        sb.append("+");
                    }
                    String name = a.get(i);
                    if (!name.equals("type")) {            // TODO: make this general
                        continue;
                    }

                    if (attCount == 0) {
                        sb.append("[");
                    }

                    attCount++;

                    sb.append(name);
                    sb.append("='");

                    if (name.equals(attName)) {
                        sb.append("0");
                    } else {
                        sb.append(atts.get(a.get(i)));
                    }
                    sb.append("'");
                }
                if (attCount > 0) {
                    sb.append("]");
                }
            }
        }
        return sb;
    }

    @Override
    public void endDocument() throws SAXException
    {
        try
        {
            if (outxml != null) {
                switch (type) {
                    case FORMAT:
                        outxml.write(sbxml.toString().replaceAll("&", "&amp;"));
                        break;
                    case PATHS:
                        outxml.write(serializePaths());
                        break;
                }
            }
        } catch (Exception ex) {
            System.out.println(ex.getMessage());
        } finally {
            closeFiles();
        }
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        switch (type)
        {
            case FORMAT:
                startElementFormat(qName, atts);
                break;
            case PATHS:
                startElementPaths(qName, atts);
                break;
        }
    }

    private void startElementFormat (String qName, Attributes atts) throws SAXException {
        margin++;
        int len = sbxml.length() - 1;
        while (len >= 0 && (sbxml.charAt(len) == ' ' || sbxml.charAt(len) == '\t')) {
            len--;
        }
        if (len >= 0) {
            sbxml.setLength(len+1);
        }

        if (len >=0 && sbxml.charAt(len) == '\n') {
            sbxml.append(getMargin());
        }

        sbxml.append(String.format("<%s", qName));

        len = atts.getLength();
        for (int i = 0; i < len; i++) {
            sbxml.append(String.format(" %s='%s'", atts.getQName(i), atts.getValue(i)));
        }
        sbxml.append(">");
    }


    private void startElementPaths (String qName, Attributes atts) throws SAXException {
        pushContext(qName, atts);
        if (atts.getValue(attName) != null) {
            savePath("");
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        switch (type) {
            case FORMAT:
                endElementFormat (qName);
                break;
            case PATHS:
                endElementPaths ();
                break;
        }
    }

    private void endElementFormat (String qName) {
        int idx = sbxml.length() - 1;
        while (idx >= 0 && (sbxml.charAt(idx) == ' ' || sbxml.charAt(idx) == '\t')) {
            idx--;
        }
        boolean slashOnly = false;
        boolean newline = true;
        if (idx >=0) {
            newline = sbxml.charAt(idx) == '\n';
            sbxml.setLength(idx+1);
        }
        if (newline) {
            sbxml.append(getMargin());
        }
        else {   // check for <foo></foo> and output </foo> instead
            if (sbxml.charAt(idx) == '>') {
                slashOnly = true;
                int len = qName.length();
                int start = idx - len - 1;
                if (sbxml.charAt(start) == '<') {
                    start++;   // move past the '<'
                    for (int i=0; slashOnly && i<len; i++) {
                        if (sbxml.charAt(start+i) != qName.charAt(i))
                            slashOnly = false;
                    }
                }
                else {
                    slashOnly = false;
                }
            }
        }

        if (slashOnly) {
            sbxml.setLength(idx);
            sbxml.append("/>");
        } else {
            sbxml.append(String.format("</%s>", qName));
        }
        margin--;
    }

    private void endElementPaths () throws SAXException {

        if (buffer.indexOf("[[") != -1) {
            if (peekContext()._atts.get(attName) == null) {
                savePath("*");
            }
        } else {
            StringBuilder sb = createPath();
            textMissing.put(sb.toString(), true);
        }

        buffer.setLength(0);
        popContext();
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        String s = new String(ch, start, length);
        switch (type) {
            case FORMAT:
                charactersFormat(s);
                break;
            case PATHS:
                buffer.append(s);   // in case it has [[..]]
                break;
        }
    }

    private void charactersFormat(String s) {
        sbxml.append(s);
    }

    @Override
    public void comment(char[] ch, int start, int length) throws SAXException {
        sbxml.append("<!--");
        sbxml.append(new String(ch, start, length));
        sbxml.append("-->");
        if (endCommentWithLineFeed) {
            sbxml.append("\r\n");
        }
    }

    @Override
    public void processingInstruction(String target, String data) throws SAXException {
        if (type == Types.FORMAT) {
            try {
                sbxml.append("<?");
                sbxml.append(target);
                sbxml.append(" ");
                sbxml.append(data);
                sbxml.append("?>\n");
            } catch (Exception ex) {
                throw new SAXException(ex.getMessage());
            }
        }
    }

    class ContextInfo {
        String qName;
        Hashtable<String, String> _atts = new Hashtable<String, String>();
        public ContextInfo(String qName, Attributes atts) {
            this.qName = qName;
            int len = atts.getLength();
            for (int i=0; i<len; i++) {
                _atts.put(atts.getQName(i), atts.getValue(i));
            }
        }
    }

    class PathInfo implements Comparator<PathInfo>{
        String path;
        int lineNumber;
        public PathInfo () {

        }
        
        public PathInfo (String path, int lineNumber) {
            this.path = path;
            this.lineNumber = lineNumber;
        }

        public int compare(PathInfo o1, PathInfo o2) {
            return o1.lineNumber - o2.lineNumber;
        }
    }
}
