package org.sanskritlibrary.admin;

import org.sanskritlibrary.XML;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.ResourceInputStream;
import org.sanskritlibrary.RawEncoder;
import org.sanskritlibrary.RawDecoder;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.nio.charset.Charset;
import org.sanskritlibrary.FSM;
import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;

public class FileManager
{
	ArrayList<Integer> pageCounts = new ArrayList<Integer>();
	StringBuilder lsb = new StringBuilder();		// an attempt to reduce garbage collection
	String directoryName;
	XML doc;
	ResourceInputStream ris;
	int _bufferSize;
	String _eol;
	String _nameTemplate;
	Hashtable<String, FSM[]> fsmh;
	Hashtable<String, String> duplicateVerse = new Hashtable<String, String>();
	boolean createAltEncodings = true;
	String textName;
	int totalPages;
	boolean unanalyzed = false;

	public FileManager(String directoryName, String configFileName,
						ResourceInputStream ris, boolean createAltEncodings,
						String textName) throws Exception
	{
		if (!directoryName.endsWith("/"))
		{
			directoryName += "/";
		}
		this.directoryName = directoryName;
		this.doc = new XML(configFileName);
		this.ris = ris;
		this.createAltEncodings = createAltEncodings;
		this.textName = textName;
		fsmh = new Hashtable<String, FSM[]>();
	}

	public ArrayList<String> getTextNames() throws Exception
	{
		ArrayList<String> names = new ArrayList<String>();
		NodeList texts = doc.getNodeList("/texts/text");
		for (int i=0; i<texts.getLength(); i++)
		{
			Node textNode = texts.item(i);
			String name = XML.getRelativeNode(textNode, "name").getTextContent();
			names.add(name);
		}

		return names;
	}

	public void updateFiles() throws Exception
	{
		Node pagination = doc.getNode("/texts/pagination");
		_bufferSize = Integer.parseInt(XML.getOptionalValue(pagination, "bufferSize", "10"));
		_eol = XML.getOptionalValue(pagination, "eol", "<br/>");
		_nameTemplate = XML.getOptionalValue(pagination, "nameTemplate", "%spages/%s_page%d.%s");

		NodeList texts = doc.getNodeList("/texts/text");
		NodeList altEncodingsGlobal = doc.getNodeList("/texts/altEncodings/altEncoding");
		for (int i = 0; i < texts.getLength(); i++)
		{
			Node textNode = texts.item(i);
			String name = XML.getRelativeNode(textNode, "name").getTextContent();
			if (textName != null && !name.equals(textName))  // remove
				continue;
			NodeList chapters = XML.getRelativeNodeList(textNode, "structure/book/chapter");
			String fileName = XML.getRelativeNode(textNode, "fileName").getTextContent();
			String originalEncoding = XML.getRelativeNode(textNode, "originalEncoding/encoding").getTextContent();
			String originalFilePath = String.format("%s%s.%s", directoryName, fileName, originalEncoding);
			this.unanalyzed = XML.getRelativeNode(textNode, "unanalyzed") != null;
			Node versePattern = XML.getRelativeNode(textNode, "versePattern");
			// Do encodings that each file should have
			for (int j = 0; j < altEncodingsGlobal.getLength(); j++)
			{
				Node altEncodingNode = altEncodingsGlobal.item(j);
				doAltEncoding(chapters, versePattern, altEncodingNode, originalEncoding, originalFilePath, fileName);
			}

			// Do encodings that are specific to this file
			NodeList altEncodings = XML.getRelativeNodeList(textNode, "altEncoding");
			for (int j = 0; j < altEncodings.getLength(); j++)
			{
				Node altEncodingNode = altEncodings.item(j);
				doAltEncoding(chapters, versePattern, altEncodingNode, originalEncoding, originalFilePath, fileName);
			}
		}
	}

	private void doAltEncoding(NodeList chapters, Node versePattern, Node altEncodingNode, String originalEncoding, String originalFilePath, String originalFileName)
			  throws Exception
	{
		String targetEncoding = XML.getRelativeNode(altEncodingNode, "encoding").getTextContent();
		RawEncoder encoder = FSM.getTargetEncoder(originalEncoding, targetEncoding, fsmh);
		if (!targetEncoding.equals(originalEncoding))
		{
			String map = String.format("*:%s->%s", originalEncoding, targetEncoding);
			TransformMap tm = new TransformMap(map, ris, fsmh);
			RawDecoder decoder = FSM.getSourceDecoder(originalEncoding, targetEncoding, fsmh);
			String targetFilePath = String.format("%s%s.%s", directoryName, originalFileName, targetEncoding);

			if (createAltEncodings)
			{
				BufferedReader br = IOAssistance.openReader(originalFilePath, decoder);
				BufferedWriter bw = IOAssistance.openWriter(targetFilePath, encoder);
				String line = null;
				int debug = 0;
				while ((line = br.readLine()) != null)
				{
					line = removeRedundantCharacters(line);
					VerseInfo vi = new VerseInfo(line, versePattern);
					if (!vi.isError)
					{
						line = line.substring(0, vi.textOffset) +
									WebServices.transformString(line.substring(vi.textOffset), tm);
					}
					else
					{
						line = null;
						// line = WebServices.transformString(line, tm);  //!! what to do here?
					}
					if (line != null)
					{
						bw.write(line);
						bw.write("\n");
					}
				}
				br.close();
				bw.close();
			}
		}
		Node paginateNode = XML.getRelativeNode(altEncodingNode, "paginate");

		createPaginatedFiles(directoryName, originalFileName, versePattern, targetEncoding, paginateNode, encoder);
	}

	private void createPaginatedFiles(String directoryName, String fileName,
			  Node versePattern, String targetEncoding, Node paginateNode,
			  RawEncoder encoder)
			  throws Exception
	{
		int debug = 0;
		this.duplicateVerse.clear();
		try
		{
			String pagemap = "";
			String delim = "";

			// possibly override values of bufferSize, eol or nameTemplate
			int bufferSize = _bufferSize;
			String eol = _eol;
			String nameTemplate = _nameTemplate;
			if (paginateNode != null)
			{
				bufferSize = Integer.parseInt(XML.getOptionalValue(paginateNode, "bufferSize", Integer.toString(_bufferSize)));
				eol = XML.getOptionalValue(paginateNode, "eol", _eol);
				nameTemplate = XML.getOptionalValue(paginateNode, "nameTemplate", _nameTemplate);
			}

			bufferSize *= 1024;

			// If original encoding is not slp1 then create a TransformMap to map each word
			// to slp1 as a parameter to getWord
			String mapslp1 = null;
			TransformMap tmslp1 = null;
			if (!targetEncoding.equals("slp1"))
			{
				mapslp1 = String.format("*:%s->slp1", targetEncoding);
				tmslp1 = new TransformMap(mapslp1, ris, fsmh);
			}

			String sourceFilePath = String.format("%s%s.%s", directoryName, fileName, targetEncoding);
			BufferedReader br = IOAssistance.openReader(sourceFilePath, encoder != null ? new RawDecoder() : null);
			BufferedWriter bw = null;
			int pageNum = 1;
			String line = null;
			int verseID = 1;
			String lineBuffer = "";
			VerseInfo prevvi = null;
			VerseInfo buffervi = null;
			boolean usePageCount = pageCounts.size() > 0;
			int currentPageCount = 0;
			int pageCountIndex = 0;
			if (usePageCount)
				currentPageCount = pageCounts.get(0);
			while (true)
			{
				String pageFilePath = String.format(nameTemplate, directoryName, fileName, pageNum, targetEncoding);
				bw = IOAssistance.openWriter(pageFilePath, encoder);
				int length = 0;
				int numVerses = 0;
				while (usePageCount ? numVerses < currentPageCount : length < bufferSize)
				{
					debug++;
					line = br.readLine();
					if (line == null)
					{
						if (lineBuffer.length() > 0)		// handle last buffer if any
						{
							length += writeLine(bw, eol, lineBuffer, verseID, buffervi, tmslp1);
							verseID++;
							numVerses++;
						}
						break;
					}
					else
					{
						VerseInfo vi = new VerseInfo(line, versePattern);
						if (!vi.isError)
						{
							// The following logic assumes that the verses are in strictly ascending order.
							// See checkFiles below and the compare method of VerseInfo
							if (vi.pada == -1)  // if no pada, dump buffered pada'ed verse if any and write un pada-ed line
							{
								if (lineBuffer.length() > 0)
								{
									length += writeLine(bw, eol, lineBuffer, verseID, buffervi, tmslp1);
									verseID++;
									numVerses++;
									lineBuffer = "";
									buffervi = null;
								}
								length += writeLine(bw, eol, vi.prefix(line), verseID, vi, tmslp1);
								verseID++;
								numVerses++;
							}
							else if (prevvi != null && prevvi.compareNoPada(vi) < 0)  // has pada and begins new verse
							{
								if (lineBuffer.length() > 0)
								{
									length += writeLine(bw, eol, lineBuffer, verseID, buffervi, tmslp1);
									verseID++;
									numVerses++;
								}
								lineBuffer = vi.prefix(line);		// start a new line buffer
								buffervi = vi;
							}
							else  // add new pada to previous padas for the book.chapter.verse
							{
								if (buffervi == null)
									buffervi = vi;
								if (lineBuffer.length() != 0)		// add <br/> to separate from previous pada
									lineBuffer += "<br/>";
								lineBuffer += vi.prefix(line);
							}

							prevvi = vi;
						}
					}
				}
				bw.close();

				if (usePageCount)
				{
					pageCountIndex++;
					if (pageCountIndex < pageCounts.size())
						currentPageCount = pageCounts.get(pageCountIndex);
					else
					{
						line = line != null ? br.readLine() : null;
						if (line != null)
							throw new Exception("page counts out of sync.");
					}
				}
				else
					pageCounts.add(numVerses);		// accumulate for next pagination

				pagemap += String.format("%s%d", delim, numVerses);
				delim = ",";

				if (line == null)
				{
					break;
				}
				pageNum++;
			}

			// write _map file even though all will be identical. I test for the existance of the file in TextViewer.java
			String pageMapPath = String.format("%spages/%s_map.%s", directoryName, fileName, targetEncoding);
			bw = IOAssistance.openWriter(pageMapPath);
			bw.write(pagemap);
			bw.close();

			br.close();
		}
		catch (Exception ex)
		{
			debug++;
		}
	}

	private int writeLine(BufferedWriter bw, String eol, String line,
									int verseID, VerseInfo vi, TransformMap tmslp1)
									throws Exception
	{
		int length = eol.length() + line.length();
		line = createLinks(verseID, vi, line, tmslp1);

		bw.write(line);
		if (eol.equals("\\n"))
			bw.write("\n");
		else
			bw.write(eol);

		return length;
	}

	// returns true if verse has already been encountered.
	private boolean checkForDuplicateVerse(VerseInfo vi)
	{
		int n = 0;
		if (duplicateVerse.containsKey(vi.toString()))
			vi = null;
		else
			duplicateVerse.put(vi.toString(), "x");

		return vi == null;
	}

	private String createLinks(int verseID, VerseInfo vi, String line,
			  TransformMap tm)
			  throws Exception
	{
		if (vi.isError)
		{
			return "error";
		}
		lsb.setLength(0);
		lsb.append(String.format("<p id='%d' class='verse'>", verseID));

		int textOffset = vi.textOffset;

		String span = String.format("<span class='verseid'>%s</span><br/>", line.substring(0, textOffset-1));
		lsb.append(span);
		line = vi.fixup(line.substring(textOffset));
		if (!unanalyzed)
		{
			String[] words = line.split("[\\s]+");

			for (String w : words)
			{
                if (w.contains("-") || w.contains("="))
                {
                    int len = w.length();
                    int start = 0;
                    for (int i=0; i<len; i++)
                    {
                        char ch = w.charAt(i);
                        if (ch == '-' || ch == '=')
                        {
                            _createLinks(w.substring(start, i), tm, lsb);
                            lsb.append(ch);
                            lsb.append(" ");
                            start = i + 1;   // move to start of next segment
                        }
                    }
                    _createLinks(w.substring(start), tm, lsb);
                }
                else
                    _createLinks(w, tm, lsb);
			}
		}
		else
		{
			String slp1 = null;
			if (tm != null)
				slp1 = WebServices.transformString(line, tm);
			else
				slp1 = line;
			slp1 = slp1.replaceAll("=", " ");
			slp1 = slp1.replaceAll("\\.", "");
			slp1 = slp1.replaceAll(",", "");
			lsb.append(String.format("<a class='ml' onclick=\"parseVerse('%s')\">%s</a>", slp1, line));
		}

		lsb.append("</p>");
		return lsb.toString();
	}


    private void _createLinks(String w, TransformMap tm, StringBuilder lsb) throws Exception
    {
        int len = w.length();
		boolean sc = false;
		if (w.endsWith(";"))
		{
			w = w.substring(0, len - 1);
			sc = true;
		}
		else if (w.endsWith("."))
		{
            w = w.substring(0, len - 1);
		}

		String slpw = w;
		if (tm != null)
		{
            slpw = WebServices.transformString(w, tm);
        }

		lsb.append(String.format("<a class='ml' onclick='getWord(\"%s\", this)'>%s</a> ", slpw, w));
		if (sc)
		{
			lsb.append("<br/>");
		}
    }

	public String removeRedundantCharacters(String line)
	{
		if (line == null)
		{
			return line;
		}

		char satlongudatta = '\u00DA';	// {shift option 1: dead key preceding the vowel it accents}
		char satlongsvarita = '\u00D9';
		char satmacron = '\u00CF';
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < line.length(); i++)
		{
			char current = line.charAt(i);
			if (skip(line, i, current, satlongudatta) ||
					  skip(line, i, current, satlongsvarita) ||
					  skip(line, i, current, satmacron))
			{
				continue;
			}
			sb.append(current);
		}
		return sb.toString();
	}

	private boolean skip(String s, int i, char current, char ch)
	{
		if (i == s.length() - 1)
		{
			return current == ch;
		} else
		{
			return current == ch && s.charAt(i + 1) == ch;
		}
	}

	public String checkFiles() throws Exception
	{
		int currentBook = -1;
		int currentChapter = -1;
		int lastVerseNumber = 0;
		int vidx = 0;
		int currentVerseCount = 0;
		int lineNumber = 0;
		StringBuilder sb = new StringBuilder();
		VerseInfo prevvi = null;
		try
		{
		String targetEncoding = "roman";  //!! hack, since unlikely that source file is in roman Unicode
		NodeList texts = doc.getNodeList("/texts/text");
		for (int i = 0; i < texts.getLength(); i++)
		{
			Node textNode = texts.item(i);
			String name = XML.getRelativeNode(textNode, "name").getTextContent();
			sb.append(String.format("Document: %s<br/>", name));
			if (textName != null && !name.equals(textName))
				continue;
			NodeList chapters = XML.getRelativeNodeList(textNode, "structure/book/chapter");
			int[] verseCounts = new int[chapters.getLength()];
			for (int j = 0; j < verseCounts.length; j++)
			{
				Node node = chapters.item(j);
				verseCounts[j] = Integer.parseInt(node.getAttributes().getNamedItem("numVerses").getNodeValue());
			}
			String fileName = XML.getRelativeNode(textNode, "fileName").getTextContent();
			String originalEncoding = XML.getRelativeNode(textNode, "originalEncoding/encoding").getTextContent();
			String originalFilePath = String.format("%s%s.%s", directoryName, fileName, originalEncoding);
			Node versePattern = XML.getRelativeNode(textNode, "versePattern");
			String map = String.format("*:%s->%s", originalEncoding, targetEncoding);
			TransformMap tm = new TransformMap(map, ris, fsmh);
			RawDecoder decoder = FSM.getSourceDecoder(originalEncoding, targetEncoding, fsmh);
			BufferedReader br = IOAssistance.openReader(originalFilePath, decoder);

			String line = null;
			currentChapter = -1;
			lastVerseNumber = 0;
			vidx = 0;
			currentVerseCount = verseCounts[vidx];
			lineNumber = 0;
			while ((line = br.readLine()) != null)
			{
				lineNumber++;
				if (lineNumber == 50)
					lineNumber = 50;
				VerseInfo vi = new VerseInfo(line, versePattern);
				if (!vi.isError)
				{
					int missingVerses = 0;
					int missingFirstVerses = 0;
					if (currentChapter == -1)
					{
						currentBook = vi.book;
						currentChapter = vi.chap;
					}
					if (vi.book != currentBook || vi.chap != currentChapter)
					{
						if (lastVerseNumber != currentVerseCount)  // check if any verses missing at the end of the old chapter
						{
							missingVerses = currentVerseCount - lastVerseNumber;
						}

						if (vi.verse != 1)		// check if verses missing from beginning of new chapter
							missingFirstVerses = vi.verse - 1;

						vidx++;
						currentVerseCount = verseCounts[vidx];
					}
					else if (lastVerseNumber + 1 != vi.verse)
					{
						missingVerses = vi.verse - lastVerseNumber - 1;
					}

					// add holders for missing verses, if any
					int errorCount = 0;
					for (int j = 0; j < missingVerses + missingFirstVerses; j++)
					{
						lastVerseNumber++;
						String temp = null;
						int cc = -1;
						int cv = -1;
						int cb = -1;
						if (j < missingVerses) // generate missing verses in current chapter
						{
							cb = currentBook;
							cc = currentChapter;
							cv = lastVerseNumber;
						}
						else	// generate missing verses at beginning of new chapter
						{
							cb = vi.book;
							cc = vi.chap;
							cv = j - missingVerses + 1;
						}
						if (vi.book == -1)
						{
							temp = String.format("%d.%d\t%s",
									  cc, cv, "Missing");
						}
						else
						{
							temp = String.format("%d.%d.%d\t%s",
									  cb, cc, cv, "Missing");
						}

						VerseInfo tvi = new VerseInfo(temp, versePattern);
						if (checkForDuplicateVerse(tvi))
							sb.append(String.format("%d) %s: missing, duplicate<br/>", lineNumber, tvi.toString()));
						else if (errorCount < 5)
							sb.append(String.format("%d) %s: missing<br/>", lineNumber, tvi.toString()));
						else if (errorCount == 5)
							sb.append(String.format("%d) %d more...<br/>", lineNumber, missingVerses - errorCount - 1));
						errorCount++;
						prevvi = tvi;
					}
					if (vi.pada<2 && checkForDuplicateVerse(vi))
						sb.append(String.format("%d) duplicate: %s<br/>", lineNumber, vi.toString()));
					lastVerseNumber = vi.verse;
					currentChapter = vi.chap;
					currentBook = vi.book;
					if (prevvi != null)
					{
						int rc = prevvi.compare(vi);
						if (rc >= 0)		// if prevvi is greater or equal to vi
							sb.append(String.format("%d) %s should be before $s", lineNumber, prevvi.toString(), vi.toString()));
					}
					else
						prevvi = vi;
				}
				else
				{
					int len = line.length();
					if (len > 40)
						len = 40;
					sb.append(String.format("%d) Unmatched line: %s ...<br/>", lineNumber, line.substring(0, len)));
				}
			}
		}
		return sb.toString();
		}
		catch (Exception ex)
		{
			String s = ex.getMessage();
			sb.append(String.format("%d) Exception: %s", lineNumber, ex.getMessage()));
			return sb.toString();
		}
	}


	public ArrayList<WordInfo> createWordList(String... encoding) throws Exception
	{
		ArrayList<WordInfo> a = new ArrayList<WordInfo>();
		NodeList texts = doc.getNodeList("/texts/text");
		for (int i = 0; i < texts.getLength(); i++)
		{
			Node textNode = texts.item(i);
			String name = XML.getRelativeNode(textNode, "name").getTextContent();
			if (textName != null && !name.equals(textName))
				continue;
			String fileName = XML.getRelativeNode(textNode, "fileName").getTextContent();
			String originalEncoding = "roman";
			String originalFilePath = null;
			if (encoding.length != 0)
				originalEncoding = encoding[0];
			originalFilePath = String.format("%s%s.%s", directoryName, fileName, originalEncoding);
			Node versePattern = XML.getRelativeNode(textNode, "versePattern");

			Hashtable<String, Integer> h = new Hashtable<String, Integer>();
			BufferedReader br = IOAssistance.openReader(originalFilePath);
			String line;
			while ((line=br.readLine())!= null)
			{
				VerseInfo vi = new VerseInfo(line, versePattern);
				String[] words = vi.text.split("[,;=*.\\s\\)]+");
				for (String word : words)
				{
                    Regex re = new Regex(word, "(\\d)");  // ignore numbers
                    if (re.find())
                        continue;
                    word = word.replace("(", "");
                    word = word.replace(")", "");
                    word = word.replace("[", "");
                    word = word.replace("]", "");
					if (word.length() > 0)
					{

						if (!h.containsKey(word))
							h.put(word, 1);
						else
							h.put(word, h.get(word)+1);
					}
				}
			}

			Enumeration<String> en = h.keys();
			while(en.hasMoreElements())
			{
				String word = en.nextElement();
				a.add(new WordInfo(word, h.get(word)));
			}

			WordInfo comp = new WordInfo();
			WordInfo.sortByWord = true;
			Collections.sort(a, comp);
			WordInfo.sortByWord = false;
			Collections.sort(a, comp);
		}

		return a;
	}

	public void writeWordList(String wordList) throws Exception
	{
		NodeList texts = doc.getNodeList("/texts/text");
		for (int i = 0; i < texts.getLength(); i++)
		{
			Node textNode = texts.item(i);
			String name = XML.getRelativeNode(textNode, "name").getTextContent();
			if (!name.equals(textName))
				continue;

			String filePath = String.format("%spages/wordlists/%s.html", directoryName, name);
			BufferedWriter bw = IOAssistance.openWriter(filePath, Charset.forName("UTF-8"));
			bw.write(wordList);
			bw.close();
		}
	}

	public String getWordList(int currentPage, int totalPages, int wordsPerPage) throws Exception
	{
		StringBuilder sb = new StringBuilder();
		boolean computeNumPages = false;
		if (currentPage == 0)
		{
			currentPage = 1;
			computeNumPages = true;
			this.totalPages = 0;
		}
		else
			this.totalPages = totalPages;

		NodeList texts = doc.getNodeList("/texts/text");
		for (int i = 0; i < texts.getLength(); i++)
		{
			Node textNode = texts.item(i);
			String name = XML.getRelativeNode(textNode, "name").getTextContent();
			if (textName != null && !name.equals(textName))
				continue;

			String filePath = String.format("%spages/wordlists/%s.html", directoryName, name);
			BufferedReader br = IOAssistance.openReader(filePath);
			String line = null;
			int lineNumber = 0;
			int firstLine = (currentPage-1) * wordsPerPage;
			int lastLine = firstLine + wordsPerPage - 1;
			while ((line=br.readLine())!=null)
			{
				if (lineNumber % wordsPerPage == 0)
					this.totalPages++;

				lineNumber++;
				if (lineNumber >= firstLine && lineNumber <= lastLine)
				{
					sb.append(line);
					sb.append("\n");
				}

				if (lineNumber == lastLine && !computeNumPages)
					break;
			}
			break;
		}

		return sb.toString();
	}

	public String findVerses(String word) throws Exception
	{
		StringBuilder sb = new StringBuilder();
		NodeList texts = doc.getNodeList("/texts/text");
		Regex regex = new Regex("", String.format("\\b%s\\b", word));
		for (int i = 0; i < texts.getLength(); i++)
		{
			Node textNode = texts.item(i);
			String name = XML.getRelativeNode(textNode, "name").getTextContent();
			if (textName != null && !name.equals(textName))
				continue;

			String filePath = String.format("%stexts/%s.roman", directoryName, name);
			BufferedReader br = IOAssistance.openReader(filePath);
			String line = null;
			while ((line=br.readLine())!=null)
			{
				regex.reset(line);
				if (regex.find())
				{
					int ofs = regex.getOffset(0);
					sb.append(String.format("<p>%s<span style='color:red'>%s</span>%s</p>",
							  line.substring(0, ofs),
							  word,
							  line.substring(ofs+word.length())));
				}
			}
			break;
		}

		return sb.toString();
	}

	public void transformWordLists(ResourceInputStream ris) throws Exception
	{
		NodeList texts = doc.getNodeList("/texts/text");
		Hashtable<String, FSM[]> fsmh = new Hashtable<String, FSM[]>();
		for (int i = 0; i < texts.getLength(); i++)
		{
			Node textNode = texts.item(i);
			String name = XML.getRelativeNode(textNode, "name").getTextContent();
			String filePath = String.format("%spages/wordlists/%s.html",
					  directoryName, name);

			for (String encoding : new String[]{"slp1", "roman", "deva"})
			{
				String map = String.format("*:roman->%s;", encoding);
				BufferedReader br = IOAssistance.openReader(filePath);
				String outputPath = String.format("%spages/wordlists/%s.%s",
						  directoryName, name, encoding);
				BufferedWriter bw = IOAssistance.openWriter(outputPath, Charset.forName("UTF-8"));
				String line = null;
				while ((line = br.readLine())!=null)
				{
					String frequency = null;
					String word = null;
					Regex re = new Regex(line, "^(\\d+)\\s+<a>([^\\\\]*)</a>");
					if (re.find())
					{
						frequency = re.getGroup(1);
						word= re.getGroup(2);
					}
					else
						throw new Exception(String.format("Unexpected pattern in %s: %s", filePath, line));

					try
					{
						String result = word;
						if (!encoding.equals("roman"))
						{
							TransformMap tm = new TransformMap(map, ris, fsmh);
							result = WebServices.transformString(word, tm);
						}
						bw.write(String.format("%s\t%s\n", frequency, result));
					}
					catch (Exception ex)
					{
						String lastError = ex.getMessage();
						throw new Exception(lastError);
					}

				}
				bw.close();
				br.close();
			}
		}
	}

	public int getTotalPages()
	{
		return totalPages;
	}

}
