package org.sanskritlibrary.admin;
import org.sanskritlibrary.XML;
import org.sanskritlibrary.Regex;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.util.ArrayList;

public class VerseInfo
{
	public int book = -1;
	public int chap = -1;
	public int verse = -1;
	public int pada = -1;
	public String text = null;
	public int textOffset = 0;
	public boolean isError = false;
	public ArrayList<ReplacementInfo> replacementInfo = new ArrayList<ReplacementInfo>();

	public VerseInfo(String line, Node versePattern) throws Exception
	{
		String pattern = XML.getOptionalValue(versePattern, "regexp", "");
		Node replacements = XML.getRelativeNode(versePattern, "replacements");
		if (replacements != null)
		{
			NodeList list = XML.getRelativeNodeList(replacements, "replacement");
			for (int i=0; i<list.getLength(); i++)
			{
				replacementInfo.add(new ReplacementInfo(list.item(i)));
			}
		}
		int bookGroup = Integer.parseInt(XML.getOptionalNumber(versePattern, "book", "-1"));
		int chapGroup = Integer.parseInt(XML.getOptionalNumber(versePattern, "chapter", "-1"));
		int verseGroup = Integer.parseInt(XML.getOptionalNumber(versePattern, "verse", "-1"));
		int padaGroup = Integer.parseInt(XML.getOptionalNumber(versePattern, "pada", "-1"));
		int textGroup = Integer.parseInt(XML.getOptionalValue(versePattern, "text", "-1"));
		Regex re = new Regex(line, pattern);
		if (!re.find())
			isError = true;
		else
		{
			if (bookGroup != -1)
				book = parseInt(re.getGroup(bookGroup));
			if (chapGroup != -1)
				chap = parseInt(re.getGroup(chapGroup));
			if (verseGroup != -1)
				verse = parseInt(re.getGroup(verseGroup));
			if (padaGroup != -1)
				pada = parseInt(re.getGroup(padaGroup));
			if (textGroup != -1)
			{
				text = re.getGroup(textGroup);
				if (text != null)
					textOffset = re.getOffset(textGroup);
			}
		}
	}

	public int compare(VerseInfo vi)
	{
		int rc = this.book - vi.book;
		if (rc == 0)		// books are equal
		{
			rc = this.chap - vi.chap;
			if (rc == 0)	// books and chapters are equal
			{
				rc = this.verse - vi.verse;
				if (rc == 0)	// books, chapters and verses are equal
				{
					if (this.pada != -1 && vi.pada != -1)
						rc = this.pada - vi.pada;
					else if (this.pada != -1)
						rc = 1;		// verse with a pada is after one that doesn't have a pada, everything else being equal
					else if (vi.pada != -1)
						rc = -1;
					// if neither verse has a pada the verses are equal.
				}
			}
		}
		
		return rc;
	}

	public int compareNoPada(VerseInfo vi)
	{
		int rc = this.book - vi.book;
		if (rc == 0)		// books are equal
		{
			rc = this.chap - vi.chap;
			if (rc == 0)	// books and chapters are equal
				rc = this.verse - vi.verse;
		}

		return rc;
	}

	public String fixup(String line)
	{
		for (ReplacementInfo ri : replacementInfo)
		{
			line = Regex.replace(line, ri.source, ri.target);
		}

		return line;
	}

	private int parseInt(String num)
	{
		if (num == null)
			return -1;
		else
			return Integer.parseInt(num);
	}

	String prefix(String line)
	{
		return line.substring(0, textOffset) + text;
	}


	@Override
	public String toString()
	{
		if (book != -1)
			return String.format("%d.%d.%d", book, chap, verse);
		else
			return String.format("%d.%d", chap, verse);
	}
}
