package org.sanskritlibrary.admin;
import org.sanskritlibrary.XML;
import org.sanskritlibrary.Regex;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.util.ArrayList;

public class VerseInfoEx
{
    public String[] names = null;
    public int[] positions = null;
    public String[] values = null;
    public boolean[] transcode = null;
	public boolean isError = false;
	public ArrayList<ReplacementInfo> replacementInfo = new ArrayList<ReplacementInfo>();

	public VerseInfoEx(String line, Node versePattern) throws Exception
	{
		String pattern = XML.getOptionalValue(versePattern, "regexp", "");
		Node replacements = XML.getRelativeNode(versePattern, "replacements");
		if (replacements != null)
		{
			NodeList list = XML.getRelativeNodeList(replacements, "replacement");
			for (int i=0; i<list.getLength(); i++)
			{
				replacementInfo.add(new ReplacementInfo(list.item(i)));
			}
		}

        NodeList fields = XML.getRelativeNodeList(versePattern, "fields/field");
        int numFields = fields.getLength();
        names = new String[numFields];
        positions = new int[numFields];
        values = new String[numFields];
        transcode = new boolean[numFields];
        for (int i=0; i<numFields; i++)
        {
            Node field = fields.item(i);
            names[i] = field.getAttributes().getNamedItem("name").getNodeValue();
            positions[i] = Integer.parseInt(field.getAttributes().getNamedItem("pos").getNodeValue());
            Node n = field.getAttributes().getNamedItem("transcode");
            transcode[i] = n != null && n.getNodeValue().toLowerCase().equals("y");
        }

        line = fixup(line);
        
        Regex re = new Regex(line, pattern);
        if (re.find())
        {
          if (re.getGroupCount() != numFields)
              throw new Exception(String.format("regexp has %d groups but there are %d fields", re.getGroupCount(), numFields));
          else
          {
              for (int i=0; i<numFields;  i++)
              {
                int n = positions[i];
                if (n <= 0 || n > numFields)
                    throw new Exception(String.format("value of pos attribute must be between 1 and %d", numFields));
                else
                    values[i] = re.getGroup(n);
              }
          }
        }
        else
            isError = true;
	}


	public String fixup(String line)
	{
		for (ReplacementInfo ri : replacementInfo)
		{
			line = Regex.replace(line, ri.source, ri.target);
		}

		return line;
	}

	private int parseInt(String num)
	{
		if (num == null)
			return -1;
		else
			return Integer.parseInt(num);
	}

}
