package org.sanskritlibrary.webservice;

import java.net.Socket;
import java.net.URL;
import java.io.PrintWriter;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Hashtable;
import java.util.Enumeration;

public class UniversalProxy
{

	private byte[] b = null;
	private int numBytes = 0;
	private int contentStart = 0;
	private boolean headersFetched = false;
	private String error;
	private int bufferSize = 1024;
	private int bufferIncrement = 1024;
	String response;
	private String transferEncoding = null;
	private int contentLength = 0;

	public boolean getResponse(String method,
			  String target,
			  Hashtable<String, String> headers,
			  byte[] data)
	{
		InputStream in = null;
		PrintWriter out = null;
		Socket s = null;

		try
		{
			URL url = new URL(target);
			StringBuffer request = new StringBuffer();
			int portNumber = url.getPort();
			if (portNumber == -1)
				portNumber = 80;
			s = new Socket(url.getHost(), portNumber);
			OutputStream os = s.getOutputStream();
			out = new PrintWriter(os, true);

			String queryParams = headers.get("x-proxy-params");
			if (queryParams == null)
				queryParams = headers.get("X-Proxy-Params");
			if (queryParams == null)
				request.append(String.format("%s %s HTTP/1.1\r\n",
					  method, url.getPath()));
			else
				request.append(String.format("%s %s?%s HTTP/1.1\r\n",
					  method, url.getPath(), queryParams));


			// Build the first line of request

			// Set the request headers
			Enumeration<String> en = headers.keys();
			boolean connectionHeaderFound = false;
			while (en.hasMoreElements())
			{
				String headerName = en.nextElement();
				String lcname = headerName.toLowerCase();
				if (!lcname.equals("cookie"))
				{
					String value = null;
					if (lcname.equals("x-proxy-cookie"))
					{
						value = headers.get(headerName);
						headerName="Cookie";		// send x-proxy-cookies as cookies
					}
					else if (lcname.equals("connection"))
					{
						connectionHeaderFound = true;
						value = "close";
					}
					else
					{
						value = headers.get(headerName);
					}
					request.append(String.format("%s: %s\r\n", headerName, value));
				}
			}

			if (!connectionHeaderFound)
			{
				request.append("Connection: close\r\n");
			}

			request.append("\r\n");

			// Send request to server
            try {
                out.print(request.toString());
                out.flush(); // important!!
                if (data != null) {
                    os.write(data);
                }
                in = s.getInputStream();

                // Read response from server
                b = new byte[bufferSize];
                int n;
                while ((n = in.read()) != -1) {
                    if (numBytes == bufferSize) // increase buffer size
                    {
                        byte[] bb = new byte[bufferSize + bufferIncrement];
                        for (int i = 0; i < numBytes; i++) {
                            bb[i] = b[i];
                        }
                        b = bb;
                        bufferSize = b.length;
                    }
                    b[numBytes++] = (byte) n;
                }

                return true;
            } catch (Exception e) {
                System.out.println(e.getMessage());
                return false;
            }
		} catch (Exception ex)
		{
			error = String.format("%s: %s", ex.getClass().getName(), ex.getMessage());
			return false;
		} finally
		{
			try
			{
				if (in != null)
				{
					in.close();
				}
				if (out != null)
				{
					out.close();
				}
				if (s != null)
				{
					s.close();
				}
			} catch (Exception ex)
			{
			}
		}
	}


	public int getStatus()
	{
		return 200;
	}

	public Hashtable<String, String> getHeaders() throws Exception
	{
		headersFetched = true;
		Hashtable<String, String> h = new Hashtable<String, String>();
		String setProxyCookie = "";
		String delim = "";

		byte[] buffer = new byte[b.length];		// !! optimize this
		for (int i = 0, j = 0; i < numBytes; i++)
		{
			buffer[j++] = b[i];
			if (b[i] == 0x0a)
			{
				if (j == 2)  // have reached blank line
				{
					contentStart = i + 1;
					break;
				}
				String header = new String(buffer, 0, j - 2, "UTF-8");
				String[] field = header.split(":");
				if (field.length == 2)
				{
					boolean useHeader = true;
					String fieldName = field[0].trim();
					String lcname = fieldName.toLowerCase();		// for convenience
					String value = field[1].trim();
					if (lcname.equals("transfer-encoding"))
					{
						transferEncoding = field[1].trim().toLowerCase();
						useHeader = false;
					}
					else if (lcname.equals("content-length"))
						contentLength = Integer.parseInt(field[1].trim());
					else if (lcname.equals("set-cookie"))
					{
						setProxyCookie += delim + value;	// accumulate all cookies sent back from server
						delim += "; ";
					}

					if (useHeader)
						h.put(fieldName, value);
				}
				j = 0;
			}
		}
		
		if (setProxyCookie.length() > 0)
		{
			h.put("X-Set-Proxy-Cookie", setProxyCookie);
		}
		return h;
	}

	public ByteBuffer getContent() throws Exception
	{
		ByteBuffer bb = null;
		byte[] result = null;
		if (!headersFetched)
		{
			throw new Exception("Need to call getHeaders first");
		}

		if (transferEncoding == null)	// assume regular content
		{
			result = new byte[numBytes - contentStart];
			for (int i = 0; i < result.length; i++)
			{
				result[i] = b[contentStart + i];
			}
			bb = new ByteBuffer(result.length, result);
		}
		else if (transferEncoding.equals("chunked"))
		{
			// first compute combined size of chunks
			int len = 0;
			byte[] numbuf = new byte[10];
			int j = 0;
			for (int i=contentStart; i<numBytes; i++)
			{
				if (b[i] == 0x0a)  // at end of chunk size
				{
					int chunkSize = Integer.parseInt(new String(numbuf, 0, j).trim(), 16);
					len += chunkSize;
					i += chunkSize + 2;	// move to byte before next chunk length
					j = 0;			// reset numbuf index
				}
				else if (b[i] != 0x0d)
					numbuf[j++] = b[i];  // accumulate chunk size
			}

			// Now copy each chunk into result buffer
			result = new byte[len];
			int n = 0;
			for (int i=contentStart; i<numBytes; i++)
			{
				if (b[i] == 0x0a)  // at end of chunk size
				{
					int chunkSize = Integer.parseInt(new String(numbuf, 0, j).trim(), 16);
					j = 0;
					i++;
					for (int k=0; k<chunkSize; k++)
					{
						result[n++] = b[i++];
					}
					i++;   // should  be at 0x0a following chunk
				}
				else
					numbuf[j++] = b[i];	// accumulate chunk size
			}

			bb = new ByteBuffer(result.length, result);
		}
		else
		{
			throw new Exception(String.format("%s is unknown Transfer-Encoding", transferEncoding));
		}

		return bb;
	}

	public void setInitialBufferSize(int bufferSize)
	{
		this.bufferSize = bufferSize;
	}

	public void setBufferIncrement(int bufferIncrement)
	{
		this.bufferIncrement = bufferIncrement;
	}

	public String getError()
	{
		return error;
	}

	public int getBufferSize()
	{
		return numBytes;
	}
}
