package org.sanskritlibrary.webservice;

import org.sanskritlibrary.Regex;
import org.sanskritlibrary.FSM;

public class WebServices
{
	// Transform string s using transforms in tm.
	// The hashtable ish contains InputStreams for all required fsm xml files.
	// How the InputStreams were obtained depends on whether the caller is a Servlet
	// or a regular application either console or Swing.

	static public String transformString(String s, TransformMap tm, String regex) throws Exception
	{
		if (!tm.isEntireString() && regex != null)
			throw new Exception("Regex specfied for a transformation where map is not for entire string");
		StringBuilder sb = new StringBuilder();
		Regex re = new Regex(s, regex);
		int offset = 0;
		while (re.find())
		{
			String grp = re.getGroup(1);
			int n = re.getOffset(1);
			sb.append(s.substring(offset, n));
			offset = n + grp.length();
			sb.append(transformString(grp, tm));
		}
		sb.append(s.substring(offset));		// add remaining text
		return sb.toString();
	}


	static public String transformString(String s, TransformMap tm, int start, StringBuilder accumulateBuffer) throws Exception
	{
		boolean debugging = false;

		if (tm.isEntireString()) // No tags in s, i.e., map looks like *:encoding1->encoding2;
		{
            return doTransform(tm, s, start, accumulateBuffer);
		}

		// Else use tags in input string s to guide transformation.
		StringBuilder result = new StringBuilder();
		int i = 0;
		int len = s.length();
		// proceed left to right through s searching for tags mentioned in the
		// transform map. For each tagged area, convert it using appropriate FSM
		while (i < len)
		{
			boolean bTagFound = false;
			int startTagCandidate = s.substring(i).indexOf("<");
			if (startTagCandidate != -1)  // if we might have a tag
			{
				int idxStart = i + startTagCandidate;
				// Have a "<", now see if what follows it matches any of the tags in the TransformMap
				for (TransformMapEntry tme : tm)
				{
					// see what start tag is
					int tagLength = tme.tag.length();
					int last = idxStart + tagLength;
					if (last < len && s.substring(idxStart, last).equals(tme.tag))
					{
						int adj = 0;
						if (!tme.retain)  // if tag is not supposed to be in output
							adj = tagLength;
						// start tag recognized, now find end of tag
						String endTag = "</" + tme.tag.substring(1);
						int idxEnd = idxStart + s.substring(idxStart).indexOf(endTag);
						if (idxEnd > idxStart)  // if end tag found convert content
						{
							result.append(s.substring(i, last-adj)); // append stuff preceding tag that is not in any tag
							String text = s.substring(idxStart + tagLength, idxEnd);
							if (debugging)
								result.append(String.format("***%s***", text));
							else
                                result.append(doTransform(tm, text, start, accumulateBuffer));
							if (tme.retain)
								result.append(endTag);
							i = idxEnd + endTag.length();  // position i after end tag.
							bTagFound = true;
							break;
						}
					}
				}
				if (!bTagFound)
				{
					result.append(s.charAt(i)); // treat the < as data, not a tag delimeter
					i++;
				}
			}
			else  // "<" not found so copy rest of string to result
			{
				result.append(s.substring(i));
				break;
			}
		}
		return result.toString();
	}

    public static String transformString(String s, TransformMap tm) throws Exception
    {
        return transformString(s, tm, 0, null);
    }

    private static String doTransform(TransformMap tm, String s, int start, StringBuilder ab) throws Exception
    {
        TransformMapEntry tme = tm.iterator().next();

        if (ab != null)
            return FSM.transformString(tm, s, ab, tme.fsm[0]);           // only do the x-->slp1 part of the transformat
        else if (start == 1)
            return FSM.transformString(tm, s, ab, tme.fsm[1]);           // only do the slp1-->y part of the transform
        else            // do entire 
        {
            FSM fsmSource = tme.fsm[0];
			FSM fsmTarget = tme.fsm[1];
			return FSM.transformString(tm, s, ab, fsmSource, fsmTarget);
        }
    }
}
